/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@parc.xerox.com
				chiba@is.s.u-tokyo.ac.jp
*/

// Compile-time Meta Level

#include "rtmop.h"

#define BUF_NAME	"mBuffer"
#define INDEX_VAR	"mIndex"
#define METAOBJ		"MetaObj"
#define DISPATCH	"Reify"
#define TRAP		"Reflect"

RtmopClass::RtmopClass(Ptree* def, Ptree* meta)
: Class(def, nil)
{
    Ptree* name;

    if(Ptree::Match(meta, "[%?]", &name))
	rt_metaclass = name;
    else
	rt_metaclass = nil;
}

Ptree* RtmopClass::TranslateSelf(Environment* env)
{
    Ptree* code = nil;
    for(int i = 1; ; ++i){
	Ptree* name = NthMemberName(i);
	if(name == nil)
	    break;
	else
	    code = MakeDispatcher(env, i, name, code);
    }

    AppendAfterToplevel(Ptree::Make(
	"void %p::" DISPATCH "(int member, char* buffer){\n"
	"int i = 0;\n"
	"switch(member){\n"
        "%p }}",
	Name(), code));

    return Class::TranslateSelf(env);
}

Ptree* RtmopClass::TranslateBaseClasses(Environment* env, Ptree* bases)
{
    bases = Class::TranslateBaseClasses(env, bases);
    if(bases == nil){
	if(rt_metaclass == nil)
	    return Ptree::Make(": public " METAOBJ);
	else
	    return Ptree::Make(": public %p", rt_metaclass);
    }
    else{
	Class* baseclass = NthBaseClass(env, 1);
	if(baseclass == nil || strcmp(baseclass->MetaclassName(),
				      "RtmopClass") != 0)
	    ErrorMessage("The base class is not RtmopClass.", Name());

	return bases;
    }
}

Ptree* RtmopClass::TranslateBody(Environment* env, Ptree* body)
{
    Ptree* decl = Ptree::Make(
	"private: void " DISPATCH "(int, char*);");
    return Ptree::Append(body, decl);
}

Ptree* RtmopClass::MakeDispatcher(Environment* env, int nth, Ptree* name,
				  Ptree* code)
{
    TypeInfo ftype;
    TypeInfo argtype;
    int i;

    PtreeId whatis = name->WhatIs();
    if(whatis == LeafClassNameId || whatis == PtreeDestructorId)
	return code;		// constructor or destructor

    if(!LookupMemberType(env, name, ftype) || ftype.WhatIs() != FunctionType)
	return code;		// data member

    Ptree* args = nil;
    for(i = 1; ftype.NthArgument(i, argtype); ++i)
	args = Ptree::Snoc(args, LoadArgument(i, argtype));

    return Ptree::Make("%p case %d:{\n %p %p}\nbreak;\n",
		       code, nth, args,
		       StoreReterunValue(ftype, name, i - 1));
}

Ptree* RtmopClass::LoadArgument(int n, TypeInfo& argtype)
{
    Ptree* code;
    TypeInfoId whatis = argtype.WhatIs();
    if(whatis == BuiltInType)
	if(argtype.IsBuiltInType() & (FloatType | DoubleType))
	    return Ptree::Make("double p%d = *(double*)&buffer[i];"
			       "i+=sizeof(double);\n",
			       n);
	else
	    return Ptree::Make("int p%d = *(int*)&buffer[i];"
			       "i+=sizeof(int);\n",
			       n);
    else if(whatis == ClassType)
	return Ptree::Make("%p p%d(&buffer[i], i);\n",
			   argtype.ClassMetaobject()->Name(), n);
    else if(whatis == PointerType){
	Ptree* tname = argtype.MakePtree(nil);
	return Ptree::Make("%p p%d = *(%p*)&buffer[i];"
			   "i+=sizeof(%p);\n",
			   tname, n, tname, tname);
    }
    else{
	MopErrorMessage("sorry, cannot handle this type",
			"RtmopClass::LoadArgument()");
	return nil;
    }
}

Ptree* RtmopClass::StoreReterunValue(TypeInfo& ftype, Ptree* member_name,
				     int nargs)
{
    TypeInfo rtype;
    Ptree* code;

    ftype.Dereference(rtype);
    TypeInfoId whatis = rtype.WhatIs();
    if(whatis == BuiltInType){
	uint tag = rtype.IsBuiltInType();
	if(tag & (FloatType | DoubleType))
	    code = Ptree::Make("*(double*)buffer = %p(",
			       member_name);
	else if(tag & VoidType)
	    code = Ptree::Make("%p(", member_name);
	else
	    code = Ptree::Make("*(int*)buffer = %p(", member_name);
    }
    else if(whatis == ClassType)
	code = Ptree::Make("%p(", member_name);
    else if(whatis == PointerType)
	code = Ptree::Make("*(%p*)buffer = %p(",
			   rtype.MakePtree(nil), member_name);
    else if(whatis == UndefType)
	code = Ptree::Make("*(int*)buffer = %p(", member_name);
    else{
	MopErrorMessage("sorry, cannot handle this type",
			"RtmopClass::StoreReterunValue()");
	return nil;
    }

    if(nargs > 0){
	Ptree* args = Ptree::Make("p%d", nargs);
	while(--nargs > 0)
	    args = Ptree::Make("p%d,%p", nargs, args);

	code = Ptree::List(code, args);
    }

    if(whatis == ClassType)
	return Ptree::Make("%p).Store(buffer, i);", code);
    else
	return Ptree::Make("%p);", code);
}

Ptree* RtmopClass::TranslateMemberCall(Environment* env, Ptree* object,
				       Ptree* op, Ptree* member,
				       Ptree* arglist)
{
    TypeInfo ftype;
    TypeInfo argtype;
    TypeInfo rtype;
    uint tag;

    if(!LookupMemberType(env, member, ftype)){
	ErrorMessage("no such a member", member);
	return nil;
    }

    Ptree* code = nil;
    PtreeIter next(Ptree::Second(arglist));
    for(int i = 1; ftype.NthArgument(i, argtype); ++i){
	code = Ptree::Snoc(code, StoreArgument(env, argtype, next()));
	next();		// skip ,
    }

    code = Ptree::Make("(" INDEX_VAR "= 0, %p "
		       "%p%p" TRAP "(%d," BUF_NAME "," INDEX_VAR "),",
		       code, TranslateExpression(env, object), op,
		       IsMember(member));

    ftype.Dereference(rtype);
    switch(rtype.WhatIs()){
    case BuiltInType :
	tag = rtype.IsBuiltInType();
	if(tag & (FloatType | DoubleType))
	    code = Ptree::Make("%p *(double*)" BUF_NAME ")", code);
	else if(tag & VoidType)
	    code = Ptree::Make("%p 0)", code);
	else
	    code = Ptree::Make("%p *(int*)" BUF_NAME ")", code);
	break;
    case ClassType :
	code = Ptree::Make("%p %p(" BUF_NAME "))",
			   code,
			   rtype.ClassMetaobject()->Name());
	break;
    case PointerType :
	code = Ptree::Make("%p *(%p*)" BUF_NAME ")",
			   code, rtype.MakePtree(nil));
	break;
    case UndefType :
	code = Ptree::Make("%p *(int*)" BUF_NAME ")", code);
	break;
    default :
	code = Ptree::Make("%p 0)", code);	// error
	break;
    }

    return code;
}

Ptree* RtmopClass::StoreArgument(Environment* env, TypeInfo& argtype,
				 Ptree* expr)
{
    Ptree* t;
    expr = TranslateExpression(env, expr);
    switch(argtype.WhatIs()){
    case BuiltInType :
	if(argtype.IsBuiltInType() & (FloatType | DoubleType))
	    return Ptree::Make("*(double*)&" BUF_NAME "[" INDEX_VAR "]= %p,"
			       INDEX_VAR "+=sizeof(double),",
			       expr);
	else
	    return Ptree::Make("*(int*)&" BUF_NAME "[" INDEX_VAR "]= %p,"
			       INDEX_VAR "+=sizeof(int),",
			       expr);
    case ClassType :
	return Ptree::Make("(%p).Store(" BUF_NAME "," INDEX_VAR "),",
			   expr);
    case PointerType :
	t = argtype.MakePtree(nil);
	return Ptree::Make("*(%p*)&" BUF_NAME "[" INDEX_VAR "]= %p,"
			   INDEX_VAR "+=sizeof(%p),",
			   t, expr, t);
    default :
	return nil;
    }
}
