/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@parc.xerox.com
				chiba@is.s.u-tokyo.ac.jp
*/

#include "wrapper.h"

// TranslateBody() inserts declarations for wrapper functions.

Ptree* WrapperClass::TranslateBody(Environment* env, Ptree* body)
{
    Ptree* decl = Ptree::Make("public:\n");
    Ptree* name;
    TypeInfo t;
    int i = 1;
    while((name = NthMemberName(i++)) != nil){
	PtreeId whatis = name->WhatIs();
	if(whatis != LeafClassNameId && whatis != PtreeDestructorId)
	    if(LookupMemberType(env, name, t) && t.WhatIs() == FunctionType){
		// if the member is not a constructor, a destructor,
		// or a data member, insert the declaration for the wrapper.
		Ptree* m = t.MakePtree(WrapperName(name));
		decl = Ptree::Make("%p %p;\n", decl, m);
	    }
    }

    return Ptree::Append(body, decl);
}

Ptree* WrapperClass::WrapperName(Ptree* name)
{
    return Ptree::Make("_wrap_%p", name);
}

/*
  TranslateMemberFunctionBody() defines wrapper functions.
  They are defined just after the wrapped function is defined.
  Note that if they are defined in the header file, their duplicated
  copies would be unnecessarily produced.
*/
Ptree* WrapperClass::TranslateMemberFunctionBody(Environment* env,
						 Ptree* name, BOOL inlined,
						 Ptree* body)
{
    TypeInfo t;

    PtreeId whatis = name->WhatIs();
    if(whatis != LeafClassNameId && whatis != PtreeDestructorId)
	if(LookupMemberType(env, name, t))
	    MakeWrapper(env, name, WrapperName(name), t, inlined);

    return Class::TranslateMemberFunctionBody(env, name, inlined, body);
}

/*
  MakeWrapper() generates the member function such as:

  <return type> CLASS::WRAPPER_NAME(<type> p1, <type> p2, ...)
  {
      <The code retured by WrapperBody()>
  }
*/
void WrapperClass::MakeWrapper(Environment* env, Ptree* name,
			       Ptree* wrapper_name, TypeInfo& t, BOOL inlined)
{
    TypeInfo atype;
    TypeInfo rtype;
    int i;
    Ptree* arglist = nil;

    for(i = 1; t.NthArgument(i, atype); ++i){
	Ptree* arg = atype.MakePtree(Ptree::Make("p%d", i));
	if(i == 1)
	    arglist = arg;
	else
	    arglist = Ptree::Make("%p,%p", arglist, arg);
    }

    t.Dereference(rtype);
    Ptree* body = WrapperBody(env, name, wrapper_name, i - 1, t);
    Ptree* head = Ptree::Make("%p::%p(%p)", Name(), wrapper_name, arglist);
    if(rtype.WhatIs() != UndefType)
	head = rtype.MakePtree(head);

    if(inlined)
	AppendAfterToplevel(Ptree::Make("inline %p{%p}\n", head, body));
    else
	AppendAfterToplevel(Ptree::Make("%p{%p}\n", head, body));
}

/*
  WrapperBody() returns the body of the wrapper.  NAME is the name of
  the member function wrapped by this wrapper.  WRAPPER_NAME is the
  name of this wrapper.  NUM_OF_ARGS is the number of the arguments
  needed to call the wrapped member function.  The arguments are
  (p1, p2, ..., p<NUM_OF_ARGS>).

  This function returns a return statement:

      return <the wrapped function>(p1, p2, ...);
*/
Ptree* WrapperClass::WrapperBody(Environment* env, Ptree* name,
				 Ptree* wrapper_name, int num_of_args,
				 TypeInfo& ftype)
{
    // first make the argument list needed to call the wrapped function.

    Ptree* arglist = nil;
    while(num_of_args > 0){
	if(num_of_args > 1)
	    arglist = Ptree::Make(",p%d%p", num_of_args, arglist);
	else
	    arglist = Ptree::Make("p1%p", arglist);

	--num_of_args;
    }

    // then make a statement that calls the wrapped member function

    ftype.Dereference();
    if(ftype.IsBuiltInType() & VoidType)
	return Ptree::Make("%p(%p);", name, arglist);
    else
	return Ptree::Make("return %p(%p);", name, arglist);
}

//  TranslateMemberCall() replaces the called function with the wrapper
//  function.

Ptree* WrapperClass::TranslateMemberCall(Environment* env,
					 Ptree* object, Ptree* op,
					 Ptree* member, Ptree* arglist)
{
    return Class::TranslateMemberCall(env, object, op, WrapperName(member),
				      arglist);
}

Ptree* WrapperClass::TranslateMemberCall(Environment* env,
					 Ptree* member, Ptree* arglist)
{
    return Class::TranslateMemberCall(env, WrapperName(member), arglist);
}
