/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/

#include "binding.h"
#include "typeinfo.h"
#include "hash.h"
#include "ptree.h"
#include "syntax.h"
#include "mop.h"
#include "function.h"
#include "y.tab.h"

// class Environment

Environment::Environment()
{
    htable = new BigHashTable;		// base environment
    client_data = nil;
    metaclass_declarations = nil;
    class_info = nil;
    outermost_block = FALSE;
    inserted_decls = nil;
    next = nil;
}

Environment::Environment(Environment* e)
{
    htable = new HashTable;
    client_data = nil;
    metaclass_declarations = nil;
    class_info = nil;
    outermost_block = FALSE;
    inserted_decls = nil;
    next = e;
}

Environment::~Environment()
{
    delete htable;
    delete client_data;
    delete metaclass_declarations;
    delete next;
}

Environment* Environment::UnlinkTop()
{
    Environment* e = next;
    next = nil;
    return e;
}

BOOL Environment::IsEmpty()
{
    return BOOL(htable->IsEmpty());
}

Environment* Environment::GetBottom()
{
    Environment* p;
    for(p = this; p->next != nil; p = p->next)
	;

    return p;
}

BOOL Environment::Lookup(Ptree* name, TypeInfo& t)
{
    BOOL is_type_name;
    return Lookup(name, is_type_name, t);
}

/*
  Syntax::LookupType() called by Lookup() may cause a MOP error message
  "MOP error: ... in Environment::Lookup()".
*/
BOOL Environment::Lookup(Ptree* name, BOOL& is_type_name, TypeInfo& t)
{
    if(this == nil){
	MopErrorMessage("Environment::Lookup()", "nil enviornment");
	return FALSE;
    }
    else if(name == nil){
	t.Undef();
	return FALSE;
    }
    else
	return Syntax::LookupType(name, this, is_type_name, t);
}

Class* Environment::LookupClassMetaobject(Ptree* name)
{
    BOOL dummy;
    TypeInfo t;

    if(Lookup(name, dummy, t))
	return t.ClassMetaobject();
    else
	return nil;
}

BOOL Environment::LookupTop(const char* name, int len, Binding*& t)
{
    return htable->Lookup((char*)name, len, (HashValue*)&t);
}

Binding* Environment::LookupAll(Ptree* name)
{
    Binding* b = nil;

    if(!name->IsLeaf())
	MopErrorMessage("LookupAll()", "bomb!");

    if(LookupAll(name->GetPosition(), name->GetLength(), b))
	return b;
    else
	return nil;
}

int Environment::LookupAll(const char* name, int len, Binding*& t)
{
    Environment* p;

    for(p = this; p != nil; p = p->next)
	if(p->htable->Lookup((char*)name, len, (HashValue*)&t))
	    return TRUE;

    return FALSE;
}

int Environment::AddEntry(char* name, int len, Binding* b)
{
    htable->AddEntry(name, len, b);
}

int Environment::AddEntry(Ptree* qname, Binding* b)
{
    Environment* env = this;
    ClassTypeInfo* c;

    switch(qname->WhatIs()){
    case LeafNameId :
	break;
    case PtreeQnameId :
	qname = Ptree::First(Ptree::Last(qname));
	c = Syntax::LookupClassOfQname(qname, env);
	if(c != nil)
	    env = c->GetEnv();

	switch(qname->WhatIs()){
	case PtreeDestructorId :
	case PtreeOperatorFuncId :
	    // don't add this to the environment
	    return -1;
	default :
	    break;
	}
	break;
    case PtreeDestructorId :
    case PtreeOperatorFuncId :
    default :
	// don't add this to the environment
	return -1;
    }

    return env->AddEntry(qname->GetPosition(), qname->GetLength(), b);
}

Binding* Environment::MostEffectiveClass()
{
    for(Environment* env = this; env != nil; env = env->next)
	if(env->class_info != nil)
	    return env->class_info;

    return nil;
}

Ptree* Environment::LookupMetaclassDeclaration(Ptree* name)
{
    for(Environment* env = this; env != nil; env = env->next){
	PtreeArray* decls = env->metaclass_declarations;
	if(decls != nil){
	    int i = decls->Number();
	    while(--i >= 0)
		if(Ptree::Eq(name, Ptree::Second(decls->Ref(i))))
		    return decls->Ref(i);
	}
    }

    return nil;
}

void Environment::AddMetaclassDeclaration(Ptree* decl)
{
    if(metaclass_declarations == nil)
	metaclass_declarations = new PtreeArray;

    metaclass_declarations->Append(decl);
}

BOOL Environment::AddClientData(Class* metaobject, Ptree* key,
				Environment::ClientData* data)
{
    if(this == nil || LookupClientData(metaobject, key))
	return FALSE;
    else{
	if(data != nil)
	    data->Ref();

#ifdef DONT_GC
	ClientDataLink* entry = new ClientDataLink;
#else
	ClientDataLink* entry = new (NoGC) ClientDataLink;
#endif
	entry->next = client_data;
	entry->metaobject = metaobject;
	entry->key = key;
	entry->data = data;
	client_data = entry;
	return TRUE;
    }
}

Environment::ClientData* Environment::LookupClientData(Class* metaobject,
						       Ptree* key)
{
    for(Environment* env = this; env != nil; env = env->next)
	for(ClientDataLink* c = env->client_data; c != nil; c = c->next)
	    if(c->metaobject == metaobject && Ptree::Eq(key, c->key))
		return c->data;

    return nil;
}

BOOL Environment::DeleteClientData(ClientData* ent)
{
    for(Environment* env = this; env != nil; env = env->next){
	if(env->client_data == nil)
	    continue;
	else if(env->client_data->data == ent){
	    ClientDataLink* c = env->client_data;
	    env->client_data = env->client_data->next;
	    c->next = nil;
	    delete c;
	    return TRUE;
	}

	for(ClientDataLink* c = env->client_data; c->next != nil; c = c->next){
	    ClientDataLink* c2 = c->next;
	    if(c2->data == ent){
		c->next = c2->next;
		c2->next = nil;
		delete c2;
		return TRUE;
	    }
	}
    }

    return FALSE;
}

BOOL Environment::InsertDeclaration(Ptree* decl)
{
    return InsertDeclaration(decl, nil, nil, nil);
}

BOOL Environment::InsertDeclaration(Ptree* decl,
				    Class* c, Ptree* key, ClientData* ent)
{
    for(Environment* env = this; env != nil; env = env->next)
	if(env->outermost_block){
	    env->inserted_decls = Ptree::Snoc(env->inserted_decls, decl);
	    if(ent == nil)
		return TRUE;
	    else
		return env->AddClientData(c, key, ent);
	}

    Ptree::RecordBeforeToplevel(decl);
    if(ent == nil)
	return TRUE;
    else
	return GetBottom()->AddClientData(c, key, ent);
}

Environment::ClientDataLink::~ClientDataLink()
{
    if(data != nil)
	data->Delete();

    delete next;
}

Environment::ClientData::~ClientData() {}

void Environment::ClientData::Delete()
{
    if(--refcount <= 0)
	delete this;
}


// class Binding

Binding::~Binding() {}

Environment* Binding::GetEnv()
{
    MopErrorMessage("Binding::GetEnv()", "bomb!");
}

Class* Binding::GetMetaobject()
{
    MopErrorMessage("Binding::GetMetaobject()", "bomb!");
}

uint NameInfo::GetToken()
{
    return Identifier;
}

TypeInfoId VariableNameInfo::WhatIs()
{
    return VariableName;
}

TypeInfoId DataMemberNameInfo::WhatIs()
{
    return DataMemberName;
}

TypeInfoId ImmutableVarNameInfo::WhatIs()
{
    return ImmutableVarName;
}

TypeInfoId FunctionNameInfo::WhatIs()
{
    return FunctionName;
}

TypeInfoId MemberFunctionNameInfo::WhatIs()
{
    return MemberFunctionName;
}

TypeInfoId TypedefTypeInfo::WhatIs()
{
    return TypedefType;
}

uint TypedefTypeInfo::GetToken()
{
    return TypeName;
}

uint BaseTypeInfo::GetToken()
{
    return TypeName;
}

ClassTypeInfo::ClassTypeInfo(Ptree* t, Environment* e)
: BaseTypeInfo(t)
{
    env = e;
    metaobject = nil;
}

ClassTypeInfo::~ClassTypeInfo()
{
    delete env;
    delete metaobject;
}

TypeInfoId ClassTypeInfo::WhatIs()
{
    return ClassType;
}

Environment* ClassTypeInfo::GetEnv()
{
    return env;
}

Class* ClassTypeInfo::GetMetaobject()
{
    return metaobject;
}

uint ClassTypeInfo::GetToken()
{
    return ClassName;
}

/*
  UpdateInfo() is called when a class is redefined.
*/
void ClassTypeInfo::UpdateInfo(Ptree* t, Environment* e)
{
    if(env == nil){
	typespec = t;
	env = e;
    }
    else{
	e->UnlinkTop();
	delete e;
    }
}

TypeInfoId EnumTypeInfo::WhatIs()
{
    return EnumType;
}

TypeInfoId TemplateInfo::WhatIs()
{
    return TemplateType;
}

uint TemplateInfo::GetToken()
{
    return TemplateName;
}

TemplateFunctionInfo::TemplateFunctionInfo(Ptree* def, Environment* e)
: Binding(nil)
{
    metaobject = new TemplateFunction(def);
    env = e;
}

TemplateFunctionInfo::~TemplateFunctionInfo()
{
    delete metaobject;
}

TypeInfoId TemplateFunctionInfo::WhatIs()
{
    return TemplateFunctionType;
}

uint TemplateFunctionInfo::GetToken()
{
    return Identifier;
}

TypeVarInfo::TypeVarInfo(TypeInfo& t)
: Binding(nil), type(t)
{
}

TypeInfoId TypeVarInfo::WhatIs()
{
    return TypeVariable;
}

uint TypeVarInfo::GetToken()
{
    return TypeName;
}

TypeInfoId ThisVariableInfo::WhatIs()
{
    return ThisVariable;
}

uint ThisVariableInfo::GetToken()
{
    return THIS;
}
