/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/

#ifndef _binding_h
#define _binding_h

#include "types.h"
#include "gc/gc_cpp.h"
#include "typeinfo.h"

class Ptree;
class HashTable;
class Binding;
class TypeInfo;
class Class;
class PtreeArray;
class TemplateFunction;

/*
  An Environement object binds a name to:
  a) its declaration (a NameInfo object) if the name is a variable name
     or a typedef type name, or
  b) its definition (a BaseType object) if the name is a fundamental
     type name.
*/

class Environment {
public:
    BOOL Lookup(Ptree* name, BOOL& is_type_name, TypeInfo& t);
    BOOL Lookup(Ptree* name, TypeInfo& t);
    Class* LookupClassMetaobject(Ptree* name);

    class ClientData {
    public:

#ifndef DONT_GC
	void* operator new(size_t size) {
	    return GC_MALLOC_UNCOLLECTABLE(size);
	}

	void operator delete(void* obj) { GC_FREE(obj); }
#endif
	ClientData() { refcount = 0; }
	virtual ~ClientData();
	void Ref() { ++refcount; }
	void Delete();
    private:
	int refcount;
    };

    BOOL AddClientData(Class*, Ptree*, ClientData*);
    ClientData* LookupClientData(Class*, Ptree*);
    BOOL DeleteClientData(ClientData*);

    BOOL InsertDeclaration(Ptree*);
    BOOL InsertDeclaration(Ptree*, Class*, Ptree*, ClientData*);

protected:
    Environment();
    Environment(Environment* e);
    ~Environment();

    Environment* UnlinkTop();
    Environment* Next() { return next; }
    BOOL IsEmpty();

    Environment* GetBottom();

    BOOL LookupTop(const char* name, int len, Binding*& t);
    Binding* LookupAll(Ptree*);
    BOOL LookupAll(const char* name, int len, Binding*& t);
    BOOL AddEntry(char* key, int len, Binding* b);
    BOOL AddEntry(Ptree* name, Binding* b);

    void RecordClassInfo(Binding* info) { class_info = info; }
    Binding* MostEffectiveClass();

    Ptree* LookupMetaclassDeclaration(Ptree*);
    void AddMetaclassDeclaration(Ptree*);

protected:
    class ClientDataLink {
    public:
	~ClientDataLink();
	ClientDataLink*	next;
	Class*		metaobject;
	Ptree*		key;
	ClientData*	data;
    };

    Environment*	next;
    HashTable*		htable;
    PtreeArray*		metaclass_declarations;
    Binding*		class_info;	// nil if this is not a class env
    BOOL		outermost_block;
    Ptree*		inserted_decls;
    ClientDataLink*	client_data;

friend class Token;
friend class Syntax;
friend class Ptree;
friend class PtreeMetaclass;
friend class PtreeFunction;
friend class PtreeTemplateDecl;
friend class ClassTypeInfo;	// for destruction
friend class TemplateFunction;
};

class FunctionEnvironment : public Environment {
public:
    FunctionEnvironment(Environment* e) : Environment(e) {
	outermost_block = TRUE;
    }

    Ptree* InsertedDeclarations() { return inserted_decls; }
};


// the tags of class Binding are defined in typeinfo.h

class Binding : public gc_cleanup {
public:
    Binding(Ptree* p){ typespec = p; }
    virtual ~Binding();
    virtual TypeInfoId WhatIs() = nil;
    virtual Environment* GetEnv();
    virtual Class* GetMetaobject();

protected:
    /*
      GetToken() is called by the lexical analyzer.  It must return
      a token, such as ClassName, for the name associated with that type.
    */
    virtual uint GetToken() = nil;
    Ptree* typespec;

    friend class Token;
    friend class Syntax;
    friend class TypeInfo;
};

class NameInfo : public Binding {
protected:
    NameInfo(Ptree* p, Ptree* q) : Binding(p) { declarator = q; }
    uint GetToken();

    Ptree* declarator;

    friend class Syntax;
};

class VariableNameInfo : public NameInfo {
public:
    VariableNameInfo(Ptree* type, Ptree* decl) : NameInfo(type, decl) {}
    TypeInfoId WhatIs();
};

class DataMemberNameInfo : public VariableNameInfo {
public:
    DataMemberNameInfo(Ptree* t, Ptree* d) : VariableNameInfo(t, d) {}
    TypeInfoId WhatIs();
};

/*
  This class is used to avoid memory leak.
  Different pointers must not indicate an object of this class.
*/
class ImmutableVarNameInfo : public VariableNameInfo {
public:
    ImmutableVarNameInfo(ImmutableVarNameInfo* obj)
		: VariableNameInfo(obj->typespec, obj->declarator) {}
    ImmutableVarNameInfo(Ptree* type, Ptree* decl)
			: VariableNameInfo(type, decl) {}
    TypeInfoId WhatIs();
};

class FunctionNameInfo : public NameInfo {
public:
    FunctionNameInfo(Ptree* type, Ptree* decl) : NameInfo(type, decl) {}
    TypeInfoId WhatIs();
};

class MemberFunctionNameInfo : public FunctionNameInfo {
public:
    MemberFunctionNameInfo(Ptree* t, Ptree* d) : FunctionNameInfo(t, d) {}
    TypeInfoId WhatIs();
};

class TypedefTypeInfo : public NameInfo {
public:
    TypedefTypeInfo(Ptree* t, Ptree* d) : NameInfo(t, d) {}
    TypeInfoId WhatIs();

protected:
    uint GetToken();
};

class BaseTypeInfo : public Binding {
protected:
    BaseTypeInfo(Ptree* p) : Binding(p) {}
    uint GetToken();
};

class ClassTypeInfo : public BaseTypeInfo {
public:
    ClassTypeInfo(Ptree*, Environment*);
    ~ClassTypeInfo();
    TypeInfoId WhatIs();
    void UpdateInfo(Ptree*, Environment*);
    Environment* GetEnv();
    Class* GetMetaobject();

protected:
    uint GetToken();
    void SetMetaobject(Class* c) { metaobject = c; }

protected:
    Environment* env;
    Class* metaobject;

friend class PtreeMetaclass;
friend class PtreeTemplateDecl;
};

class EnumTypeInfo : public BaseTypeInfo {
public:
    EnumTypeInfo(Ptree* p) : BaseTypeInfo(p) {}
    TypeInfoId WhatIs();
};

class TemplateInfo : public ClassTypeInfo {
public:
    TemplateInfo(Ptree* p, Environment* e) : ClassTypeInfo(p, e) {}
    TypeInfoId WhatIs();

protected:
    uint GetToken();
};

class TemplateFunctionInfo : public Binding {
public:
    TemplateFunctionInfo(Ptree*, Environment*);
    ~TemplateFunctionInfo();
    TypeInfoId WhatIs();
    Environment* GetEnv() { return env; }
    TemplateFunction* GetTempFuncMetaobject() { return metaobject; }

protected:
    uint GetToken();

    Environment* env;
    TemplateFunction* metaobject;
};

// TypeVarInfo is used to implement function templates

class TypeVarInfo : public Binding {
public:
    TypeVarInfo(TypeInfo&);
    TypeInfoId WhatIs();
    TypeInfo& GetTypeInfo() { return type; }

protected:
    uint GetToken();

    TypeInfo type;
};

class ThisVariableInfo : public Binding {
public:
    ThisVariableInfo(ClassTypeInfo* c) : Binding(nil) {
	class_info = c;
    }
    TypeInfoId WhatIs();
    ClassTypeInfo* ClassInfo() { return class_info; }

protected:
    uint GetToken();

private:
    ClassTypeInfo* class_info;
};

#endif /* _binding_h */
