/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/

#include "ptree.h"
#include "typeinfo.h"
#include "binding.h"
#include "function.h"

TemplateFunction::TemplateFunction(Ptree* def)
{
    Ptree *p, *classname;

    definition = def;
    PtreeIter next(Ptree::Third(def));
    while((p = next()) != nil){
	if(Ptree::Match(p, "[[[class %?]] []]", &classname))
	    arguments.Append(classname);
        else
	    arguments.Append(nil);	// not class parameter

	next();		// skip ,
    }
}

Ptree* TemplateFunction::TranslateSelf(Environment* env)
{
    PtreeArray before, after;
    Ptree* def2;

    Ptree* decl = Ptree::Nth(definition, 4);
    Ptree* decl2 = decl->WalkMemberDeclaration(env, before, after);

    if(decl == decl2)
	def2 = definition;
    else{
	Ptree *keyword, *lbracket, *args, *rbracket;
	if(!Ptree::Match(definition, "[%? %? %? %? %_]",
			&keyword, &lbracket, &args, &rbracket)){
	    definition->Display();
	    MopErrorMessage("TemplateFunction::TranslateSelf()",
			    "bomb!");
	}

	def2 = new PtreeTemplateDecl(keyword, lbracket, args, rbracket,
				     decl2);
    }

    if(before.Number() > 0 || after.Number() > 0)
	def2 = Ptree::List(before.All(), def2, after.All());

    return def2;
}

Ptree* TemplateFunction::TranslateFuncall(Environment* template_env,
					  Environment* argenv, Ptree* fcall)
{
    Ptree *name, *args, *args2, *decl, *def2;

    Environment* newenv = Ptree::MakeEnvironment(template_env);
    name = Ptree::First(fcall);
    args = Ptree::Second(fcall);

    decl = Ptree::Nth(definition, 4);
    Ptree::RecordDeclarator(newenv, Ptree::First(decl),
			    Ptree::Second(decl));

    if(BindAllTypeVariables(newenv, argenv,
			    name, Ptree::Second(args), args2))
	args2 = Ptree::List(Ptree::First(args), args2, Ptree::Third(args));
    else
	args2 = args;

    def2 = MakeInstance(newenv, name, args2);
    Ptree::DeleteTopEnvironment(newenv);
    if(def2 != nil)
	return def2;
    else
	if(args == args2)
	    return fcall;
	else
	    return Ptree::List(name, args2);
}

BOOL TemplateFunction::BindAllTypeVariables(Environment* env,
					    Environment* argenv,
					    Ptree* name,
					    Ptree* arglist, Ptree*& arglist2)
{
    Ptree *expr, *expr2, *args2;
    TypeInfo ftype, fatype, aatype;  // function type, formal argument type,
				     // actual argument type

    if(!env->Lookup(name, ftype))
	MopErrorMessage("TemplateFunction::BindAllTypeVariables()", "bomb!");

    PtreeIter next(arglist);
    int i = 0;
    BOOL changed = FALSE;
    args2 = nil;
    while((expr = next()) != nil){
	aatype.Undef();
	expr2 = expr->WalkExpr(argenv, aatype);
	if(expr != expr2)
	    changed = TRUE;

	args2 = Ptree::Snoc(args2, expr2);
	ftype.NthArgument(i++, fatype);
	for(;;){
	    TypeInfoId whatis = fatype.WhatIs();
	    if(whatis == UndefType)
		break;
	    else if(whatis == ClassType){
		BindTypeVariable(env, fatype, aatype);
		break;
	    }
	    else{
		fatype.Dereference();
		aatype.Dereference();
	    }
	}

	expr = next();		// skip ,
	if(expr != nil)
	    Ptree::Snoc(args2, expr);
    }

    arglist2 = args2;
    return changed;
}

void TemplateFunction::BindTypeVariable(Environment* env,
					TypeInfo& fatype, TypeInfo& aatype)
{
    Ptree* name = fatype.FullTypeName();
    if(name == nil)
	MopErrorMessage("TemplateFunction::BindTypeVaraible()", "bomb!");

    if(aatype.WhatIs() == UndefType)
	return;

    /* If fatype is a qualified class name, we check only the first
       class name (e.g. if the name is X::Y::Z, we check only X.)
       This is a bug although I know. */

    if(name->WhatIs() == PtreeQClassNameId){
	Ptree* aname = aatype.FullTypeName();
	if(aname == nil || aname->WhatIs() != PtreeQClassNameId)
	    return;

	aname = Ptree::First(aname);
	if(!env->Lookup(aname, aatype))
	    return;

	name = Ptree::First(name);
    }

    for(int i = 0; i < arguments.Number(); ++i)
	if(Ptree::Eq(name, arguments[i])){
	    TypeVarInfo* type = new TypeVarInfo(aatype);
	    if(env->AddEntry(name->GetPosition(), name->GetLength(),
			     type) < 0)
		delete type;

	    break;
	}
}

Ptree* TemplateFunction::MakeInstance(Environment* env, Ptree* name,
				      Ptree* args)
{
    Ptree *e1, *e2, *e3, *e4, *decl, *decl2;
    PtreeArray before, after;

    if(!Ptree::Match(definition, "[%? %? %? %? %?]",
		     &e1, &e2, &e3, &e4, &decl))
	MopErrorMessage("TemplateFunction::MakeInstance()", "bomb!");

    decl2 = decl->WalkMemberDeclaration(env, before, after);

    if(before.Number() > 0)
	Ptree::RecordBeforeToplevel(before.All());

    if(after.Number() > 0)
	Ptree::RecordAfterToplevel(after.All());

    if(decl == decl2)
	return nil;	// don't have to instantiate
    else{
	Ptree* name2 = Ptree::GenSym();
	decl2 = Ptree::ReplaceAll(decl2, name, name2);
	Ptree* tempfunc = Ptree::Make("%p %p %p %p\n%p\n",
				      e1, e2, e3, e4, decl2);
	Ptree::RecordBeforeToplevel(tempfunc);
	return Ptree::Make("%p%p", name2, args);
    }
}
