/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/

#include <iostream.h>
#include <string.h>
#include "hash.h"
#include "gc/gc.h"
#include "types.h"

struct HashTableEntry {
    char*	key;		// nil: unused, -1: deleted
    HashValue	value;
};

/* convenient functions.  It makes a copy of character string. */

char* KeyString(char* key) {
    char* str = new char[strlen(key) + 1];
    strcpy(str, key);
    return str;
}

char* KeyString(char* key, int len) {
    char* str = new char[len + 1];
    bcopy(key, str, len);
    str[len] = '\0';
    return str;
}

char* ConcatKeyString(char* key1, char* key2, char* key3)
{
    int len = key1 != nil ? strlen(key1) : 0;
    len += key2 != nil ? strlen(key2) : 0;
    len += key3 != nil ? strlen(key3) : 0;
    char* str = new char[len + 1];
    str[0] = '\0';
    if(key1 != nil)
	strcat(str, key1);

    if(key2 != nil)
	strcat(str, key2);

    if(key3 != nil)
	strcat(str, key3);

    return str;
}

HashTable::HashTable()
{
    Size = 251;
    Prime2 = 127;
    MakeTable();
}

void HashTable::MakeTable()
{
#ifdef DONT_GC
    entries = new HashTableEntry[Size];
#else
    entries = (HashTableEntry*)GC_MALLOC_UNCOLLECTABLE(sizeof(HashTableEntry)
						       * Size);
#endif
    for(int i = 0; i < Size; ++i)
	entries[i].key = nil;
}

BigHashTable::BigHashTable() : HashTable(0)
{
    Size = 2053;	// 1021
    Prime2 = 1021;	// 509
    MakeTable();
}

HashTable::~HashTable()
{
    for(int i = 0; i < Size; ++i)
	if(entries[i].key != (char*)-1){
	    DeleteKeyString(entries[i].key);
	    // delete entries[i].value;		don't delete elements
	}

#ifdef DONT_GC
    delete [] entries;
#else
    GC_free(entries);
#endif
}

void HashTable::DeleteAll()
{
    for(int i = 0; i < Size; ++i)
	if(entries[i].key != (char*)-1){
	    DeleteKeyString(entries[i].key);
#ifdef DONT_GC
	    delete entries[i].value;
#endif
	    entries[i].value = nil;
	}

#ifdef DONT_GC
    delete [] entries;
#else
    GC_free(entries);
#endif
}

int HashTable::IsEmpty()
{
    for(int i = 0; i < Size; ++i)
	if(entries[i].key != (char*)-1)
	    return FALSE;

    return TRUE;
}

int HashTable::Lookup(char* key, HashValue* value)
{
    int i;
    return Lookup2(key, value, &i);
}

int HashTable::Lookup(char* key, int len, HashValue* value)
{
    int i;
    return Lookup2(key, len, value, &i);
}

int HashTable::Lookup2(char* key, HashValue* value, int* index)
{
    unsigned int p = StringToInt(key);
    for(int i = 0; i < Size; ++i){
	int j = HashFunc(p, i);
	if(entries[j].key == nil){
	    return FALSE;		// not found
	}
	else if(entries[j].key != (char*)-1
				&& strcmp(entries[j].key, key) == 0){
	    *value = entries[j].value;
	    *index = j;
	    return TRUE;
	}
    }

    return FALSE;
}

int HashTable::Lookup2(char* key, int len, HashValue* value, int* index)
{
    unsigned int p = StringToInt(key, len);
    for(int i = 0; i < Size; ++i){
	int j = HashFunc(p, i);
	if(entries[j].key == nil){
	    return FALSE;		// not found
	}
	else if(entries[j].key != (char*)-1
		&& strncmp(entries[j].key, key, len) == 0
		&& entries[j].key[len] == '\0'){
	    *value = entries[j].value;
	    *index = j;
	    return TRUE;
	}
    }

    return FALSE;
}

// AddEntry adds a new entry to the hash table.
// If succeeding, this returns an index of the added entry, otherwise -1.
// Because `key' is duplicated, you can delete `key' later on.

int HashTable::AddEntry(char* key, HashValue value, int* index)
{
    unsigned int p = StringToInt(key);
    for(int i = 0; i < Size; ++i){
	int j = HashFunc(p, i);
	if(entries[j].key == nil || entries[j].key == (char*)-1){
	    entries[j].key = KeyString(key);
	    entries[j].value = value;
	    if(index != nil)
		*index = j;

	    return j;
	}
	else if(strcmp(entries[j].key, key) == 0){
	    if(index != nil)
		*index = j;

	    return -1;		// it is already registered.
	}
    }

    cerr << "HashTable overflow (key: " << key << ")\nPanic...\n";
    if(index != nil)
	*index = 0;		// no meaning

    return -1;
}

int HashTable::AddEntry(char* key, int len, HashValue value, int* index)
{
    unsigned int p = StringToInt(key, len);
    for(int i = 0; i < Size; ++i){
	int j = HashFunc(p, i);
	if(entries[j].key == nil || entries[j].key == (char*)-1){
	    entries[j].key = KeyString(key, len);
	    entries[j].value = value;
	    if(index != nil)
		*index = j;

	    return j;
	}
	else if(strncmp(entries[j].key, key, len) == 0
		&& entries[j].key[len] == '\0'){
	    if(index != nil)
		*index = j;

	    return -1;		// it is already registered.
	}
    }

    cerr << "HashTable overflow (key: ";
    for(int i = 0; i < len; ++i)
	cerr << key[i];

    cerr << ")\nPanic...\n";
    if(index != nil)
	*index = 0;		// no meaning

    return -1;
}

HashValue HashTable::Peek(int index)
{
    return entries[index].value;
}

void HashTable::ReplaceValue(int index, HashValue val)
{
    if(0 <= index && index < Size)
	entries[index].value = val;
    else
	cerr << "HashTable: invalid index (" << index << ")\n";
}

int HashTable::RemoveEntry(char* key)
{
    HashValue	u;
    int		index;

    if(!Lookup2(key, &u, &index))
	return FALSE;		// not found
    else{
	DeleteKeyString(entries[index].key);
	// delete entries[index].value;
	entries[index].key = (char*)-1;
	return TRUE;
    }
}

int HashTable::RemoveEntry(char* key, int len)
{
    HashValue	u;
    int		index;

    if(!Lookup2(key, len, &u, &index))
	return FALSE;		// not found
    else{
	DeleteKeyString(entries[index].key);
	// delete entries[index].value;
	entries[index].key = (char*)-1;
	return TRUE;
    }
}

unsigned int HashTable::StringToInt(char* key)
{
    if(key == nil)
	return 0;

    unsigned int p = 0;
    int i, j;
    for(i = j = 0; key[i] != '\0'; ++i, ++j){
	if(j >= sizeof(unsigned int) * 8 - 7)
	    j = 0;

	p += key[i] << j;
    }

    return p;
}

unsigned int HashTable::StringToInt(char* key, int len)
{
    if(key == nil)
	return 0;

    unsigned int p = 0;
    int i, j;
    for(i = j = 0; i < len; ++i, ++j){
	if(j >= sizeof(unsigned int) * 8 - 7)
	    j = 0;

	p += key[i] << j;
    }

    return p;
}

int HashTable::HashFunc(unsigned int p, int n)
{
    return((p + (unsigned int)n * (p % Prime2)) % Size);
}


#ifdef TEST

#include <stdio.h>

int main()
{
    char	buf[128];
    int		v;
    HashTable	t;

    while(gets(buf) != NULL){
	unsigned int k = t.StringToInt(buf);
	printf("key %s (%d): %d\n", buf, k, t.HashFunc(k, 0));
	printf("key %s (%d): %d\n", buf, k, t.HashFunc(k, 1));

	printf("add %d, ", t.AddEntry(buf, 1));
	printf("lookup %d", t.Lookup(buf, v));
	printf("(%d)\n", v);
    }
}

#endif /* TEST */
