/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/

#include <stdio.h>
#include <fstream.h>
#include <string.h>
#include <unistd.h>
#include "token.h"
#include "buffer.h"
#include "mop.h"

#define VERSION		"2.4a"

#define COPYING		"Copyright (c) 1995, 1996 Xerox Corporation. All Rights Reserved."

#define COMPILER	"g++"

// g++ recognizes a .ii file as preprocessed C++ source code.

#define OUTPUT_EXT	".ii"
#define CPP_EXT		".occ"

extern "C" {
    int execvp(...);
    int wait(int*);
}

const NARGS = 16;

static BOOL showProgram;
static BOOL doCompile;
static BOOL verboseMode;
static BOOL makeExecutable;

static int cppArgc = 0;
static char* cppArgv[NARGS];
static int ccArgc = 0;
static char* ccArgv[NARGS];

extern long unsigned int GC_gc_no;

static char* ParseOptions(int argc, char** argv);
static void AddCppOption(char* arg);
static void RecordSourceFile(char* arg, char*& source_file);
static void AddCcOption(char* arg);
static void StripSuffix(char* str);

main(int argc, char** argv)
{
    static char errmsg[] = " Error(s).  OpenC++ Stops.\n";
    char dest[64];

    char* src = ParseOptions(argc, argv);
    if(src == nil){
	cerr << "** No source file.  Read from stdin.\n";
	Token::SetSource(new ProgramFromStdin());
	yyparse();
	if(!showProgram)
	    if(NumOfErrors() == 0)
		Token::WriteText(cout, "stdout");
	    else{
		cerr << NumOfErrors() << errmsg;
		exit(1);
	    }
    }
    else{
	FILE* sfptr;
	FILE* dfptr;
	int status;

	char* cppfile = new char[strlen(src) + sizeof(CPP_EXT)];
	strcpy(cppfile, src);
	StripSuffix(cppfile);
	strcat(cppfile, CPP_EXT);

	char* dest = new char[strlen(src) + sizeof(OUTPUT_EXT)];
	strcpy(dest, src);
	StripSuffix(dest);
	strcat(dest, OUTPUT_EXT);

	AddCppOption("-D__opencxx");
	AddCppOption("-E");
	AddCppOption("-o");
	AddCppOption(cppfile);
	AddCppOption(src);
	AddCppOption((char*)0);

	if(verboseMode)
	    cerr << "Preprocess...";

	if(fork() == 0){
	    execvp(COMPILER, cppArgv);
	    perror("cannot invoke a compiler");
	}
	else{
	    wait(&status);
	    if(status != 0)
		exit(1);
	}

	if((sfptr = fopen(cppfile, "r")) == NULL){
	    perror(src);
	    exit(1);
	}

	if((dfptr = fopen(dest, "w")) == NULL){
	    perror(dest);
	    exit(1);
	}

	Token::SetSource(new ProgramFile(sfptr), cppfile);
	if(verboseMode)
	    cerr << "Translate...";

	yyparse();

	if(NumOfErrors() == 0){
	    ofstream out(fileno(dfptr));
	    Token::WriteText(out, dest);
	}
	else{
	    cerr << NumOfErrors() << errmsg;
	    exit(1);
	}

	fclose(dfptr);
	fclose(sfptr);

	unlink(cppfile);

	if(doCompile){
	    if(!makeExecutable)
		AddCcOption("-c");

	    AddCcOption(dest);
	    AddCcOption((char*)0);

	    if(verboseMode)
		cerr << "Compile...";

	    if(fork() == 0){
		execvp(COMPILER, ccArgv);
		perror("cannot invoke a compiler");
	    }
	    else{
		wait(&status);
		if(status != 0)
		    exit(1);
	    }
	}

	delete dest;
	delete cppfile;
    }

    if(verboseMode){
	cerr << "done.\n";
#ifndef DONT_GC
	cerr << "GC happened " << GC_gc_no << " times.\n";
#endif
    }

    exit(0);
}

static char* ParseOptions(int argc, char** argv)
{
    char* source_file = nil;
    showProgram = FALSE;
    doCompile = TRUE;
    verboseMode = FALSE;
    makeExecutable = TRUE;

    AddCppOption(COMPILER);
    AddCcOption(COMPILER);

    for(int i = 1; i < argc; ++i)
	if(strcmp("-E", argv[i]) == 0)
	    doCompile = FALSE;
	else if(strcmp("-c", argv[i]) == 0)
	    makeExecutable = FALSE;
	else if(strcmp("-d", argv[i]) == 0)	// hidden option
	    Token::ignore_line_directive = TRUE;
	else if(strcmp("-l", argv[i]) == 0){
	    cout << "[Loaded metaclasses...]\n";
	    opcxx_ListOfMetaclass::PrintAllMetaclasses();
	    exit(0);
	}
	else if(strcmp("-s", argv[i]) == 0){
	    showProgram = TRUE;
	    ShowProgram(TRUE);
	}
	else if(strcmp("-v", argv[i]) == 0)
	    verboseMode = TRUE;
	else if(strcmp("--", argv[i]) == 0){
	    while(++i < argc)
		RecordSourceFile(argv[i], source_file);

	    break;
	}
	else if(*argv[i] == '-'
		&& (argv[i][1] == 'D' || argv[i][1] == 'I'))
	    AddCppOption(argv[i]);
	else if(*argv[i] == '-'){
	    cerr << "OpenC++ version " VERSION "\n" COPYING "\nUsage: "
		 << argv[0]
		 << " [-l][-s][-v][-E][-c][-I<directory>][-D<macro>]\n";
	    exit(1);
	}
	else
	    RecordSourceFile(argv[i], source_file);

    return source_file;
}

static void AddCppOption(char* arg)
{
    if(cppArgc < NARGS)
	cppArgv[cppArgc++] = arg;
    else{
	cerr << "too many arguments\n";
	exit(1);
    }
}

static void RecordSourceFile(char* arg, char*& source_file)
{
    static char* errmsg
	= "more than one .cc files are not accepted at a time.\n";

    if(arg != nil && arg[0] != '-'){
	int len = strlen(arg);
	if(len >= 3){
	    char* ext = arg + len - 3;
	    if(ext[0] == '.' && ext[1] == 'c' && ext[2] == 'c')
		if(source_file == nil){
		    source_file = arg;
		    return;
		}
		else{
		    cerr << errmsg;
		    exit(1);
		}
	}
    }

    AddCcOption(arg);
    return;
}

static void AddCcOption(char* arg)
{
    if(ccArgc < NARGS)
	ccArgv[ccArgc++] = arg;
    else{
	cerr << "too many arguments\n";
	exit(1);
    }
}

static void StripSuffix(char* str)
{
    char* dot = &str[strlen(str)];
    for(char* ptr = str; *ptr != '\0'; ++ptr)
	if(*ptr == '.')
	    dot = ptr;

    *dot = '\0';
}

