/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/

#include <iostream.h>
#include <string.h>
#include "mop.h"
#include "ptree.h"
#include "token.h"
#include "syntax.h"
#include "y.tab.h"
#include "quote-class.h"

Class* Class::class_list = nil;

static Ptree* TranslateBases(Environment* env, Ptree* bases);

Class::Class(Ptree* def, Ptree* meta_args)
{
    definition = def;
    if(meta_args != nil)
	WarningMessage("meta arguments are ignored.");

    next = class_list;
    class_list = this;
}

Ptree* Class::Name()
{
    return Ptree::Second(definition);
}

Ptree* Class::BaseClasses()
{
    return Ptree::Third(definition);
}

Ptree* Class::Members()
{
    return Ptree::Second(Ptree::Nth(definition, 3));
}

char* Class::MetaclassName()
{
    return "Class";
}

Class* Class::NthBaseClass(Environment* env, int n)
{
    PtreeIter next(BaseClasses());
    Ptree* p;
    BOOL is_type;
    TypeInfo t;

    next();	// skip :
    while(p = next()){
	if(n-- == 0)
	    if(env->Lookup(Ptree::First(Ptree::Last(p)), is_type, t))
		return t.ClassMetaobject();
	    else
		return nil;

	next();	// skip ,
    }

    return nil;
}

Ptree* Class::NthMemberName(int nth)
{
    return Ptree::FindNthMemberName(Members(), nth + 1);
    // Note: In FindNthMemberName(), nth starts from 1.
}

int Class::IsMember(Ptree* name)
{
    return Ptree::FindMemberName(Members(), name) - 1;
}

BOOL Class::LookupMemberType(Environment* env, Ptree* name,
				   TypeInfo& mem_type)
{
    BOOL is_type;
    TypeInfo class_type;

    mem_type.Undef();
    if(env != nil && env->Lookup(Name(), is_type, class_type))
	return BOOL(Syntax::LookupMemberType(name, class_type, mem_type)
		    != UndefType);
    else
	return FALSE;
}

// This metaobject cannot handle templates

BOOL Class::AcceptTemplate()
{
    return FALSE;
}

/*
   At the beginning, Initialize() is once called on each metaclass.
   The subclasses of Class may define their own Initialize().
   Initialize() returns FALSE if the initialization fails.
*/
BOOL Class::Initialize()
{
    return TRUE;
}

/*
  Finalize() is called on each metaobject at the end.  The returned
  code is appended to the generated file.
*/
Ptree* Class::Finalize()
{
    return nil;
}

Ptree* Class::TranslateNewType(Environment* env, Ptree* type_name)
{
    if(type_name->WhatIs() == PtreeParenId){
	Ptree* t = Ptree::Second(type_name);
	Ptree* t2 = Ptree::WalkCastType(env, t);
	if(t == t2)
	    return type_name;
	else
	    return new PtreeParen(Ptree::First(type_name), t2,
				  Ptree::Third(type_name));
    }
    else
	return Ptree::WalkCastType(env, type_name);
}

Ptree* Class::TranslateArgumentList(Environment* env, Ptree* list)
{
    Ptree* args2;

    Ptree* args = Ptree::Second(list);
    if(args == nil)
	return list;

    // args may be a new's placement, which is an expression
    // in the OpenC++ grammer.
    if(IsArglist(args))
	args2 = Ptree::WalkExprList(env, args);
    else
	return TranslateExpression(env, list);

    if(args == args2)
	return list;
    else
	return new PtreeParen(Ptree::First(list), args2, Ptree::Third(list));
}

BOOL Class::IsArglist(Ptree* args)
{
    if(args == nil)
	return TRUE;
    else if(args->IsLeaf())
	return FALSE;
    else if(Ptree::Length(args) == 1)
	return TRUE;
    else if(args->WhatIs() == PtreeInfixExprId)
	return FALSE;
    else
	return Ptree::Eq(Ptree::Second(args), ',');
}

Ptree* Class::TranslateIndex(Environment* env, Ptree* index)
{
    Ptree* expr = Ptree::Second(index);
    Ptree* expr2 = TranslateExpression(env, expr);
    if(expr == expr2)
	return index;
    else
	return new PtreeBracket(Ptree::First(index), expr2,
				Ptree::Third(index));
}

Ptree* Class::TranslateExpression(Environment* env, Ptree* expr)
{
    TypeInfo t;

    return TranslateExpression(env, expr, t);
}

Ptree* Class::TranslateExpression(Environment* env, Ptree* expr,
				  TypeInfo& type)
{
    if(expr == nil)
	return expr;
    else{
	type.Undef();
	return expr->WalkExpr(env, type);
    }
}

Ptree* Class::TranslateStatement(Environment* env, Ptree* expr)
{
    if(expr == nil)
	return expr;
    else
	return expr->WalkStatement(env);
}

void Class::InsertBeforeToplevel(Ptree* p)
{
    Ptree::RecordBeforeToplevel(p);
}

void Class::AppendAfterToplevel(Ptree* p)
{
    Ptree::RecordAfterToplevel(p);
}

void Class::RegisterNewModifier(char* str)
{
    if(!Token::RecordKeyword(str, UserKeyword))
	MopErrorMessage("Class::RegisterNewModifier(): already used", str);
}

void Class::RegisterNewAccessSpecifier(char* str)
{
    if(!Token::RecordKeyword(str, UserKeyword4))
	MopErrorMessage("Class::RegisterNewAccessSpecifier(): already used",
			str);
}

void Class::RegisterNewWhileStatement(char* str)
{
    if(!Token::RecordKeyword(str, UserKeyword))
	MopErrorMessage("Class::RegisterNewWhileStatement(): already used",
			str);
}

void Class::RegisterNewForStatement(char* str)
{
    if(!Token::RecordKeyword(str, UserKeyword3))
	MopErrorMessage("Class::RegisterNewForStatement(): already used",
			str);
}

void Class::RegisterNewClosureStatement(char* str)
{
    if(!Token::RecordKeyword(str, UserKeyword2))
	MopErrorMessage("Class::RegisterNewClosureStatement(): already used",
			str);
}

void Class::RegisterMetaclass(char* keyword, char* class_name)
{
    if(!PtreeMetaclass::RegisterKeyword(keyword, class_name))
	MopErrorMessage("Class::RegisterMetaclass(): "
			"the keyword is already used",
			keyword);
}

// TranslateSelf() cannot be overridden.

Ptree* Class::TranslateSelf(Environment* env)
{
    Ptree *head, *name, *bases, *lparen, *body, *rparen;
    Ptree *name2, *bases2, *body2;
    PtreeArray before, after;

    if(!Ptree::Match(Definition(), "[%? %? %? [%? %? %?]]",
		     &head, &name, &bases, &lparen, &body, &rparen))
	MopErrorMessage("Class::TranslateSelf()", "bomb!");

    name2 = TranslateClassName(env, nil, name);
    bases2 = TranslateBaseClasses(env, bases);
    body2 = TranslateBody(env, Ptree::WalkClassBody(env, body, before, after));
    InsertBeforeToplevel(before.All());
    AppendAfterToplevel(after.All());
    if(name == name2 && bases == bases2 && body == body2)
	return Definition();
    else
	return Ptree::List(head, name2, bases2,
			   new PtreeBrace(lparen, body2, rparen));
}

Ptree* Class::TranslateClassName(Environment*,
				 Ptree* keyword, Ptree* name)
{
    if(keyword != nil)
	ErrorMessage("unsupported user keyword", keyword);

    return name;
}

Ptree* Class::TranslateBaseClasses(Environment* env, Ptree* bases)
{
    if(bases == nil)
	return bases;
    else{
	Ptree* b = Ptree::Rest(bases);
	Ptree* b2 = TranslateBases(env, b);
	if(b == b2)
	    return bases;
	else
	    return Ptree::Cons(Ptree::First(bases), b2);
    }
}

static Ptree* TranslateBases(Environment* env, Ptree* bases)
{
    Ptree *rest, *rest2, *base, *basename, *basename2;
    Class* metaobject;
    BOOL is_type;
    TypeInfo t;

    if(bases == nil)
	return nil;
    else{
	rest = Ptree::ListTail(bases, 2);
	rest2 = TranslateBases(env, rest);
	base = Ptree::First(bases);
	basename = basename2 = Ptree::First(Ptree::Last(base));
	if(env->Lookup(basename, is_type, t)){
	    metaobject = t.ClassMetaobject();
	    if(metaobject != nil)
		basename2 = metaobject->TranslateClassName(env, nil, basename);
	}

	if(basename != basename2){
	    Ptree *fst, *snd;
	    fst = Ptree::First(base);
	    snd = Ptree::Second(base);
	    if(fst == basename)
		base = Ptree::List(basename2);
	    else if(snd == basename)
		base = Ptree::List(fst, basename2);
	    else
		base = Ptree::List(fst, snd, basename2);
	}
	else if(rest == rest2)
	    return bases;

	if(rest2 == nil)
	    return Ptree::List(base);
	else
	    return Ptree::Cons(base, Ptree::Cons(Ptree::Second(bases), rest2));
    }
}

// body may be partly translated by other metaobjects.

Ptree* Class::TranslateBody(Environment*, Ptree* body)
{
    return body;
}

Ptree* Class::TranslateMemberFunctionBody(Environment* env, Ptree* name,
					  BOOL inlined, Ptree* body)
{
    return PtreeBrace::WalkStatements(env, body);
}

// INIT is "= <expression>" or "( <expression> )".
//
Ptree* Class::TranslateInitializer(Environment* env, Ptree* init,
				   PtreeArray& before, PtreeArray& after)
{
    Ptree* expr = Ptree::Second(init);
    return Ptree::Substitute(init, 2, TranslateExpression(env, expr), expr);
}

Ptree* Class::TranslateAssign(Environment* env, Ptree* object, Ptree* op,
			      Ptree* expr)
{
    return Ptree::List(TranslateExpression(env, object),
		       op,
		       TranslateExpression(env, expr));
}

// overloaded function call operator ()

Ptree* Class::TranslateFunctionCall(Environment* env, Ptree* object,
				    Ptree* args)
{
    return Ptree::List(TranslateExpression(env, object),
		       TranslateArgumentList(env, args));
}

Ptree* Class::TranslateSubscript(Environment* env, Ptree* object,
				 Ptree* index)
{
    return Ptree::List(TranslateExpression(env, object),
		       TranslateIndex(env, index));
}

Ptree* Class::TranslatePostfix(Environment* env, Ptree* object,
				 Ptree* op)
{
    return Ptree::List(TranslateExpression(env, object), op);
}

Ptree* Class::TranslateUnary(Environment* env, Ptree* op, Ptree* object)
{
    return Ptree::List(op, TranslateExpression(env, object));
}

Ptree* Class::TranslateBinary(Environment* env, Ptree* lexpr, Ptree* op,
			      Ptree* rexpr)
{
    return Ptree::List(TranslateExpression(env, lexpr), op,
		       TranslateExpression(env, rexpr));
}

Ptree* Class::TranslateNew(Environment* env, Ptree* header,
			   Ptree* op, Ptree* placement, Ptree* tname,
			   Ptree* arglist)
{
    if(header != nil && !header->Eq("::"))
	ErrorMessage("unsupported user keyword", op);

    Ptree* placement2 = TranslateArgumentList(env, placement);
    Ptree* tname2 = TranslateNewType(env, tname);
    if(arglist == nil)
	return Ptree::List(header, op, placement2, tname2);
    else
	return Ptree::List(header, op, placement2, tname2,
			   TranslateArgumentList(env, arglist));
}

Ptree* Class::TranslateMemberRead(Environment* env, Ptree* object,
				  Ptree* op, Ptree* member)
{
    return Ptree::List(TranslateExpression(env, object), op, member);
}

Ptree* Class::TranslateMemberRead(Environment* env, Ptree* member)
{
    return member;
}

Ptree* Class::TranslateMemberWrite(Environment* env, Ptree* object,
				   Ptree* op, Ptree* member, Ptree* assign_op,
				   Ptree* expr)
{
    return Ptree::List(TranslateExpression(env, object),
		       op, member, assign_op,
		       TranslateExpression(env, expr));
}

Ptree* Class::TranslateMemberWrite(Environment* env, Ptree* member,
				   Ptree* assign_op, Ptree* expr)
{
    return Ptree::List(member, assign_op, TranslateExpression(env, expr));
}

Ptree* Class::TranslateMemberCall(Environment* env, Ptree* object,
				  Ptree* op, Ptree* member, Ptree* arglist)
{
    return Ptree::List(TranslateExpression(env, object), op, member,
		       TranslateArgumentList(env, arglist));
}

Ptree* Class::TranslateMemberCall(Environment* env,
				  Ptree* member, Ptree* arglist)
{
    return Ptree::List(member, TranslateArgumentList(env, arglist));
}

Ptree* Class::TranslateUserStatement(Environment* env, Ptree* object,
				     Ptree* op,
				     Ptree* keyword, Ptree* rest)
{
    ErrorMessage("unsupported user statement", keyword);
    return nil;
}

Ptree* Class::TranslateStaticUserStatement(Environment* env,
					   Ptree* keyword, Ptree* rest)
{
    ErrorMessage("unsupported user statement", keyword);
    return nil;
}

//
// class opcxx_ListOfMetaclass	--- not documented class
//

opcxx_ListOfMetaclass* opcxx_ListOfMetaclass::head = nil;

static Class* CreateClass(Ptree* def, Ptree* marg)
{
    return new Class(def, marg);
}

static opcxx_ListOfMetaclass classCreator("Class", CreateClass,
				          Class::Initialize());

static Class* CreateQuoteClass(Ptree* def, Ptree* marg)
{
    return new QuoteClass(def, marg);
}

static opcxx_ListOfMetaclass QuoteClassCreator("QuoteClass",
					       CreateQuoteClass,
					       QuoteClass::Initialize());

opcxx_ListOfMetaclass::opcxx_ListOfMetaclass(char* n,
					     opcxx_MetaclassCreator c,
					     BOOL success)
{
    if(!success)
	MopErrorMessage("the initialization process failed", n);

    proc = c;
    name = n;
    if(AlreadyRecorded(n))
	next = nil;
    else{
	next = head;
	head = this;
    }
}

Class* opcxx_ListOfMetaclass::New(Ptree* name, Ptree* def, Ptree* marg)
{
    if(name != nil){
	opcxx_ListOfMetaclass* p = head;
	while(p != nil){
	    if(name->Eq(p->name))
		return (*p->proc)(def, marg);
	    else
		p = p->next;
	}
    }

    return nil;		// the metaclass is not loaded.
}

Class* opcxx_ListOfMetaclass::New(char* name, Ptree* def, Ptree* marg)
{
    if(name != nil){
	opcxx_ListOfMetaclass* p = head;
	while(p != nil){
	    if(strcmp(name, p->name) == 0)
		return (*p->proc)(def, marg);
	    else
		p = p->next;
	}
    }

    return nil;		// the metaclass is not loaded.
}

BOOL opcxx_ListOfMetaclass::AlreadyRecorded(char* name)
{
    for(opcxx_ListOfMetaclass* p = head; p != nil; p = p->next)
	if(strcmp(name, p->name) == 0)
	   return TRUE;

    return FALSE;
}

void opcxx_ListOfMetaclass::PrintAllMetaclasses()
{
    for(opcxx_ListOfMetaclass* p = head; p != nil; p = p->next)
	cout << p->name << '\n';
}
