/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/

#include "token.h"
#include "buffer.h"
#include "ptree.h"
#include "mop.h"

#ifndef DONT_GC
#include "gc/gc_cpp.h"
#endif

#define is_blank(c)	((c) == ' ' || (c) == '\t' || (c) == '\f')
#define is_digit(c)	('0' <= (c) && (c) <= '9')

void Token::WriteText(ostream& out, char* name)
{
    Class* c;
    Ptree* p;

    file->Write(out, name);
    for(c = Class::class_list; c != nil; c = c->next){
	p = c->Finalize();
	if(p != nil){
	    p->Write(out);
	    out << '\n';
	}
    }
}

void Token::ReplaceText(Ptree* oldtext, Ptree* newtext)
{
    file->Replace(PtreeLeftMost(oldtext), PtreeRightMost(oldtext), newtext);
}

void Token::InsertText(Ptree* oldtext, Ptree* before_text, Ptree* after_text)
{
    char* p;

    if(before_text != nil){
	p = PtreeLeftMost(oldtext);
	file->Replace(p, p, before_text);
    }

    if(after_text != nil){
	p = PtreeRightMost(oldtext);
	file->Replace(p, p, after_text);
    }
}

char* Token::PtreeLeftMost(Ptree* p)
{
    if(p == nil)
	return nil;
    else if(p->IsLeaf())
	return p->GetPosition();
    else{
	while(p != nil){
	    char* i = PtreeLeftMost(p->Car());
	    if(i != nil)
		return i;
	    else
		p = p->Cdr();
	}
	return nil;
    }
}

char* Token::PtreeRightMost(Ptree* p)
{
    if(p == nil)
	return nil;
    else if(p->IsLeaf())
	return p->GetPosition() + p->GetLength();
    else{
	int n = Ptree::Length(p);
	while(n > 0){
	    char* i = PtreeRightMost(Ptree::Nth(p, --n));
	    if(i != nil)
		return i;
	}

	return nil;
    }
}

/*
  MinimumReplace() replaces only changed nodes comparing oldtext and newtext.
  It always substitutes a subtree below a PtreeChange object.
*/
void Token::MinimumReplace(Ptree* oldtext, Ptree* newtext)
{
    if(MinimumReplace2(oldtext, newtext))
	ReplaceText(oldtext, newtext);
}

BOOL Token::MinimumReplace2(Ptree* oldtext, Ptree* newtext)
{
    if(oldtext == newtext)
	return FALSE;
    else if(oldtext == nil || newtext == nil)
	return TRUE;
    else if(newtext->IsA(PtreeChangedId))
	return TRUE;
    else if(oldtext->IsLeaf() || newtext->IsLeaf())
	return TRUE;
    else if(oldtext->Car() == nil && oldtext->Cdr() == nil)
	return TRUE;
    else if(oldtext == newtext->Cdr()){
	InsertText(oldtext, newtext->Car(), nil);
	return FALSE;
    }
    else if(oldtext->Car() == Ptree::Second(newtext)){
	InsertText(oldtext->Car(), newtext->Car(), nil);
	newtext = Ptree::ListTail(newtext, 2);
	if(MinimumReplace2(oldtext->Cdr(), newtext))
	    if(oldtext->Cdr() == nil)
		InsertText(oldtext->Car(), nil, newtext);
	    else
		ReplaceText(oldtext->Cdr(), newtext);
	else
	    return FALSE;
    }
    else{
	BOOL dirty1 = MinimumReplace2(oldtext->Car(), newtext->Car());
	BOOL dirty2 = MinimumReplace2(oldtext->Cdr(), newtext->Cdr());
	if(dirty1 == dirty2)
	    return dirty1;
	else if(dirty1)
	    if(oldtext->Cdr() == nil && newtext->Cdr() == nil)
		return TRUE;
	    else if(oldtext->Car() == nil)
		InsertText(oldtext->Cdr(), newtext->Car(), nil);
	    else
		ReplaceText(oldtext->Car(), newtext->Car());
	else
	    if(oldtext->Car() == nil && newtext->Car() == nil)
		return TRUE;
	    else if(oldtext->Cdr() == nil)
		InsertText(oldtext->Car(), nil, newtext->Cdr());
	    else
		ReplaceText(oldtext->Cdr(), newtext->Cdr());

	return FALSE;
    }
}

void Program::Replace(char* startpos, char* endpos, Ptree* text)
{
    if(startpos == nil || endpos == nil)
	return;

    uint start = uint(startpos - buf);
    uint end = uint(endpos - buf);
    Replacement* p = replacement;
#ifdef DONT_GC
    if(p == nil)
	replacement = new Replacement(nil, start, end, text);
    else if(p->next == nil)
	if(start <= p->startpos && end < p->endpos)
	    replacement = new Replacement(p, start, end, text);
	else
	    p->next = new Replacement(nil, start, end, text);
    else{
	for(; p->next != nil; p = p->next)
	    if(start <= p->next->startpos && end < p->next->endpos)
		break;

	p->next = new Replacement(p->next, start, end, text);
    }
#else
    if(p == nil)
	replacement = new (NoGC) Replacement(nil, start, end, text);
    else if(p->next == nil)
	if(start <= p->startpos && end < p->endpos)
	    replacement = new (NoGC) Replacement(p, start, end, text);
	else
	    p->next = new (NoGC) Replacement(nil, start, end, text);
    else{
	for(; p->next != nil; p = p->next)
	    if(start <= p->next->startpos && end < p->next->endpos)
		break;

	p->next = new (NoGC) Replacement(p->next, start, end, text);
    }
#endif
}

Program::Replacement::Replacement(Replacement* n, uint st, uint ed, Ptree* t)
{
    next = n;
    startpos = st;
    endpos = ed;
    text = t;
}

void Program::Write(ostream& out, char* file_name)
{
    Replacement* rep = replacement;
    uint pos;
    uint nlines = 1;
    uint line_number = 1;
    uint index = 0;
    char c;

    filename = 0;
    filename_length = 0;

    if(Ref(index) == '#')
	line_number = ReadLineDirective(index + 1, line_number);

    for(; rep != nil; rep = rep->next){
	pos = rep->startpos;
	while(index < pos){
	    c = Ref(index++);
	    if(c == '\0'){
		--index;
		break;
	    }

	    out << c;
	    if(c == '\n'){
		++nlines;
		++line_number;
		if(Ref(index) == '#')
		    line_number = ReadLineDirective(index + 1, line_number);
	    }
	}

	if(index > 0 && Ref(index - 1) != '\n'){
	    out << '\n';
	    ++nlines;
	}

	out << "# " << nlines + 1 << '\"' << file_name << "\"\n";
	nlines += rep->text->Write(out) + 3;

	pos = rep->endpos;
	while(index < pos){
	    c = Ref(index++);
	    if(c == '\0'){
		--index;
		break;
	    }
	    else if(c == '\n'){
		++line_number;
		if(Ref(index) == '#')
		    line_number = ReadLineDirective(index + 1, line_number);
	    }
	}

	out << "\n# " << line_number << ' ';
	if(filename_length > 0)
	    for(int i = 0; i < filename_length; ++i)
		out << (char)Ref(filename + i);
	else
	    out << "\"unknown\"";

	out << '\n';
    }

    while((c = Ref(index++)) != '\0'){
	out << c;
	if(c == '\n')
	    ++nlines;
    }

    out << "\n# " << nlines + 2 << " \"" << file_name << "\"\n";
}

int Program::ReadLineDirective(uint index, int line_number)
{
    char c;

    do{
	c = Ref(index++);
    }while(is_blank(c));

    if(is_digit(c)){		/* # <line> <file> */
	int num = c - '0';
	for(;;){
	    c = Ref(index++);
	    if(is_digit(c))
		num = num * 10 + c - '0';
	    else
		break;
	}

	line_number = num - 1;	/* line_number'll be incremented soon */

	if(is_blank(c)){
	    do{
		c = Ref(index++);
	    }while(is_blank(c));
	    if(c == '"'){
		uint fname_start = index - 1;
		do{
		    c = Ref(index++);
		} while(c != '"');
		if(index > fname_start + 2){
		    filename = fname_start;
		    filename_length = index - fname_start;
		}
	    }
	}
    }

    return line_number;
}
