/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/
/*
  OpenC++ parser

  This parser may depend on Bison.
  It causes 34 shift/reduce conflicts.
*/

%{
#include <iostream.h>
#include <stdlib.h>
#include <string.h>

#include "token.h"
#include "ptree.h"
#include "syntax.h"

/*
  YYSTYPE specifies the type of 'yylval', 'yyval', and the value stack.
*/
#define YYSTYPE		TokenVal

int yylex();
void yyerror(char*);

int    gSuppressErrorMsg;
static int showPtree;

%}

/*
  This parser assumes that yylex() distinguishes identifiers, typedef-ed
  type names, template names, qualified class names, and qualified
  constructor names.

  The token value for ClassName is a PtreeQClassName, PtreeTemplateName,
  or LeafClassName object.

  The token value for QConstructorName is a PtreeQClassName.
*/
%token Identifier	TypeName	TemplateName	ClassName
%token Constant		CharConst	StringL		QConstructorName
%token QName

%token AssignOp		EqualOp		RelOp		ShiftOp
%token LogOrOp		LogAndOp	IncOp		Scope
%token Ellipsis		PmOp		ArrowOp

%token '!' '%' '^' '&' '*' '(' ')' '-' '+' '=' '{' '}'
%token '|' '~' '[' ']' ';' ':' '<' '>' '?' ',' '.' '/'

/* invalid token */
%token BadToken

/*
  The following two tokens are for backtracking.  The function style cast
  and declaration cause ambiguity.  To disambiguate it, the parser must
  be able to backtrack.  This parser fisrt try to parse tokens asuming
  that they are interpreted as declaration.  If parsing fails, the parser
  makes error, which is captured by the error recovery rule.  The parser
  then asks the lexical analyzer to rewind the token stream, and parse
  the tokens again assuming that they are interpreted as expressions.
*/
%token ErrorOk		DeclCloseParen

/* C++ reserved */

%token AUTO		CHAR		CLASS		CONST
%token DELETE		DOUBLE		ENUM		EXTERN
%token FLOAT		FRIEND		INLINE		INT
%token LONG		NEW		OPERATOR	PRIVATE
%token PROTECTED	PUBLIC		REGISTER	SHORT
%token SIGNED		STATIC		STRUCT		TYPEDEF
%token UNION		UNSIGNED	VIRTUAL		VOID
%token VOLATILE		TEMPLATE

/* new C++ keywords */

%token MUTABLE

%token BREAK	CASE	CONTINUE	DEFAULT		DO
%token ELSE	FOR	GOTO		IF		RETURN
%token SIZEOF	SWITCH	THIS		WHILE

/* G++ reserved */
%token ATTRIBUTE

/* OpenC++ reserved */
%token METACLASS

/* OpenC++ token */
%token UserKeyword
%token UserKeyword2	/* object->keyword(arg.decl.list){ ... } */
%token UserKeyword3	/* object->keyword(... ; ... ; ... ) { ... } */
%token UserKeyword4	/* access control */

%%

/*
  In this part, YYDEBUG is always defined to be 0 or 1.
  do not use '#ifdef YYDEBUG' to express 'if this compilation is
  for debugging', but use '#if YYDEBUG' instead.
*/

/*
  Note: if a rule has no action, Bison copies the value of $1 into $$.
*/

program
    : nomore.backtrack
    | nomore.backtrack definitions
    ;

definitions
    : definition
	{ if(showPtree)
	      $1.ptree->Display();

	  Ptree::AppendToplevel($1.ptree);
        }
    | definitions definition
	{ if(showPtree)
	      $2.ptree->Display();

	  Ptree::AppendToplevel($2.ptree);
        }
    | error ';'
    | error '}'
    | definitions error ';'
    | definitions error '}'
    ;

definition
    : ';'
	{ $$.ptree = new PtreeDeclaration(NIL, NIL, $1); }
    | type.def
    | linkage.specification
    | declaration
    | template.declaration
    | meta.declaration
    ;

/* OpenC++ only */

meta.declaration
    : METACLASS Identifier ':' Identifier ';'
	{ $$.ptree = new PtreeMetaclass($1, new LeafClassName($2),
					$3, $4, NIL, $5); }
    | METACLASS Identifier ':' Identifier '(' meta.arguments ')' ';'
	{ $$.ptree = new PtreeMetaclass($1, new LeafClassName($2),
					$3, $4,
					new PtreeParen($5, $6, $7),
					$8); }
    | METACLASS ClassName ':' Identifier ';'
	{ $$.ptree = new PtreeMetaclass($1, $2.ptree, $3, $4, NIL, $5); }
    | METACLASS ClassName ':' Identifier '(' meta.arguments ')' ';'
	{ $$.ptree = new PtreeMetaclass($1, $2.ptree, $3, $4,
					new PtreeParen($5, $6, $7),
					$8); }
    ;

meta.arguments
    : meta.argument
	{ $$.ptree = Ptree::List($1.ptree); }
    | meta.arguments ',' meta.argument
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree)); }
    ;

meta.argument
    : const.argument
    | primary.type.name
    | '(' opt.comma.expression ')'
	{ $$.ptree = new PtreeParen($1, $2, $3); }
    ;

const.argument
    : Constant
	{ $$.ptree = new LeafConstValue($1); }
    | CharConst
	{ $$.ptree = new LeafConstValue($1); }
    | StringL
	{ $$.ptree = new LeafConstValue($1); }
    | Identifier
	{ $$.ptree = new LeafName($1); }
    ;

type.def
    : TYPEDEF type.specifier typedef.declarator.list ';'
	{ Ptree* p = new PtreeTypedef($1, $2.ptree, $3.ptree, $4);
	  /* $3 and $4 have been concatenated. */
	  for(Ptree* q = $3.ptree; q != nil; q = q->Cdr())
	      if(!q->Car()->Eq(',') && !q->Car()->Eq(';'))
	          Token::RecordTypedef(p, q->Car());

	  $$.ptree = p;
	}
    ;

typedef.declarator.list
    : typedef.declarator
	{ $$.ptree = Ptree::List($1.ptree); }
    | typedef.declarator.list ',' typedef.declarator
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree)); }
    ;

typedef.declarator
    : declarator
    | ClassName
    | TypeName
	{ $$.ptree = new LeafTypeName($1); }
    ;

linkage.specification
    : EXTERN StringL declaration 
	{ $$.ptree = new PtreeLinkage($1, $2, $3.ptree); }
    | EXTERN StringL '{' '}'
	{ $$.ptree = new PtreeLinkage($1, $2, new PtreeBrace($3, NIL, $4)); }
    | EXTERN StringL '{' linkage.body '}'
	{ $$.ptree = new PtreeLinkage($1, $2, new PtreeBrace($3, $4, $5)); }
    ;

linkage.body
    : definition
	{ $$.ptree = Ptree::List($1.ptree); }
    | linkage.body definition
	{ $$.ptree = Ptree::Snoc($1.ptree, $2.ptree); }
    ;

template.declaration
    : keyword.template '<' temp.arg.list '>' declaration
	{ Ptree* name = Syntax::DeclaredTemplateName($5.ptree);
	  $$.ptree = new PtreeTemplateDecl($1.ptree, $2, $3.ptree, $4,
					   $5.ptree);
	  if(name != nil)
	      Token::RecordTemplate(name, $$.ptree, Token::ExitClassBody());
	  else
	      Token::ExitScope();
        }
    ;

keyword.template
    : TEMPLATE
	{ Token::EnterNewScope();
	  $$.ptree = new LeafReserved($1); }
    ;

temp.arg.list
    : temp.argument
	{ $$.ptree = Ptree::List($1.ptree); }
    | temp.arg.list ',' temp.argument
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree)); }
    ;

temp.argument
    : type.specifier arg.declarator
	{ $$.ptree = Ptree::List($1.ptree, $2.ptree); }
    ;

declaration
    : declaration.statement
    | declarator ';'
	{ $$.ptree = Syntax::OldFuncDeclaration($1.ptree, new Leaf($2)); }
    | declarator
	{ $$.func_decl.env = Token::EnterFunctionScope($1.ptree); }
      func.body
	{ Token::ExitFunctionScope($2.func_decl.env);
	  $$.ptree = Syntax::OldFuncDeclaration($1.ptree, $3.ptree); }
    | declarator ':' expression ';'		/* bit field */
	{ Ptree* p = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree));
	  $$.ptree = new PtreeDeclaration(NIL, Ptree::List(p), $4); }
    | type.specifier declarator
	{ $$.func_decl.env = Token::EnterFunctionScope($2.ptree); }
      func.body
	{ Token::ExitFunctionScope($3.func_decl.env);
	  $$.ptree = Syntax::FuncDeclaration($1.ptree, $2.ptree, $4.ptree); }
    | qualified.constructor
    | constructor.or.bitfield
    ;

declaration.statement
    : type.specifier ';'
	{ $$.ptree = new PtreeDeclaration($1.ptree, NIL, $2); }
    | type.specifier declarator.list ';'
	{ $$.ptree = Syntax::Declaration($1.ptree, $2.ptree, new Leaf($3)); }
    ;

qualified.constructor
    : qualified.constructor.head '(' arg.decl.list ')' ';'
	{ Token::ExitNestedClassScope($1.func_decl.env);
	  $$.ptree = Syntax::QualifiedConstructor($1.func_decl.ptree,
						  $2, $3.ptree, $4,
						  new Leaf($5));
	}
    | qualified.constructor.head '(' arg.decl.list ')' func.body
	{ Token::ExitNestedClassScope($1.func_decl.env);
	  $$.ptree = Syntax::QualifiedConstructor($1.func_decl.ptree,
						  $2, $3.ptree, $4, $5.ptree);
	}
    | qualified.constructor.head '(' arg.decl.list ')'
      ':' member.initializers func.body
	{ Token::ExitNestedClassScope($1.func_decl.env);
	  $$.ptree = Syntax::QualifiedConstructor($1.func_decl.ptree,
						  $2, $3.ptree, $4,
						  new Leaf($5), $6.ptree,
						  $7.ptree);
	}
    ;

qualified.constructor.head
    : QConstructorName
	{ $$.func_decl.ptree = Syntax::QualifiedConstructorHead(NIL,
								$1.ptree);
	  $$.func_decl.env = Token::EnterNestedClassScope($1.ptree); }
    | type.specifier QConstructorName
	{ $$.func_decl.ptree = Syntax::QualifiedConstructorHead($1.ptree,
								$2.ptree);
	  $$.func_decl.env = Token::EnterNestedClassScope($2.ptree); }
    ;

/*
  This rules are for class members.  So we don't have to change the scope
  to parse the arguments and the function body for constructors.
*/
constructor.or.bitfield
    : type.specifier '(' arg.decl.list ')' ';'
	{ $$.ptree = Syntax::Constructor($1.ptree, $2, $3.ptree, $4,
					 new Leaf($5)); }
    | type.specifier '(' arg.decl.list ')' func.body
	{ $$.ptree = Syntax::Constructor($1.ptree, $2, $3.ptree, $4,
					 $5.ptree); }
    | type.specifier '(' arg.decl.list ')' ':' member.initializers func.body
	{ $$.ptree = Syntax::Constructor($1.ptree, $2, $3.ptree, $4,
					 new Leaf($5), $6.ptree, $7.ptree); }
    | type.specifier ':' expression ';'
	{ Ptree* p = Ptree::List(NIL, new Leaf($2), $3.ptree);
	  $$.ptree = new PtreeDeclaration($1.ptree, Ptree::List(p), $4); }
    ;

declarator.list
    : init.declarator.or.bitfield
	{ $$.ptree = Ptree::List($1.ptree); }
    | declarator.list ',' init.declarator.or.bitfield
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree)); }
    ;

init.declarator.or.bitfield
    : declarator
    | declarator '=' initialize.expr
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree)); }
    | declarator ':' expression
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree)); }
    ;

/*
  The action rule for functions assumes that a function that returns
  a function is invalid.
*/
declarator
    : qualified.name
	{ $$.ptree = Ptree::List($1.ptree); }
    | ptr.operator declarator
	{ $$.ptree = Ptree::Nconc($1.ptree, $2.ptree); }
    | func.declarator arg.decl.list.or.initializer ')'
	{ Token::ExitFunctionScope($1.func_decl.env);
	  $$.ptree = Ptree::Nconc($1.func_decl.ptree,
				  Ptree::List($2.ptree, new Leaf($3))); }
    | func.declarator arg.decl.list.or.initializer ')' cv.qualifiers
	{ Token::ExitFunctionScope($1.func_decl.env);
	  $$.ptree = Ptree::Nconc($1.func_decl.ptree,
				  Ptree::List($2.ptree, new Leaf($3)),
				  $4.ptree); }
    | declarator '[' opt.comma.expression ']'
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree,
					      new Leaf($4))); }
    | '(' declarator ')'
	{ $$.ptree = Ptree::List(new PtreeParen($1, $2, $3)); }
    ;

func.declarator
    : declarator '('
	{ $$.func_decl.env = Token::EnterFunctionScope($1.ptree);
	  $$.func_decl.ptree = Ptree::Snoc($1.ptree, new Leaf($2)); }
    ;

/*
  This rule must accept function.arguments to parse declarations like:
	Point p(1, 3);
  "(1, 3)" is arg.decl.list.or.initializer.

  Although declarator may cause backtracking, it never causes backtracking
  if the state shifts into this rule.  Thus the machanism of nested
  backtracking is not necessary.  See expression.or.declaration and
  backtrackable.statement.
*/
arg.decl.list.or.initializer
    : prepare.backtrack arg.decl.list nomore.backtrack
	{ $$.ptree = $2.ptree; }
    | prepare.backtrack error do.backtrack ErrorOk
      function.arguments
	{ $$.ptree = $5.ptree; }
    ;

ptr.operators
    : ptr.operator
    | ptr.operators ptr.operator
	{ $$.ptree = Ptree::Nconc($1.ptree, $2.ptree); }
    ;

ptr.operator
    : ptr.op.element
	{ $$.ptree = Ptree::List($1.ptree); }
    | ptr.op.element cv.qualifiers
	{ $$.ptree = Ptree::List($1.ptree, $2.ptree); }
    ;

ptr.op.element
    : '*'
	{ $$.ptree = new LeafPointer($1); }
    | '&'
	{ $$.ptree = new LeafReference($1); }
    | ClassName Scope '*'
	{ $$.ptree = new PtreePtrToMember($1, $2, $3); }
    ;

cv.qualifiers
    : cv.qualifier
	{ $$.ptree = new PtreeCvQualifier($1); }
    | cv.qualifiers cv.qualifier
	{ $$.ptree = Ptree::Snoc($1.ptree, new PtreeCvQualifier($2)); }
    ;

cv.qualifier
    : CONST
    | VOLATILE
    | MUTABLE
    ;

qualified.name
    : symbol.name
    | Scope symbol.name
	{ $$.ptree = new PtreeQname($1, $2); }
    | QName symbol.name
	{ Ptree::Snoc($1.ptree, $2.ptree);
	  $$.ptree = $1.ptree; }
    ;

symbol.name
    : Identifier
	{ $$.ptree = new LeafName($1); }
    | '~' ClassName
	{ Syntax::Destructor($2.ptree);
	  $$.ptree = new PtreeDestructor($1, $2); }
    | op.func.name
    ;

/* operator functions */

op.func.name
    : OPERATOR operator
	{ $$.ptree = new PtreeOperatorFunc($1, $2); }
    ;

operator
    : op.symbol
	{ $$.ptree = new Leaf($1); }
    | new.or.delete
	{ $$.ptree = new LeafReserved($1); }
    | new.or.delete '[' ']'
	{ /* this rule makes a shift/reduce conflict.  But shifting
	     is fine. */
	  $$.ptree = Ptree::List(new LeafReserved($1), new Leaf($2),
				 new Leaf($3)); }
    | '(' ')'
	{ $$.ptree = Ptree::List(new Leaf($1), new Leaf($2)); }
    | '[' ']'
	{ $$.ptree = Ptree::List(new Leaf($1), new Leaf($2)); }
    | type.specifier
	{ $$.ptree = Syntax::CastOperator($1.ptree); }
    | type.specifier ptr.operators
	{ $$.ptree = Ptree::Nconc($1.ptree, $2.ptree); }
    ;

new.or.delete
    : NEW
    | DELETE
    ;

op.symbol
    : '+' | '-' | '*' | '/' | '%' | '^' | '&' | '|' | '~'
    | '!' | '=' | '<' | '>' | AssignOp | ShiftOp | EqualOp
    | RelOp | LogAndOp | LogOrOp | IncOp | ',' | PmOp | ArrowOp
    ;

/* arguments */

arg.decl.list
    : /* empty */
	{ $$.ptree = nil; }
    | Ellipsis
	{ $$.ptree = Ptree::List(new Leaf($1)); }
    | arg.declaration.list
    | arg.declaration.list Ellipsis
	{ $$.ptree = Ptree::Snoc($1.ptree, new Leaf($2)); }
    | arg.declaration.list ',' Ellipsis
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), new Leaf($3))); }
    ;

arg.declaration.list
    : arg.declaration
	{ $$.ptree = Ptree::List($1.ptree); }
    | arg.declaration.list ',' arg.declaration
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree)); }
    ;

arg.declaration
    : type.specifier arg.declarator
	{ $$.ptree = Ptree::List($1.ptree, $2.ptree); }
    | type.specifier arg.declarator '=' expression
	{ $$.ptree = Ptree::List($1.ptree, $2.ptree, new Leaf($3),
				 $4.ptree); }
    ;

arg.declarator
    : /* empty */
	{ $$.ptree = nil; }
    | qualified.name
	{ $$.ptree = Ptree::List($1.ptree); }
    | ptr.operator arg.declarator
	{ $$.ptree = Ptree::Nconc($1.ptree, $2.ptree); }
    | arg.declarator '(' arg.decl.list ')'
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree,
					      new Leaf($4))); }
    | arg.declarator '(' arg.decl.list ')' cv.qualifiers
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree,
					      new Leaf($4), $5.ptree)); }
    | arg.declarator '[' opt.comma.expression ']'
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree,
					      new Leaf($4))); }
    | '(' arg.declarator ')'
	{ $$.ptree = Ptree::List(new PtreeParen($1, $2, $3)); }
    ;

/* type sepcification */

/*
  Note: type.specifier matches only "const" if the given tokens are
  "const A = 1;"
*/
type.specifier
    : primary.type.name
    | type.modifiers
    | type.modifiers primary.type.name
	{ $$.ptree = Ptree::Nconc($1.ptree, $2.ptree); }
    | type.modifiers primary.type.name cv.qualifiers
	{ $$.ptree = Ptree::Nconc($1.ptree, $2.ptree, $3.ptree); }
    | primary.type.name cv.qualifiers
	{ $$.ptree = Ptree::Nconc($1.ptree, $2.ptree); }
    ;

primary.type.name
    : builtin.type.names
    | TypeName
	{ $$.ptree = Ptree::List(new LeafTypeName($1)); }
    | ClassName
	{ $$.ptree = Ptree::List($1.ptree); }
    | class.specifier
	{ $$.ptree = Ptree::List($1.ptree); }
    | enum.specifier
	{ $$.ptree = Ptree::List($1.ptree); }
    ;

type.modifiers
    : type.modifier
	{ $$.ptree = Ptree::List($1.ptree); }
    | type.modifiers type.modifier
	{ $$.ptree = Ptree::Snoc($1.ptree, $2.ptree); }
    ;

type.modifier
    : type.modifier2
	{ $$.ptree = new LeafReservedModifier($1); }
    | userdef.keyword
    ;

type.modifier2
    : cv.qualifier
    | STATIC
    | EXTERN
    | AUTO
    | REGISTER
    | INLINE
    | VIRTUAL
    | FRIEND
    ;

builtin.type.names
    : builtin.type.name
	{ $$.ptree = Ptree::List(new LeafReserved($1)); }
    | builtin.type.names builtin.type.name
	{ $$.ptree = Ptree::Snoc($1.ptree, new LeafReserved($2)); }
    ;

builtin.type.name
    : CHAR
    | INT
    | SHORT
    | LONG
    | SIGNED
    | UNSIGNED
    | FLOAT
    | DOUBLE
    | VOID
    ;

/* user-defined keywords (OpenC++ only) */

userdef.keyword
    : UserKeyword
	{ $$.ptree = new PtreeUserKeyword($1, NIL); }
    | UserKeyword '(' function.arguments ')'
	{ $$.ptree = new PtreeUserKeyword($1, new PtreeParen($2, $3, $4)); }
    ;

/* enum declaration */

enum.specifier
    : ENUM Identifier
	{ $$.ptree = new PtreeEnum($1, $2);
	  Token::RecordEnum($2, $$.ptree); }
    | ENUM TypeName
	{ $$.ptree = new PtreeEnum($1, $2); }
    | ENUM enum.body
	{ $$.ptree = new PtreeEnum($1, $2.ptree); }
    | ENUM Identifier enum.body
	{ $$.ptree = new PtreeEnum($1, $2, $3.ptree);
	  Token::RecordEnum($2, $$.ptree); }
    ;

enum.body
    : '{' enum.list '}'
	{ $$.ptree = new PtreeBrace($1, $2, $3); }
    | '{' enum.list ',' '}'
	{ $$.ptree = new PtreeBrace($1, Ptree::Snoc($2.ptree, new Leaf($3)),
				    $4); }
    ;

enum.list
    : enumerator
	{ $$.ptree = Ptree::List($1.ptree); }
    | enum.list ',' enumerator
	{ $$.ptree = Ptree::Nconc($1.ptree,
				 Ptree::List(new Leaf($2), $3.ptree)); }
    ;

enumerator
    : Identifier
	{ $$.ptree = new Leaf($1); }
    | Identifier '=' expression
	{ $$.ptree = Ptree::List(new Leaf($1), new Leaf($2), $3.ptree); }
    ;

/* --- class declaration --- */

class.specifier
    : class.specifier.with.name
    | class.specifier.with.name class.body
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(nil, $2.class_body.ptree));
	  Token::RecordClass($1.ptree->Cadr(), $$.ptree, $2.class_body.env); }
    | class.specifier.with.name ':' base.specifier class.body
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(Ptree::Cons(new Leaf($2),
							  $3.ptree),
					      $4.class_body.ptree));
	  Token::RecordClass($1.ptree->Cadr(), $$.ptree, $4.class_body.env); }
    | class.key class.body
	{ $$.ptree = new PtreeClass($1, nil);
	  Ptree::Nconc($$.ptree, Ptree::List(nil, $2.class_body.ptree)); }
    ;

/*
  RecordClass() must be called here.  Otherwise, the class body cannot
  be correctly parsed.
*/
class.specifier.with.name
    : class.key Identifier
	{ Ptree* p = new LeafClassName($2);
	  Token::RecordClass(p, nil, nil);
	  $$.ptree = new PtreeClass($1, p); }
    | class.key ClassName
	{ Syntax::IsNotQualifiedClassName($2.ptree);
	  $$.ptree = new PtreeClass($1, $2.ptree); }
    ;

class.key
    : CLASS
    | STRUCT
    | UNION
    ;

class.body
    : begin.class.body end.class.body
	{ $$.class_body.env = $2.class_body.env;
	  $$.class_body.ptree = new PtreeBrace($1, NIL,
					       $2.class_body.ptree); }
    | begin.class.body class.mems end.class.body
	{ $$.class_body.env = $3.class_body.env;
	  $$.class_body.ptree = new PtreeBrace($1, $2.ptree,
					       $3.class_body.ptree); }
    ;

begin.class.body
    : '{'
	{ Token::EnterNewScope();
	  $$ = $1; }
    ;

end.class.body
    : '}'
	{ $$.class_body.env = Token::ExitClassBody();
	  $$.class_body.ptree = new Leaf($1); }
    ;

class.mems
    : class.member
	{ $$.ptree = Ptree::List($1.ptree); }
    | class.mems class.member
	{ $$.ptree = Ptree::Snoc($1.ptree, $2.ptree); }
    ;

class.member
    : access.spec ':'
	{ $$.ptree = new PtreeAccessCtrl($1, $2); }
    | userdef.access.spec
    | declaration
    | type.def
    | meta.declaration		/* OpenC++ only */
    | ';'
	{ $$.ptree = new PtreeDeclaration(NIL, NIL, $1); }
    ;

userdef.access.spec		/* OpenC++ only */
    : UserKeyword4 ':'
	{ $$.ptree = new PtreeUserKeyword($1, NIL, $2); }
    | UserKeyword4 '(' function.arguments ')' ':'
	{ $$.ptree = new PtreeUserKeyword($1, new PtreeParen($2, $3, $4),
					  $5); }
    ;

base.specifier
    : base.class.specifier
	{ $$.ptree = Ptree::List($1.ptree); }
    | base.specifier ',' base.class.specifier
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree)); }
    ;

base.class.specifier
    : ClassName
	{ $$.ptree = Ptree::List($1.ptree); }
    | VIRTUAL ClassName
	{ $$.ptree = Ptree::List(new LeafReserved($1), $2.ptree); }
    | access.spec ClassName
	{ $$.ptree = Ptree::List(new LeafReserved($1), $2.ptree); }
    | VIRTUAL access.spec ClassName
	{ $$.ptree = Ptree::List(new LeafReserved($1), new LeafReserved($2),
				 $3.ptree); }
    | access.spec VIRTUAL ClassName
	{ $$.ptree = Ptree::List(new LeafReserved($1), new LeafReserved($2),
				 $3.ptree); }

    ;

access.spec
    : PRIVATE
    | PROTECTED
    | PUBLIC
    ;

member.initializers
    : member.initializer
	{ $$.ptree = Ptree::List($1.ptree); }
    | member.initializers ',' member.initializer
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree)); }
    ;

member.initializer
    : qualified.name '(' function.arguments ')'
	{ $$.ptree = Ptree::List($1.ptree, new Leaf($2), $3.ptree,
				 new Leaf($4)); }
    | ClassName '(' function.arguments ')'
	{ $$.ptree = Ptree::List($1.ptree, new Leaf($2), $3.ptree,
				 new Leaf($4)); }
    ;

/* --- expressions --- */

initialize.expr
    : expression
    | '{' initializer.list '}'
	{ $$.ptree = new PtreeBrace($1, $2, $3); }
    | '{' initializer.list ',' '}'
	{ $$.ptree = new PtreeBrace($1, Ptree::Snoc($2.ptree, new Leaf($3)),
				    $4); }
    ;

initializer.list
    : initialize.expr
	{ $$.ptree = Ptree::List($1.ptree); }
    | initializer.list ',' initialize.expr
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree)); }
    ;

function.arguments
    : /* empty */
	{ $$.ptree = nil; }
    | function.args.list
    ;

function.args.list
    : expression
	{ $$.ptree = new PtreeActualArgs($1.ptree); }
    | function.args.list ',' expression
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree)); }
    ;

opt.comma.expression
    : /* empty */
	{ $$.ptree = nil; }
    | comma.expression
    ;

comma.expression
    : expression
    | comma.expression ',' expression
	{ $$.ptree = new PtreeInfixExpr($1, $2, $3); }
    ;

/* -- must be an optimal list. */
expression
    : conditional.expr
    | conditional.expr AssignOp expression
	{ $$.ptree = Syntax::MakeAssignExpr($1.ptree, $2, $3.ptree); }
    | conditional.expr '=' expression
	{ $$.ptree = Syntax::MakeAssignExpr($1.ptree, $2, $3.ptree); }
    | error
	{ /* yyparse() does not print an error message.  So we explicitly
	     insert ErrorMessage() here.  This rule makes a dummy StringL. */
	  ErrorMessage("parse error");
	  $$.ptree = new UserLeaf("error", 5); }
    ;

conditional.expr
    : logical.or.expr
    | logical.or.expr '?' comma.expression ':' conditional.expr
	{ $$.ptree = new PtreeCondExpr($1, $2, $3, $4, $5); }
    ;

logical.or.expr
    : logical.and.expr
    | logical.or.expr LogOrOp logical.and.expr
	{ $$.ptree = new PtreeInfixExpr($1, $2, $3); }
    ;

logical.and.expr
    : inclusive.or.expr
    | logical.and.expr LogAndOp inclusive.or.expr
	{ $$.ptree = new PtreeInfixExpr($1, $2, $3); }
    ;

inclusive.or.expr
    : exclusive.or.expr
    | inclusive.or.expr '|' exclusive.or.expr
	{ $$.ptree = new PtreeInfixExpr($1, $2, $3); }
    ;

exclusive.or.expr
    : and.expr
    | exclusive.or.expr '^' and.expr
	{ $$.ptree = new PtreeInfixExpr($1, $2, $3); }
    ;

and.expr
    : equality.expr
    | and.expr '&' equality.expr
	{ $$.ptree = new PtreeInfixExpr($1, $2, $3); }
    ;

equality.expr
    : relational.expr
    | equality.expr EqualOp relational.expr
	{ $$.ptree = new PtreeInfixExpr($1, $2, $3); }
    ;

relational.expr
    : shift.expr
    | relational.expr RelOp shift.expr
	{ $$.ptree = new PtreeInfixExpr($1, $2, $3); }
    | relational.expr '<' shift.expr
	{ $$.ptree = new PtreeInfixExpr($1, $2, $3); }
    | relational.expr '>' shift.expr
	{ $$.ptree = new PtreeInfixExpr($1, $2, $3); }
    ;

shift.expr
    : additive.expr
    | shift.expr ShiftOp additive.expr
	{ $$.ptree = new PtreeInfixExpr($1, $2, $3); }
    ;

additive.expr
    : multiply.expr
    | additive.expr '+' multiply.expr
	{ $$.ptree = new PtreeInfixExpr($1, $2, $3); }
    | additive.expr '-' multiply.expr
	{ $$.ptree = new PtreeInfixExpr($1, $2, $3); }
    ;

multiply.expr
    : pm.expr
    | multiply.expr '*' pm.expr
	{ $$.ptree = new PtreeInfixExpr($1, $2, $3); }
    | multiply.expr '/' pm.expr
	{ $$.ptree = new PtreeInfixExpr($1, $2, $3); }
    | multiply.expr '%' pm.expr
	{ $$.ptree = new PtreeInfixExpr($1, $2, $3); }
    ;

pm.expr
    : cast.expr
    | pm.expr PmOp cast.expr
	{ $$.ptree = new PtreePtrToMemExpr($1.ptree, $2, $3.ptree); }
    ;

cast.expr
    : unary.expr
    | '(' expression.or.declaration DeclCloseParen cast.expr
	{ $$.ptree = new PtreeCast($1, $2, $3, $4); }
    ;

unary.expr
    : postfix.expr
    | unary.operator cast.expr
	{ $$.ptree = new PtreeUnaryExpr($1, $2); }
    | SIZEOF unary.expr
	{ $$.ptree = new PtreeSizeofExpr(new LeafReserved($1), $2.ptree); }
    | SIZEOF '(' expression.or.declaration DeclCloseParen
	{ $$.ptree = new PtreeSizeofType(new LeafReserved($1),
					  new PtreeParen($2, $3, $4)); }
    | allocate.expr
    | deallocate.expr
    ;

unary.operator
    : '*' | '&' | '+' | '-' | '!' | '~' | IncOp
    ;

allocate.expr
    : NEW allocate2.expr allocate.initializer
	{ $$.ptree = new PtreeNew(nil, new LeafReserved($1),
				  $2.ptree, $3.ptree); }
    | userdef.keyword NEW allocate2.expr allocate.initializer   /* OpenC++ */
	{ $$.ptree = new PtreeNew($1.ptree, new LeafReserved($2),
				  $3.ptree, $4.ptree); }
    | Scope NEW allocate2.expr allocate.initializer
	{ $$.ptree = new PtreeNew(new Leaf($1), new LeafReserved($2),
				  $3.ptree, $4.ptree); }
    ;

allocate2.expr
    : type.specifier new.declarator
	{ $$.ptree = Ptree::List(NIL, Ptree::Cons($1.ptree, $2.ptree)); }
    | '(' expression.or.declaration DeclCloseParen
	{ $$.ptree = Ptree::List(NIL, new PtreeParen($1, $2.ptree, $3)); }
    | '(' expression.or.declaration ')' type.specifier new.declarator
	{ $$.ptree = Ptree::List(new PtreeParen($1, $2, $3),
				 Ptree::Cons($4.ptree, $5.ptree)); }
    | '(' expression.or.declaration ')' '(' type.specifier cast.declarator ')'
	{ Ptree* p = Ptree::Cons($5.ptree, $6.ptree);
	  $$.ptree = Ptree::List(new PtreeParen($1, $2, $3),
				 new PtreeParen($4, p, $7)); }
    ;

new.declarator
    : /* empty */
	{ $$.ptree = nil; }
    | ptr.operators
    | new.declarator.array
    | ptr.operators new.declarator.array
	{ $$.ptree = Ptree::Nconc($1.ptree, $2.ptree); }
    ;

new.declarator.array
    : '[' comma.expression ']'
	{ $$.ptree = Ptree::List(new Leaf($1), $2.ptree, new Leaf($3)); }
    | new.declarator.array '[' comma.expression ']'
	{ $$.ptree = Ptree::Nconc($1.ptree,
			          Ptree::List(new Leaf($2), $3.ptree,
					      new Leaf($4))); }
    ;

allocate.initializer
    : /* empty */
	{ $$.ptree = nil; }
    | '(' ')'
	{ $$.ptree = new PtreeParen($1, NIL, $2); }
    | '(' initializer.list ')'
	{ $$.ptree = new PtreeParen($1, $2, $3); }
    ;

deallocate.expr
    : DELETE cast.expr
	{ $$.ptree = new PtreeDelete(nil, new LeafReserved($1), nil,
				     $2.ptree); }
    | DELETE '[' ']' cast.expr
	{ $$.ptree = new PtreeDelete(nil, new LeafReserved($1),
				     Ptree::List(new Leaf($2), new Leaf($3)),
				     $4.ptree); }
    | Scope DELETE cast.expr
	{ $$.ptree = new PtreeDelete(new Leaf($1), new LeafReserved($2),
				     nil, $3.ptree); }
    | Scope DELETE '[' ']' cast.expr
	{ $$.ptree = new PtreeDelete(new Leaf($1), new Leaf($2),
				     Ptree::List(new Leaf($3), new Leaf($4)),
				     $5.ptree); }
    ;

postfix.expr
    : Constant
	{ $$.ptree = new LeafConstValue($1); }
    | CharConst
	{ $$.ptree = new LeafConstValue($1); }
    | StringL
	{ $$.ptree = new LeafConstValue($1); }
    | THIS
	{ $$.ptree = new LeafThis($1); }
    | qualified.name
    | '(' expression.or.declaration ')'
	{ $$.ptree = new PtreeParen($1, $2, $3); }
    | postfix.expr '[' comma.expression ']'
	{ $$.ptree = new PtreePostfixExpr($1.ptree,
					  new PtreeBracket($2, $3, $4)); }
    | postfix.expr '(' function.arguments ')'
	{ $$.ptree = Syntax::MakeFuncallExpr($1.ptree, $2, $3.ptree, $4); }
    | postfix.expr '.' qualified.name
	{ $$.ptree = new PtreeMemberAccessExpr($1.ptree, $2, $3.ptree); }
    | postfix.expr ArrowOp qualified.name
	{ $$.ptree = new PtreeMemberAccessExpr($1.ptree, $2, $3.ptree); }
    | postfix.expr IncOp
	{ $$.ptree = new PtreePostfixExpr($1.ptree, new Leaf($2)); }
    /*
      The next two rules are for fucntion-style casts and the C-style
      initialize expressions.
      Accroding to ARM, the rule should be "primary.type.name '(' ... ')'".
    */
    | TypeName '(' function.arguments ')'
	{ $$.ptree = new PtreeFstyleCast(Ptree::List(new LeafTypeName($1)),
					 $2, $3, $4); }
    | ClassName '(' function.arguments ')'
	{ $$.ptree = new PtreeFstyleCast(Ptree::List($1.ptree), $2, $3, $4); }
    | builtin.type.names '(' function.arguments ')'
	{ $$.ptree = new PtreeFstyleCast($1.ptree, $2, $3, $4); }
    /*
      OpenC++ only
    */
    | postfix.expr '.' userdef.statement
	{ $$.ptree = new PtreeUserStatement($1, $2, $3); }
    | postfix.expr ArrowOp userdef.statement
	{ $$.ptree = new PtreeUserStatement($1, $2, $3); }
    | QName userdef.statement
	{ $$.ptree = new PtreeStaticUserStatement($1, $2); }
    ;

cast.declarator
    : /* empty */
	{ $$.ptree = nil; }
    | ptr.operator cast.declarator
	{ $$.ptree = Ptree::Nconc($1.ptree, $2.ptree); }
    | cast.declarator '(' arg.decl.list ')'
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree,
					      new Leaf($4))); }
    | cast.declarator '(' arg.decl.list ')' cv.qualifiers
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree,
					      new Leaf($4), $5.ptree)); }
    | cast.declarator '[' opt.comma.expression ']'
	{ $$.ptree = Ptree::Nconc($1.ptree,
				  Ptree::List(new Leaf($2), $3.ptree,
					      new Leaf($4))); }
    | '(' cast.declarator ')'
	{ $$.ptree = Ptree::List(new PtreeParen($1, $2, $3)); }
    ;

/* OpenC++ only */

userdef.statement
    : UserKeyword '(' function.arguments ')' compound.statement
	{ $$.ptree = Ptree::List(new Leaf($1), new Leaf($2), $3.ptree,
				 new Leaf($4), $5.ptree); }
    | UserKeyword2 '(' arg.decl.list ')' compound.statement
	{ $$.ptree = Ptree::List(new Leaf($1), new Leaf($2), $3.ptree,
				 new Leaf($4), $5.ptree); }
    | UserKeyword3 '(' backtrackable.expr.statement for.expr ';' for.expr ')'
      compound.statement
	{ Ptree* p = Ptree::List($3.ptree, $4.ptree, new Leaf($5), $6.ptree);
	  $$.ptree = Ptree::List(new Leaf($1), new Leaf($2), p, new Leaf($7),
				 $8.ptree); }
    ;

/*
  This rule may make backtracking.  If the result of parsing is declaration,
  then the next token is DeclCloseParen (declaration closing parenthesis).
  Otherwise, the next token is ')'.

  The way of backtracking is simple.  First discard a look ahead token,
  tell the lexical analyzer to rewind the token stream and to return
  ErrorOk when yylex() is called next.  Because the parser skips tokens
  until it encounters ErrorOk after an error occurs, the internal state
  is guaranteed to be recovered.

  Since errors always involves invocations of yyerror(), the parser must
  suppress error messages.

  One big defect of the backrack is that we don't have a way to deallocate
  memory after we detect an error.

  yyclean is Bison's macro.  It flushes a look-ahead token.
*/
expression.or.declaration
    : prepare.backtrack type.specifier cast.declarator
	{ if(yychar == ')'){
	    yyclearin;
	    Token::PushToken(DeclCloseParen);
	    /* Caution: TokenPosition() must return the position of
			the last ')'. */
	  }
	  gSuppressErrorMsg = FALSE;
	  /* ordinary action rule. */
	  $$.ptree = Ptree::Cons($2.ptree, $3.ptree); }
    | prepare.backtrack error do.backtrack ErrorOk comma.expression
	{ $$.ptree = $5.ptree; }
    ;

/* these three non-terminals are for backtracking.  Their actions
   are not relevant to constructing a parse tree. */

prepare.backtrack
    : /* empty */
	{ Token::MarkPosition(yychar, yychar == YYEMPTY);
	  gSuppressErrorMsg = TRUE; }
    ;

do.backtrack
    : /* empty */
	{ yyclearin;
	  Token::BackTrack(ErrorOk);
	  gSuppressErrorMsg = FALSE;
        }
    ;

nomore.backtrack
    : /* empty */
	{ gSuppressErrorMsg = FALSE; }
    ;


/* --- function body --- */

func.body
    : compound.statement
    ;

compound.statement
    : empty.body
    | begin.body statements end.body
	{ $$.ptree = new PtreeBrace($1, $2, $3); }
    ;

begin.body
    : '{'
	{ Token::EnterNewScope();
	  $$ = $1; }
    ;

end.body
    : '}'
	{ Token::ExitScope();
	  $$ = $1; }
    ;

empty.body
    : begin.body prepare.backtrack end.body nomore.backtrack
	{ $$.ptree = new PtreeBrace($1, NIL, $3); }
    ;

statements
    : backtrackable.statement
	{ $$.ptree = Ptree::List($1.ptree); }
    | statements backtrackable.statement
	{ $$.ptree = Ptree::Snoc($1.ptree, $2.ptree); }
    ;

/*
  statement may cause backtracking as well as expression.or.declaration.
  The implementation technique is almost the same as theirs.
*/
backtrackable.statement
    : prepare.backtrack statement nomore.backtrack
	{ $$.ptree = $2.ptree; }
    ;

statement
    : expr.statement
    | compound.statement
    | labeled.statement
    | selection.statement
    | loop.statement
    | jump.statement
    ;

expr.statement
    : ';'
	{ $$.ptree = new PtreeDeclaration(NIL, NIL, $1); }
    | declaration.statement
    | error do.backtrack ErrorOk comma.expression ';'
	{ $$.ptree = new PtreeExprStatement($4, $5); }
    ;

/*
  A label "foo :" doesn't make a statement alone.
  It must have a following statement, which may be a labeled statement.
  So "foo: bar: a = 1;" makes a single (nested) statement.
*/
labeled.statement
    : Identifier ':' backtrackable.statement
	{ $$.ptree = new PtreeLabel(new LeafName($1), new Leaf($2),
				    $3.ptree); }
    | CASE exprs.in.statement ':' backtrackable.statement
	{ $$.ptree = new PtreeCaseLabel(new LeafReserved($1), $2.ptree,
					new Leaf($3), $4.ptree); }
    | DEFAULT ':' backtrackable.statement
	{ $$.ptree = new PtreeLabel(new LeafReserved($1), new Leaf($2),
				    $3.ptree); }
    ;

selection.statement
    : IF '(' exprs.in.statement ')' backtrackable.statement
	{ $$.ptree = Ptree::Nconc(new PtreeIfStatement($1),
				  Ptree::List(new Leaf($2), $3.ptree,
					      new Leaf($4), $5.ptree)); }
    | IF '(' exprs.in.statement ')' backtrackable.statement
      ELSE backtrackable.statement
	{ $$.ptree = Ptree::Nconc(new PtreeIfStatement($1),
				  Ptree::List(new Leaf($2), $3.ptree,
					      new Leaf($4),
					      $5.ptree,
					      new LeafReserved($6),
					      $7.ptree)); }
    | SWITCH '(' exprs.in.statement ')' backtrackable.statement
	{ $$.ptree = Ptree::Nconc(new PtreeSwitchStatement($1),
				  Ptree::List(new Leaf($2), $3.ptree,
					      new Leaf($4), $5.ptree)); }
    ;

loop.statement
    : WHILE '(' exprs.in.statement ')' backtrackable.statement
	{ $$.ptree = Ptree::Nconc(new PtreeWhileStatement($1),
				  Ptree::List(new Leaf($2), $3.ptree,
					      new Leaf($4), $5.ptree)); }
    | DO backtrackable.statement WHILE '(' exprs.in.statement ')' ';'
	{ $$.ptree = Ptree::Nconc(new PtreeDoStatement($1),
				  Ptree::List($2.ptree,
					      new LeafReserved($3),
					      new Leaf($4), $5.ptree,
					      new Leaf($6), new Leaf($7))); }
    | FOR '(' backtrackable.expr.statement for.expr ';' for.expr ')'
      backtrackable.statement
	{ $$.ptree = Ptree::Nconc(new PtreeForStatement($1),
				  Ptree::List(new Leaf($2), $3.ptree,
					      $4.ptree, new Leaf($5),
					      $6.ptree, new Leaf($7),
					      $8.ptree)); }
    ;

for.expr
    : /* empty */
	{ $$.ptree = nil; }
    | exprs.in.statement
    ;

jump.statement
    : BREAK ';'
	{ $$.ptree = Ptree::Nconc(new PtreeBreakStatement($1),
				  Ptree::List(new Leaf($2))); }
    | CONTINUE ';'
	{ $$.ptree = Ptree::Nconc(new PtreeContinueStatement($1),
				  Ptree::List(new Leaf($2))); }
    | RETURN ';'
	{ $$.ptree = Ptree::Nconc(new PtreeReturnStatement($1),
				  Ptree::List(NIL, new Leaf($2))); }
    | RETURN exprs.in.statement ';'
	{ $$.ptree = Ptree::Nconc(new PtreeReturnStatement($1),
				  Ptree::List($2.ptree, new Leaf($3))); }
    | GOTO Identifier ';'
	{ $$.ptree = Ptree::Nconc(new PtreeGotoStatement($1),
				  Ptree::List(new LeafName($2),
					      new Leaf($3))); }
    ;

backtrackable.expr.statement
    : prepare.backtrack expr.statement nomore.backtrack
	{ $$.ptree = $2.ptree; }
    ;

exprs.in.statement
    : nomore.backtrack comma.expression
	{ $$.ptree = $2.ptree; }
    ;

%%

static void check_num_errors()
{
    const MaxErrors = 8;

    if(yynerrs > MaxErrors){
	cerr << "OpenC++: *** Too many errors. Generation stops.\n";
	exit(yynerrs);
    }
}

static ostream& show_error_msg_head()
{
    cerr << Token::File() << ", line " << Token::Line() << ": ";
    return cerr;
}

#ifdef YYDEBUG
static int nbacktracks = 0;
#endif

void yyerror(char* msg)
{
    if(!gSuppressErrorMsg){
	show_error_msg_head() << msg << '\n';
	check_num_errors();
    }
    else{
#ifdef YYDEBUG
	if(yydebug)
	    show_error_msg_head() << "backtrack\n";

	++nbacktracks;
#endif
	--yynerrs;
    }
}

void ErrorMessage(char* msg, Ptree* name1, Ptree* name2)
{
    show_error_msg_head() << msg;
    if(name1 != nil){
	cerr << " \"";
	name1->Write(cerr);
	cerr << '\"';
    }

    if(name2 != nil){
	cerr << " in ";
	name2->Write(cerr);
    }

    cerr << '\n';
    ++yynerrs;
    check_num_errors();
}

void WarningMessage(char* msg, Ptree* name1, Ptree* name2)
{
    show_error_msg_head() << "warning " << msg;
    if(name1 != nil){
	cerr << " \"";
	name1->Write(cerr);
	cerr << '\"';
    }

    if(name2 != nil){
	cerr << " in ";
	name2->Write(cerr);
    }

    cerr << '\n';
}

void ShowProgram(BOOL on)
{
    showPtree = on;
}

BOOL NumOfErrors()
{
    return yynerrs;
}

#ifdef YYDEBUG

#include "buffer.h"

main(int argc, char** argv)
{
    extern int yydebug;

    yydebug = (argc > 1 && *argv[1] == 'd');
    showPtree = (argc > 1 && *argv[1] == 's');
    Token::SetSource(new ProgramFromStdin());
    cerr << "yyparse() == " << yyparse()
	 << " (error " << yynerrs
	 << ", backtrack " << nbacktracks << ")\n";
}

#endif YYDEBUG
