/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/

#ifndef _ptree_core_h
#define _ptree_core_h

#include <stdio.h>
#include <iostream.h>
#include "types.h"
#include "gc/gc_cpp.h"

#define NIL	((Ptree*)0)

union TokenVal;
class NonLeaf;
class Environment;
class TypeInfo;
class Binding;
class HashTable;
class Class;
class PtreeArray;
class ProgramString;

/*
   If the following enum is modified, PtreeIdTable in ptree2.cc
   must be updated as well.
*/

enum PtreeId {
    NonLeafId, LeafId,

    LeafTypeNameId, LeafClassNameId, PtreeQClassNameId,

// toplevel declarations
    PtreeDeclarationId, PtreeFunctionId, PtreeTemplateDeclId,
    PtreeMetaclassId, PtreeTypedefId, PtreeLinkageId,

// statements
    PtreeExprStatementId, PtreeLabelId, PtreeCaseLabelId,
    PtreeIfStatementId, PtreeSwitchStatementId, PtreeWhileStatementId,
    PtreeDoStatementId, PtreeForStatementId, PtreeBreakStatementId,
    PtreeContinueStatementId, PtreeReturnStatementId, PtreeGotoStatementId,

// expressions
    PtreeInfixExprId, PtreeCondExprId, PtreeCastId, PtreeUnaryExprId,
    PtreePostfixExprId, PtreeMemberAccessExprId,
    PtreeSizeofExprId, PtreeAssignExprId,
    PtreeSizeofTypeId, PtreePtrToMemExprId, PtreeNewId, PtreeDeleteId, 
    PtreeFstyleCastId, PtreeUserStatementId, PtreeStaticUserStatementId,

    PtreeEnumId, PtreeClassId,

    PtreeParenId, PtreeBracketId, PtreeBraceId, PtreeAngleId,

    LeafPointerId, LeafReferenceId, LeafConstValueId, LeafNameId,
    LeafReservedId, LeafThisId, LeafReservedModifierId,
    PtreePtrToMemberId, PtreeCvQualifierId, PtreeQnameId,
    PtreeDestructorId, PtreeOperatorFuncId, PtreeTemplateNameId,
    PtreeAccessCtrlId, PtreeUserKeywordId, PtreeActualArgsId,

// miscellaneous
    PtreeChangedId, PtreeNonExprId,
};

#ifdef __opencxx
metaclass Ptree : QuoteClass;		// get qMake() available
#endif

class Ptree : public gc_cleanup {
public:
    virtual ~Ptree();
    virtual BOOL IsA(PtreeId) = nil;
    virtual PtreeId WhatIs() = nil;
    BOOL IsLeaf() { return IsA(LeafId); }
    BOOL Eq(char);
    BOOL Eq(char*);
    BOOL Eq(Ptree* p) { return Eq(this, p); }

    void Display();
    void Print();
    int Write(FILE*);
    int Write(ostream&);
    char* MakeString();
    virtual void MakeString2(ProgramString&, BOOL&) = nil;
    virtual int Write(ostream&, int indent) = nil;
    virtual void Print(int indent, int maxdepth) = nil;

protected:
    Ptree* Car() { return data.nonleaf.child; }
    Ptree* Cdr() { return data.nonleaf.next; }
    Ptree* Cadr() { return Cdr()->Car(); }
    Ptree* Cddr() { return Cdr()->Cdr(); }
    char* GetPosition() { return data.leaf.position; }
    int GetLength() { return data.leaf.length; }

    void PrintIndent(int, ostream&);

private:
    void SetCar(Ptree* p) { data.nonleaf.child = p; }
    void SetCdr(Ptree* p) { data.nonleaf.next = p; }

protected:
    union {
	struct {
	    Ptree* child;
	    Ptree* next;
	} nonleaf;
	struct {
	    char* position;
	    int  length;
	} leaf;
    }data;

// in ptree2.cc
public:
    static BOOL Eq(Ptree*, char);
    static BOOL Eq(Ptree*, char*);
    static BOOL Eq(Ptree*, Ptree*);
    static Ptree* Last(Ptree*);
    static Ptree* First(Ptree*);
    static Ptree* Rest(Ptree*);
    static Ptree* Second(Ptree*);
    static Ptree* Third(Ptree*);
    static Ptree* Nth(Ptree*, int);
    static int Length(Ptree*);
    static Ptree* ListTail(Ptree*, int);

    static Ptree* Cons(Ptree*, Ptree*);
    static Ptree* List();
    static Ptree* List(Ptree*);
    static Ptree* List(Ptree*, Ptree*);
    static Ptree* List(Ptree*, Ptree*, Ptree*);
    static Ptree* List(Ptree*, Ptree*, Ptree*, Ptree*);
    static Ptree* List(Ptree*, Ptree*, Ptree*, Ptree*, Ptree*);
    static Ptree* List(Ptree*, Ptree*, Ptree*, Ptree*, Ptree*, Ptree*);
    static Ptree* List(Ptree*, Ptree*, Ptree*, Ptree*, Ptree*, Ptree*,
		       Ptree*);
    static Ptree* CopyList(Ptree*);
    static Ptree* Append(Ptree*, Ptree*);
    static Ptree* Substitute(Ptree*, int, Ptree*, Ptree*);
    static Ptree* Substitute(Ptree*, int, Ptree*, Ptree*,
			     int, Ptree*, Ptree*);
    static Ptree* Substitute(Ptree*, int, Ptree*, Ptree*,
			     int, Ptree*, Ptree*, int, Ptree*, Ptree*);
    static Ptree* Substitute(Ptree*, int, Ptree*, Ptree*,
			     int, Ptree*, Ptree*, int, Ptree*, Ptree*,
			     int, Ptree*, Ptree*);

    static Ptree* ReplaceAll(Ptree*, Ptree*, Ptree*);
    static Ptree* Replace(Ptree*, Ptree*, Ptree*);
    static Ptree* ReplaceTail(Ptree*, Ptree*, Ptree*);
    static Ptree* Map(Environment*, Ptree*,
		      Ptree* (*)(Environment*, Ptree*));
    static Ptree* Map2(Environment*, Ptree*,
		       Ptree* (*)(Environment*, Ptree*));

    /* they cause side-effect */
    static Ptree* Snoc(Ptree*, Ptree*);
    static Ptree* Nconc(Ptree*, Ptree*);
    static Ptree* Nconc(Ptree*, Ptree*, Ptree*);

    static char* IdName(PtreeId);

// in pattern.cc
public:
    static BOOL Match(Ptree*, char*, ...);
    static Ptree* Make(char* pat, ...);
    static Ptree* MakeStatement(char* pat, ...);
    static Ptree* GenSym();

    static Ptree* qMake(char*);
    static Ptree* qMakeStatement(char*);

    BOOL Reify(unsigned int&);
    BOOL Reify(char*&);

    static char* IntegerToString(int, int&);	// not documented

private:
    static char* MatchPat(Ptree*, char*);
    static char* MatchList(Ptree*, char*);
    static char* MatchWord(Ptree*, char*);

// in walk.cc and walk2.cc
public:
    virtual void WalkToplevel(Environment*);
    virtual Ptree* WalkStatement(Environment*);
    virtual Ptree* WalkMemberDeclaration(Environment*,
					 PtreeArray&, PtreeArray&);
    virtual Ptree* NthMemberName(int&);
    virtual BOOL IsMember(Ptree*, int&);

    static void AppendToplevel(Ptree*);

    static void RecordBeforeToplevel(Ptree*);
    static void RecordAfterToplevel(Ptree*);

    static Ptree* WalkClassBody(Environment*, Ptree*,
				PtreeArray&, PtreeArray&);

    static Ptree* FindNthMemberName(Ptree*, int);
    static int FindMemberName(Ptree*, Ptree*);

    static Ptree* WalkTypeSpecifier(Environment*, Ptree*);
    static Ptree* WalkMemberInitializers(Environment*, Ptree*);
    static Ptree* WalkMemberInitializersCore(Environment*, Ptree*);
    static Ptree* WalkDeclarators(Environment*, Ptree*, Ptree*,
				  PtreeArray&, PtreeArray&);
    static Ptree* WalkArguments(Environment*, Ptree*,
				PtreeArray&, PtreeArray&);
    static void RecordDeclarator(Environment*, Ptree*, Ptree*);
    static Ptree* WalkDeclarator(Environment*, Ptree*,
				 PtreeArray&, PtreeArray&);
    static Ptree* WalkQualifiedName(Environment*, Ptree*);
    static Ptree* WalkClassName(Environment*&, Ptree*);

    static Ptree* WalkExpr(Environment*, Ptree*);
    virtual Ptree* WalkExpr(Environment*, TypeInfo&);
    virtual void Type(Environment*, TypeInfo&);
    static Ptree* WalkCastType(Environment* env, Ptree* type);
    static Ptree* WalkExprList(Environment*, Ptree*);
    static Ptree* WalkExprListCore(Environment*, Ptree*);
    static Ptree* CheckEquiv(Ptree*, Ptree*);

protected:
    static Environment* MakeEnvironment(Environment*);
    static void DeleteTopEnvironment(Environment*);
    static void ClearBeforeAfterToplevel();
    static void InsertBeforeAfterToplevel(Ptree*);

    static Ptree* WalkInitializerEtc(Environment*, Ptree*, Ptree*,
				 PtreeArray&, PtreeArray&);
    static Ptree* WalkQualifiedName2(Environment*, Ptree*);

    static Ptree* program_tree;
    static Ptree* last_appended_subtree;

    static Ptree* before_toplevel;
    static Ptree* after_toplevel;

friend class Token;
friend class Syntax;
friend class Environment;
friend class PtreeIter;
friend class PtreeHead;
friend class Leaf;
friend class NonLeaf;
friend class PtreeBrace;
friend class TemplateFunction;
friend int yyparse();
};

inline ostream& operator << (ostream& s, Ptree* p)
{
    p->Write(s);
    return s;
}

// in ptree.cc

class PtreeIter {
public:
    PtreeIter(Ptree* p) { ptree = p; }
    Ptree* operator () () { return Pop(); }
    Ptree* Pop();
    BOOL Next(Ptree*&);
    void Reset(Ptree* p) { ptree = p; }

    Ptree* operator * () { return This(); }
    Ptree* operator ++ () { Pop(); return This(); }
    Ptree* operator ++ (int) { return Pop(); }	// postfix
    Ptree* This();
    BOOL Empty() { return BOOL(ptree == nil); }

private:
    Ptree* ptree;
};

class PtreeArray {
public:
    PtreeArray(int = 8);
    ~PtreeArray();
    int Number() { return num; }
    Ptree*& operator [] (uint index) { return Ref(index); }
    Ptree*& Ref(uint index);
    void Append(Ptree*);
    Ptree* All();

private:
    uint num, size;
    Ptree** array;
};

// PtreeHead is used to implement Ptree::qMake()
// The implementation is in pattern.cc

class PtreeHead {
public:
    PtreeHead() { ptree = nil; }
    operator Ptree* () { return ptree; }
    PtreeHead& operator + (Ptree*);
    PtreeHead& operator + (char*);
    PtreeHead& operator + (char);
    PtreeHead& operator + (int);
    static Ptree* Statement(Ptree*);

private:
    Ptree* Append(Ptree*, Ptree*);
    Ptree* Append(Ptree*, char*, int);

private:
    Ptree* ptree;
};

void MopErrorMessage(char* method_name, char* msg);
void MopWarningMessage(char* method_name, char* msg);
void MopMoreWarningMessage(char* msg);

// these are defined in parser.y
void ErrorMessage(char*, Ptree* = nil, Ptree* = nil);
void WarningMessage(char*, Ptree* = nil, Ptree* = nil);

#endif /* _ptree_core_h */
