/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/

#include <iostream.h>
#include <fstream.h>
#include "ptree.h"
#include "token.h"
#include "buffer.h"
#include "y.tab.h"
#include "hash.h"

const char* nilSymbol = "[]";

void MopErrorMessage(char* name, char* msg)
{
    cerr << "MOP error: " << msg << " in " << name << '\n';
    exit(1);
}

void MopWarningMessage(char* name, char* msg)
{
    cerr << "MOP warning: " << msg << " in " << name << '\n';
}

void MopMoreWarningMessage(char* msg)
{
    cerr << "             " << msg << '\n';
}

Ptree::~Ptree() {}

BOOL Ptree::Eq(char c)
{
    return(IsLeaf() && GetLength() == 1 && *GetPosition() == c);
}

BOOL Ptree::Eq(char* str)
{
    if(IsLeaf()){
	char* p = GetPosition();
	int n = GetLength();
	int i;
	for(i = 0; i < n; ++i)
	    if(p[i] != str[i] || str[i] == '\0')
		return FALSE;

	return BOOL(str[i] == '\0');
    }
    else
	return FALSE;
}

/*
  Lists are surrounded by [ and ].  The dodted pair is displayed like
  [ a @ b ].
*/
void Ptree::Display()
{
    Print();
    cout << '\n';
}

void Ptree::Print()
{
    if(this == nil)
	cout << nilSymbol;
    else
	Print(0, 32);
}

char* Ptree::MakeString()
{
    char* str;
    char* buf;
    int len;

    if(this == nil)
	return nil;
    else if(IsLeaf()){
	str = GetPosition();
	len = GetLength();
#ifdef DONT_GC
	buf = new char[len + 1];
#else
	buf = (char*)GC_MALLOC_ATOMIC(len + 1);
#endif
	bcopy(str, buf, len);
	buf[len] = '\0';
	return buf;
    }
    else{
	BOOL last = FALSE;
	ProgramString ps;
	MakeString2(ps, last);
	str = (char*)ps.Read(0);
	len = ps.Length();
#ifdef DONT_GC
	buf = new char[len + 1];
#else
	buf = (char*)GC_MALLOC_ATOMIC(len + 1);
#endif
	bcopy(str, buf, len);
	buf[len] = '\0';
	return buf;
    }
}

void NonLeaf::MakeString2(ProgramString& ps, BOOL& last)
{
    Ptree* p = this;
    for(;;){
	Ptree* head = p->Car();
	if(head != nil)
	    head->MakeString2(ps, last);

	p = p->Cdr();
	if(p == nil)
	    break;
	else if(p->IsLeaf()){
	    MopErrorMessage("NonLeaf::MakeString()", "not list");
	    break;
	}
    }
}

inline BOOL is_letter(unsigned char c)
{
    return('0' <= c && c <= '9'
		|| 'A' <= c && c <= 'Z' || 'a' <= c && c <= 'z'
		|| c == '_' || c == '$');
}

void Leaf::MakeString2(ProgramString& ps, BOOL& last)
{
    char* ptr = GetPosition();
    int len = GetLength();

    if(len <= 0)
	return;

    if(last && is_letter(*ptr))
	ps << ' ';

    while(len-- > 1)
	ps << *ptr++;

    ps << *ptr;
    last = is_letter(*ptr);
}

int Ptree::Write(FILE* fptr)
{
    ofstream out(fileno(fptr));

    if(this == nil)
	return 0;
    else
	return Write(out, 0);
}

int Ptree::Write(ostream& out)
{
    if(this == nil)
	return 0;
    else
	return Write(out, 0);
}

int Leaf::Write(ostream& out, int indent)
{
    int n = 0;
    char* ptr = GetPosition();
    int len = GetLength();
    while(len-- > 0){
	char c = *ptr++;
	if(c == '\n'){
	    PrintIndent(indent, out);
	    ++n;
	}
	else
	    out << c;
    }

    return n;
}

int NonLeaf::Write(ostream& out, int indent)
{
    int n = 0;
    Ptree* p = this;
    for(;;){
	Ptree* head = p->Car();
	if(head != nil)
	    n += head->Write(out, indent);

	p = p->Cdr();
	if(p == nil)
	    break;
	else if(p->IsLeaf()){
	    MopErrorMessage("NonLeaf::Write()", "not list");
	    break;
	}
	else
	    out << ' ';
    }

    return n;
}

void Ptree::PrintIndent(int indent, ostream& out)
{
    out << '\n';
    for(int i = 0; i < indent; ++i)
	out << "    ";
}

Ptree* PtreeIter::Pop()
{
    if(ptree == nil)
	return nil;
    else{
	Ptree* p = ptree->Car();
	ptree = ptree->Cdr();
	return p;
    }
}

BOOL PtreeIter::Next(Ptree*& car)
{
    if(ptree == nil)
	return FALSE;
    else{
	car = ptree->Car();
	ptree = ptree->Cdr();
	return TRUE;
    }
}

Ptree* PtreeIter::This()
{
    if(ptree == nil)
	return nil;
    else
	return ptree->Car();
}

PtreeArray::PtreeArray(int s)
{
    num = 0;
    size = s;
    if(s > 0){
#ifdef DONT_GC
	array = new Ptree*[s];
#else
	array = (Ptree**)GC_MALLOC_UNCOLLECTABLE(sizeof(Ptree*) * s);
#endif
    }
}

PtreeArray::~PtreeArray()
{
#ifdef DONT_GC
    delete [] array;
#else
    GC_FREE(array);
#endif
}

void PtreeArray::Append(Ptree* p)
{
    if(num >= size){
	size += 8;
#ifdef DONT_GC
	Ptree** a = new Ptree*[size];
	bcopy((char*)array, (char*)a, num * sizeof(Ptree*));
	delete [] array;
	array = a;
#else
	array = (Ptree**)GC_REALLOC(array, sizeof(Ptree*) * size);
#endif
    }

    array[num++] = p;
}

Ptree*& PtreeArray::Ref(uint i)
{
    if(i < num)
	return array[i];
    else
	MopErrorMessage("PtreeArray", "out of range");
}

Ptree* PtreeArray::All()
{
    Ptree* lst = nil;

    for(int i = Number() - 1; i >= 0; --i)
	lst = Ptree::Cons(Ref(i), lst);

    return lst;
}

NonLeaf::~NonLeaf()
{
#ifdef DONT_GC
    delete data.nonleaf.child;
    delete data.nonleaf.next;
#endif
}

PtreeId NonLeaf::WhatIs()
{
    return NonLeafId;
}

BOOL NonLeaf::IsA(PtreeId id)
{
    return(id == NonLeafId || id == WhatIs());
}

void NonLeaf::CopyAndDelete(Ptree* p)
{
    if(!p->IsA(NonLeafId))
	MopErrorMessage("NonLeaf::CopyAndDelete", "not list");

    NonLeaf* q = (NonLeaf*)p;
    SetCar(q->Car());
    SetCdr(q->Cdr());
    q->SetNil();
    delete q;
}

void NonLeaf::Print(int indent, int maxdepth)
{
    if(maxdepth <= 0){
	cout << "???";
	return;
    }

    cout << '[';
    Ptree* p = this;
    while(p != nil){
	if(p->IsLeaf()){
	    cout << "@ ";
	    p->Print(indent, maxdepth - 1);
	    break;
	}
	else{
	    Ptree* q = p->Car();
	    if(q == nil)
		cout << nilSymbol;
	    else
		q->Print(indent, maxdepth - 1);
	}

	p = p->Cdr();
	if(p != nil)
	    cout << ' ';
    }

    cout << ']';
}

Leaf::Leaf(TokenVal& t)
{
    data.leaf.position = t.leaf.position;
    data.leaf.length = t.leaf.length;
}

Leaf::~Leaf() {}

PtreeId Leaf::WhatIs()
{
    return LeafId;
}

BOOL Leaf::IsA(PtreeId id)
{
    return(id == LeafId || id == WhatIs());
}

/*
  [, ], and @ are special characters.
*/
void Leaf::Print(int, int)
{
    int i, j;
    char* pos;

    pos = GetPosition();
    j = GetLength();

    if(j == 1 && *pos == '@'){
	cout << "\\@";
	return;
    }

    for(i = 0; i < j; ++i)
	if(pos[i] == '[' || pos[i] == ']'){
	    cout << '\\';
	    cout << pos[i];
	}
	else
	    cout << pos[i];
}

UserLeaf::UserLeaf(char* str, int len) : Leaf(new char[len], len)
{
    bcopy(str, data.leaf.position, len);
}

UserLeaf::UserLeaf(char* str1, int len1, char* str2, int len2)
: Leaf(new char[len1 + len2], len1 + len2)
{
    bcopy(str1, data.leaf.position, len1);
    bcopy(str2, &data.leaf.position[len1], len2);
}

UserLeaf::~UserLeaf()
{
    delete data.leaf.position;
}

void UserLeaf::Print(int, int)
{
    int i, j;
    char* pos;

    pos = GetPosition();
    j = GetLength();

    if(j == 1 && *pos == '@'){
	cout << "\\@";
	return;
    }

    cout << '`';
    for(i = 0; i < j; ++i)
	if(pos[i] == '[' || pos[i] == ']'){
	    cout << '\\';
	    cout << pos[i];
	}
	else
	    cout << pos[i];

    cout << '`';
}

LeafTypeName::LeafTypeName(TokenVal& t) : Leaf(t) {}

PtreeId LeafTypeName::WhatIs()
{
    return LeafTypeNameId;
}

PtreeId LeafClassName::WhatIs()
{
    return LeafClassNameId;
}

PtreeId PtreeQClassName::WhatIs()
{
    return PtreeQClassNameId;
}

PtreeBlock::PtreeBlock(TokenVal& t1, TokenVal& t2, TokenVal& t3)
: NonLeaf(new Leaf(t1), List(t2.ptree, new Leaf(t3))) {}

PtreeBlock::PtreeBlock(TokenVal& t1, Ptree* p2, TokenVal& t3)
: NonLeaf(new Leaf(t1), List(p2, new Leaf(t3))) {}

PtreeBlock::PtreeBlock(TokenVal& t1, Ptree* p2, Ptree* p3)
: NonLeaf(new Leaf(t1), List(p2, p3)) {}

PtreeBlock::PtreeBlock(Ptree* p1, Ptree* p2, Ptree* p3)
: NonLeaf(p1, List(p2, p3)) {}

PtreeId PtreeParen::WhatIs()
{
    return PtreeParenId;
}

PtreeId PtreeBrace::WhatIs()
{
    return PtreeBraceId;
}

void PtreeBrace::Print(int indent, int maxdepth)
{
    Ptree* p = this->Cadr();
    if(p != nil && p->IsLeaf()){
	cout << "[{ ";
	p->Print(indent, maxdepth - 1);
	cout << " }]";
	return;
    }

    cout << "[{ [";
    while(p != nil){
	PrintIndent(indent + 1, cout);
	if(p->IsLeaf()){
	    cout << "@ ";
	    p->Print(indent + 1, maxdepth - 1);
	    break;
	}
	else{
	    Ptree* q = p->Car();
	    if(q == nil)
		cout << nilSymbol;
	    else
		q->Print(indent + 1, maxdepth - 1);
	}

	p = p->Cdr();
    }

    PrintIndent(indent, cout);
    cout << "] }]";
}

int PtreeBrace::Write(ostream& out, int indent)
{
    int n = 0;

    out << '{';
    Ptree* p = this->Cadr();
    while(p != nil){
	if(p->IsLeaf()){
	    MopErrorMessage("PtreeBrace::Write()", "non list");
	    break;
	}
	else{
	    PrintIndent(indent + 1, out);
	    ++n;
	    Ptree* q = p->Car();
	    p = p->Cdr();
	    if(q != nil)
		n += q->Write(out, indent + 1);
	}
    }

    PrintIndent(indent, out);
    ++n;
    out << '}';
    return n;
}

PtreeId PtreeBracket::WhatIs()
{
    return PtreeBracketId;
}

PtreeId PtreeAngle::WhatIs()
{
    return PtreeAngleId;
}

LeafPointer::LeafPointer(TokenVal& t) : Leaf(t) {}

LeafPointer::LeafPointer() : Leaf("*", 1) {}

PtreeId LeafPointer::WhatIs()
{
    return LeafPointerId;
}

LeafReference::LeafReference(TokenVal& t) : Leaf(t) {}

PtreeId LeafReference::WhatIs()
{
    return LeafReferenceId;
}

PtreePtrToMember::PtreePtrToMember(TokenVal& t1, TokenVal& t2, TokenVal& t3)
: NonLeaf(nil, nil)
{
    Ptree* p = List(new Leaf(t2), new Leaf(t3));
    if(t1.ptree->IsA(PtreeQClassNameId)){
	Nconc(t1.ptree, p);
	CopyAndDelete(t1.ptree);
    }
    else{
	SetCar(t1.ptree);
	SetCdr(p);
    }
}

PtreePtrToMember::PtreePtrToMember(Ptree* p1, Ptree* p2, Ptree* p3)
: NonLeaf(p1, List(p2, p3)) {}

PtreeId PtreePtrToMember::WhatIs()
{
    return PtreePtrToMemberId;
}

PtreeCvQualifier::PtreeCvQualifier(TokenVal& t)
: NonLeaf(new LeafReserved(t), nil) {}

PtreeId PtreeCvQualifier::WhatIs()
{
    return PtreeCvQualifierId;
}

LeafName::LeafName(TokenVal& t) : Leaf(t) {}

PtreeId LeafName::WhatIs()
{
    return LeafNameId;
}

PtreeQname::PtreeQname(TokenVal& t1, TokenVal& t2)
: NonLeaf(new Leaf(t1), List(t2.ptree)) {}

/*
  PtreeQname objects don't contain PtreeQClassName objects.
*/
PtreeQname::PtreeQname(TokenVal& t1, TokenVal& t2, TokenVal& t3)
: NonLeaf(nil, nil)
{
    if(t1.ptree->IsA(PtreeQClassNameId))
	CopyAndDelete(t1.ptree);
    else // if(t1.ptree->IsA(LeafClassNameId))
	SetCar(t1.ptree);

    Nconc(this, List(new Leaf(t2), t3.ptree));
}

PtreeQname::PtreeQname(Ptree* constructor_name)
: NonLeaf(nil, nil)
{
    if(constructor_name->IsA(PtreeQClassNameId))
	CopyAndDelete(constructor_name);
    else // if(constructor_name->IsA(LeafClassNameId))
	SetCar(constructor_name);
}

PtreeQname::PtreeQname(Ptree* p1, Ptree* p2)
: NonLeaf(p1, p2) {}

PtreeId PtreeQname::WhatIs()
{
    return PtreeQnameId;
}

PtreeDestructor::PtreeDestructor(TokenVal& t1, TokenVal& t2)
: NonLeaf(new Leaf(t1), List(t2.ptree)) {}

BOOL PtreeDestructor::IsA(PtreeId id)
{
    return(id == LeafNameId || NonLeaf::IsA(id));
}

PtreeId PtreeDestructor::WhatIs()
{
    return PtreeDestructorId;
}

PtreeOperatorFunc::PtreeOperatorFunc(TokenVal& t1, TokenVal& t2)
: NonLeaf(new LeafReserved(t1), List(t2.ptree)) {}

PtreeId PtreeOperatorFunc::WhatIs()
{
    return PtreeOperatorFuncId;
}

LeafReserved::LeafReserved(TokenVal& t) : Leaf(t) {}

PtreeId LeafReserved::WhatIs()
{
    return LeafReservedId;
}

LeafReservedModifier::LeafReservedModifier(TokenVal& t) : LeafReserved(t) {}

PtreeId LeafReservedModifier::WhatIs()
{
    return LeafReservedModifierId;
}

BOOL LeafReservedModifier::IsA(PtreeId id)
{
    return(id == LeafReservedModifierId || id == LeafReservedId
	   || id == LeafId);
}

LeafThis::LeafThis(TokenVal& t) : LeafReserved(t) {}

PtreeId LeafThis::WhatIs()
{
    return LeafThisId;
}

BOOL LeafThis::IsA(PtreeId id)
{
    return(id == LeafThisId || id == LeafReservedId
	   || id == LeafId);
}

PtreeTemplateName::PtreeTemplateName(TokenVal& t1, TokenVal& t2,
				     TokenVal& t3, TokenVal& t4)
: NonLeaf(new Leaf(t1), List(new PtreeAngle(t2, t3, t4))) {}

PtreeTemplateName::PtreeTemplateName(Ptree* p1, Ptree* p2, Ptree* p3,
				     Ptree* p4)
: NonLeaf(p1, List(new PtreeAngle(p2, p3, p4))) {}

PtreeId PtreeTemplateName::WhatIs()
{
    return PtreeTemplateNameId;
}

PtreeEnum::PtreeEnum(TokenVal& t1, TokenVal& t2)
: NonLeaf(new LeafReserved(t1), List(new Leaf(t2))) {}

PtreeEnum::PtreeEnum(TokenVal& t, Ptree* p)
: NonLeaf(new Leaf(t), List(nil, p)) {}

PtreeEnum::PtreeEnum(TokenVal& t1, TokenVal& t2, Ptree* p)
: NonLeaf(new Leaf(t1), List(new Leaf(t2), p)) {}

PtreeId PtreeEnum::WhatIs()
{
    return PtreeEnumId;
}

PtreeClass::PtreeClass(TokenVal& t1, Ptree* p2)
: NonLeaf(new LeafReserved(t1), List(p2)) {}

PtreeId PtreeClass::WhatIs()
{
    return PtreeClassId;
}

PtreeAccessCtrl::PtreeAccessCtrl(TokenVal& t1, TokenVal& t2)
: NonLeaf(new LeafReserved(t1), List(new Leaf(t2))) {}

PtreeId PtreeAccessCtrl::WhatIs()
{
    return PtreeAccessCtrlId;
}

PtreeDeclaration::PtreeDeclaration(Ptree* p1, Ptree* p2, TokenVal& t3)
: NonLeaf(p1, Ptree::Snoc(p2, new Leaf(t3))) {}

PtreeDeclaration::PtreeDeclaration(Ptree* p1, Ptree* p2, Ptree* p3)
: NonLeaf(p1, Ptree::Snoc(p2, p3)) {}

PtreeId PtreeDeclaration::WhatIs()
{
    return PtreeDeclarationId;
}

PtreeFunction::PtreeFunction(Ptree* p1, Ptree* p2)
: NonLeaf(p1, List(p2)) {}

PtreeFunction::PtreeFunction(Ptree* p1, Ptree* p2, Ptree* p3)
: NonLeaf(p1, List(p2, p3)) {}

PtreeFunction::PtreeFunction(Ptree* p1, Ptree* p2, Ptree* p3, Ptree* p4)
: NonLeaf(p1, List(p2, p3, p4)) {}

PtreeId PtreeFunction::WhatIs()
{
    return PtreeFunctionId;
}

PtreeTemplateDecl::PtreeTemplateDecl(Ptree* p1, TokenVal& p2, Ptree* p3,
				     TokenVal& p4, Ptree* p5)
: NonLeaf(p1, List(new Leaf(p2), p3, new Leaf(p4), p5)) {}

PtreeTemplateDecl::PtreeTemplateDecl(Ptree* p1, Ptree* p2, Ptree* p3,
				     Ptree* p4, Ptree* p5)
: NonLeaf(p1, List(p2, p3, p4, p5)) {}

PtreeId PtreeTemplateDecl::WhatIs()
{
    return PtreeTemplateDeclId;
}

PtreeMetaclass::PtreeMetaclass(TokenVal& t1, Ptree* p2, TokenVal& t3,
			       TokenVal& t4, Ptree* p5, TokenVal& t6)
: NonLeaf(new LeafReserved(t1), List(p2, new Leaf(t3),
				     new Leaf(t4), p5, new Leaf(t6))) {}

PtreeId PtreeMetaclass::WhatIs()
{
    return PtreeMetaclassId;
}

BOOL PtreeMetaclass::RegisterKeyword(char* keyword, char* metaclass_name)
{
    if(keywords == nil)
	keywords = new HashTable;

    return BOOL(Token::RecordKeyword(keyword, UserKeyword)
		&& keywords->AddEntry(keyword, metaclass_name) >= 0);
}

PtreeTypedef::PtreeTypedef(TokenVal& t1, Ptree* p2, Ptree* p3, TokenVal& t4)
: NonLeaf(new LeafReserved(t1), Cons(p2, Snoc(p3, new Leaf(t4)))) {}

PtreeId PtreeTypedef::WhatIs()
{
    return PtreeTypedefId;;
}

PtreeLinkage::PtreeLinkage(TokenVal& t1, TokenVal& t2, Ptree* p3)
: NonLeaf(new LeafReserved(t1), List(new Leaf(t2), p3)) {}

PtreeId PtreeLinkage::WhatIs()
{
    return PtreeLinkageId;
}

PtreeUserKeyword::PtreeUserKeyword(TokenVal& t1, Ptree* p2)
: NonLeaf(new Leaf(t1), List(p2)) {}

PtreeUserKeyword::PtreeUserKeyword(TokenVal& t1, Ptree* p2, TokenVal& t3)
: NonLeaf(new Leaf(t1), List(p2, new Leaf(t3))) {}

PtreeId PtreeUserKeyword::WhatIs()
{
    return PtreeUserKeywordId;
}

PtreeAssignExpr::PtreeAssignExpr(Ptree* p1, Ptree* p2, Ptree* p3)
: NonLeaf(p1, List(p2, p3)) {}

PtreeId PtreeAssignExpr::WhatIs()
{
    return PtreeAssignExprId;
}

PtreeInfixExpr::PtreeInfixExpr(TokenVal& t1, TokenVal& t2, TokenVal& t3)
: NonLeaf(t1.ptree, List(new Leaf(t2), t3.ptree)) {}

PtreeInfixExpr::PtreeInfixExpr(Ptree* p1, Ptree* p2, Ptree* p3)
: NonLeaf(p1, List(p2, p3)) {}

PtreeId PtreeInfixExpr::WhatIs()
{
    return PtreeInfixExprId;
}

PtreeCondExpr::PtreeCondExpr(TokenVal& t1, TokenVal& t2, TokenVal& t3,
			     TokenVal& t4, TokenVal& t5)
: NonLeaf(t1.ptree, List(new Leaf(t2), t3.ptree, new Leaf(t4), t5.ptree)) {}

PtreeCondExpr::PtreeCondExpr(Ptree* p1, Ptree* p2, Ptree* p3, Ptree* p4,
			     Ptree* p5)
: NonLeaf(p1, List(p2, p3, p4, p5)) {}

PtreeId PtreeCondExpr::WhatIs()
{
    return PtreeCondExprId;
}

PtreeCast::PtreeCast(TokenVal& t1, TokenVal& t2, TokenVal& t3, TokenVal& t4)
: NonLeaf(new PtreeParen(t1, t2, t3), List(t4.ptree)) {}

PtreeCast::PtreeCast(Ptree* p1, Ptree* p2)
: NonLeaf(p1, List(p2)) {}

PtreeId PtreeCast::WhatIs()
{
    return PtreeCastId;
}

PtreeUnaryExpr::PtreeUnaryExpr(TokenVal& t1, TokenVal& t2)
: NonLeaf(new Leaf(t1), List(t2.ptree)) {}

PtreeUnaryExpr::PtreeUnaryExpr(Ptree* p1, Ptree* p2)
: NonLeaf(p1, List(p2)) {}

PtreeId PtreeUnaryExpr::WhatIs()
{
    return PtreeUnaryExprId;
}

PtreePostfixExpr::PtreePostfixExpr(Ptree* p1, Ptree* p2)
: NonLeaf(p1, List(p2)) {}

PtreeId PtreePostfixExpr::WhatIs()
{
    return PtreePostfixExprId;
}

PtreeMemberAccessExpr::PtreeMemberAccessExpr(Ptree* p1, TokenVal& t2,
					     Ptree* p3)
: NonLeaf(p1, List(new Leaf(t2), p3)) {}

PtreeMemberAccessExpr::PtreeMemberAccessExpr(Ptree* p1, Ptree* p2, Ptree* p3)
: NonLeaf(p1, List(p2, p3)) {}

PtreeMemberAccessExpr::PtreeMemberAccessExpr(Ptree* p1, Ptree* p2, Ptree* p3,
					     Ptree* p4)
: NonLeaf(p1, List(p2, p3, p4)) {}

PtreeMemberAccessExpr::PtreeMemberAccessExpr(Ptree* p1, Ptree* p2, Ptree* p3,
					     Ptree* p4, Ptree* p5)
: NonLeaf(p1, List(p2, p3, p4, p5)) {}

PtreeId PtreeMemberAccessExpr::WhatIs()
{
    return PtreeMemberAccessExprId;
}

PtreePtrToMemExpr::PtreePtrToMemExpr(Ptree* p1, TokenVal& t2, Ptree* p3)
: NonLeaf(p1, List(new Leaf(t2), p3)) {}

PtreePtrToMemExpr::PtreePtrToMemExpr(Ptree* p1, Ptree* p2, Ptree* p3)
: NonLeaf(p1, List(p2, p3)) {}

PtreeId PtreePtrToMemExpr::WhatIs()
{
    return PtreePtrToMemExprId;
}

PtreeSizeofExpr::PtreeSizeofExpr(Ptree* p1, Ptree* p2)
: NonLeaf(p1, List(p2)) {}

PtreeId PtreeSizeofExpr::WhatIs()
{
    return PtreeSizeofExprId;
}

PtreeSizeofType::PtreeSizeofType(Ptree* p1, Ptree* p2)
: NonLeaf(p1, List(p2)) {}

PtreeId PtreeSizeofType::WhatIs()
{
    return PtreeSizeofTypeId;
}

LeafConstValue::LeafConstValue(TokenVal& t) : Leaf(t) {}

PtreeId LeafConstValue::WhatIs()
{
    return LeafConstValueId;
}

PtreeFstyleCast::PtreeFstyleCast(Ptree* p1, TokenVal& t2, TokenVal& t3,
				 TokenVal& t4)
: NonLeaf(p1, List(new PtreeParen(t2, t3, t4))) {}

PtreeFstyleCast::PtreeFstyleCast(Ptree* p1, Ptree* p2)
: NonLeaf(p1, List(p2)) {}

PtreeId PtreeFstyleCast::WhatIs()
{
    return PtreeFstyleCastId;
}

PtreeUserStatement::PtreeUserStatement(TokenVal& t1, TokenVal& t2,
				       TokenVal& t3)
: NonLeaf(t1.ptree, Cons(new Leaf(t2), t3.ptree)) {}

PtreeId PtreeUserStatement::WhatIs()
{
    return PtreeUserStatementId;
}

PtreeStaticUserStatement::PtreeStaticUserStatement(TokenVal& t1, TokenVal& t2)
: NonLeaf(t1.ptree, Cons(t2.ptree, nil)) {}

PtreeId PtreeStaticUserStatement::WhatIs()
{
    return PtreeStaticUserStatementId;
}

PtreeExprStatement::PtreeExprStatement(TokenVal& t1, TokenVal& t2)
: NonLeaf(t1.ptree, List(new Leaf(t2))) {}

PtreeId PtreeExprStatement::WhatIs()
{
    return PtreeExprStatementId;
}

PtreeLabel::PtreeLabel(Ptree* p1, Ptree* p2, Ptree* p3)
: NonLeaf(p1, List(p2, p3)) {}

PtreeId PtreeLabel::WhatIs()
{
    return PtreeLabelId;
}

PtreeCaseLabel::PtreeCaseLabel(Ptree* p1, Ptree* p2, Ptree* p3, Ptree* p4)
: NonLeaf(p1, List(p2, p3, p4)) {}

PtreeId PtreeCaseLabel::WhatIs()
{
    return PtreeCaseLabelId;
}

PtreeIfStatement::PtreeIfStatement(TokenVal& t)
: NonLeaf(new LeafReserved(t), nil) {}

PtreeId PtreeIfStatement::WhatIs()
{
    return PtreeIfStatementId;
}

void PtreeIfStatement::Print(int indent, int maxdepth)
{
    Ptree* cond = Third(this);
    Ptree* thenp = Nth(this, 4);
    Ptree* elsep = Nth(this, 6);

    cout << "[if ( ";
    if(cond == nil)
	cout << nilSymbol;
    else
	cond->Print(indent, maxdepth - 1);

    cout << " )";
    PrintIndent(indent + 1, cout);
    if(thenp != nil)
	thenp->Print(indent + 1, maxdepth - 1);

    if(elsep != nil){
	PrintIndent(indent, cout);
	cout << " else ";
	if(elsep->WhatIs() == PtreeIfStatementId)
	    elsep->Print(indent, maxdepth - 1);
	else{
	    PrintIndent(indent + 1, cout);
	    elsep->Print(indent + 1, maxdepth - 1);
	}
    }

    cout << ']';
}

int PtreeIfStatement::Write(ostream& out, int indent)
{
    int n = 0;
    Ptree* cond = Third(this);
    Ptree* thenp = Nth(this, 4);
    Ptree* elsep = Nth(this, 6);

    out << "if(";
    if(cond != nil)
	n += cond->Write(out, indent);

    out << ')';
    PrintIndent(indent + 1, out);
    ++n;
    if(thenp != nil)
	n += thenp->Write(out, indent + 1);

    if(elsep != nil){
	PrintIndent(indent, out);
	++n;
	out << "else ";
	if(elsep->WhatIs() == PtreeIfStatementId)
	    n += elsep->Write(out, indent);
	else{
	    PrintIndent(indent + 1, out);
	    ++n;
	    n += elsep->Write(out, indent + 1);
	}
    }

    return n;
}

PtreeSwitchStatement::PtreeSwitchStatement(TokenVal& t)
: NonLeaf(new LeafReserved(t), nil) {}

PtreeId PtreeSwitchStatement::WhatIs()
{
    return PtreeSwitchStatementId;
}

PtreeWhileStatement::PtreeWhileStatement(TokenVal& t)
: NonLeaf(new LeafReserved(t), nil) {}

PtreeId PtreeWhileStatement::WhatIs()
{
    return PtreeWhileStatementId;
}

PtreeDoStatement::PtreeDoStatement(TokenVal& t)
: NonLeaf(new LeafReserved(t), nil) {}

PtreeId PtreeDoStatement::WhatIs()
{
    return PtreeDoStatementId;
}

PtreeForStatement::PtreeForStatement(TokenVal& t)
: NonLeaf(new LeafReserved(t), nil) {}

PtreeId PtreeForStatement::WhatIs()
{
    return PtreeForStatementId;
}

PtreeBreakStatement::PtreeBreakStatement(TokenVal& t)
: NonLeaf(new LeafReserved(t), nil) {}

PtreeId PtreeBreakStatement::WhatIs()
{
    return PtreeBreakStatementId;
}

PtreeContinueStatement::PtreeContinueStatement(TokenVal& t)
: NonLeaf(new LeafReserved(t), nil) {}

PtreeId PtreeContinueStatement::WhatIs()
{
    return PtreeContinueStatementId;
}

PtreeReturnStatement::PtreeReturnStatement(TokenVal& t)
: NonLeaf(new LeafReserved(t), nil) {}

PtreeId PtreeReturnStatement::WhatIs()
{
    return PtreeReturnStatementId;
}

PtreeGotoStatement::PtreeGotoStatement(TokenVal& t)
: NonLeaf(new LeafReserved(t), nil) {}

PtreeId PtreeGotoStatement::WhatIs()
{
    return PtreeGotoStatementId;
}

PtreeNew::PtreeNew(Ptree* p1, Ptree* p2, Ptree* p3, Ptree* p4)
: NonLeaf(p1, Cons(p2, p4 == nil ? p3 : Snoc(p3, p4))) {}

PtreeId PtreeNew::WhatIs()
{
    return PtreeNewId;
}

PtreeDelete::PtreeDelete(Ptree* p1, Ptree* p2, Ptree* p3, Ptree* p4)
: NonLeaf(p1, List(p2, p3, p4)) {}

PtreeId PtreeDelete::WhatIs()
{
    return PtreeDeleteId;
}

PtreeActualArgs::PtreeActualArgs(Ptree* p)
: NonLeaf(p, nil) {}

PtreeId PtreeActualArgs::WhatIs()
{
    return PtreeActualArgsId;
}

PtreeChanged::PtreeChanged(Ptree* p)
: NonLeaf(p, nil) {}

PtreeId PtreeChanged::WhatIs()
{
    return PtreeChangedId;
}

PtreeNonExpr::PtreeNonExpr(Ptree* p)
: PtreeChanged(p) {}

PtreeId PtreeNonExpr::WhatIs()
{
    return PtreeNonExprId;
}

BOOL PtreeNonExpr::IsA(PtreeId id)
{
    return(id == PtreeNonExprId || PtreeChanged::IsA(id));
}
