/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/

#ifndef _ptree_h
#define _ptree_h

#include "ptree-core.h"

class NonLeaf : public Ptree {
public:
    NonLeaf(Ptree* p, Ptree* q) {
	data.nonleaf.child = p; data.nonleaf.next = q;
    }

    ~NonLeaf();
    BOOL IsA(PtreeId);
    PtreeId WhatIs();

    void SetCar(Ptree* p) { data.nonleaf.child = p; }
    void SetCdr(Ptree* p) { data.nonleaf.next = p; }
    void SetNil() { data.nonleaf.child = data.nonleaf.next = nil; }
    void CopyAndDelete(Ptree*);

    void Print() { Ptree::Print(); }
    void Print(int, int);
    int Write(ostream&, int);
    void MakeString2(ProgramString&, BOOL&);
};

class Leaf : public Ptree {
public:
    Leaf(char* p, int l) { data.leaf.position = p; data.leaf.length = l; }
    Leaf(TokenVal&);
    ~Leaf();
    BOOL IsA(PtreeId);
    PtreeId WhatIs();

    void Print(int, int);
    int Write(ostream&, int);
    void MakeString2(ProgramString&, BOOL&);

    char* Position() { return data.leaf.position; }
    int Length() { return data.leaf.length; }
};

class UserLeaf : public Leaf {
public:
    UserLeaf(char*, int);    // the text is duplicated at the heap area.
    UserLeaf(char*, int, char*, int);
    ~UserLeaf();
    void Print(int, int);

protected:
    void WalkToplevel(Environment*);
    Ptree* WalkStatement(Environment*);
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class LeafTypeName : public Leaf {
public:
    LeafTypeName(TokenVal& t);
    LeafTypeName(char* p, int l) : Leaf(p, l) {}
    PtreeId WhatIs();
};

class LeafClassName : public Leaf {
public:
    LeafClassName(char* p, int l) : Leaf(p, l) {}
    LeafClassName(TokenVal& t) : Leaf(t) {}
    PtreeId WhatIs();
};

// qualified class
class PtreeQClassName : public NonLeaf {
public:
    PtreeQClassName(Ptree* p) : NonLeaf(p, nil) {}
    PtreeId WhatIs();
};

class PtreeBlock : public NonLeaf {	// abstract class
public:
    PtreeBlock(TokenVal&, TokenVal&, TokenVal&);
    PtreeBlock(TokenVal&, Ptree*, TokenVal&);
    PtreeBlock(TokenVal&, Ptree*, Ptree*);
    PtreeBlock(Ptree*, Ptree*, Ptree*);

    Ptree* Body() { return this->Cadr(); }
};

class PtreeParen : public PtreeBlock {
public:
    PtreeParen(TokenVal& t1, TokenVal& t2, TokenVal& t3)
	: PtreeBlock(t1, t2, t3) {}
    PtreeParen(TokenVal& t1, Ptree* p2, TokenVal& t3)
	: PtreeBlock(t1, p2, t3) {}
    PtreeParen(Ptree* p1, Ptree* p2, Ptree* p3)
	: PtreeBlock(p1, p2, p3) {}
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeBrace : public PtreeBlock {
public:
    PtreeBrace(TokenVal& t1, TokenVal& t2, TokenVal& t3)
	: PtreeBlock(t1, t2, t3) {}
    PtreeBrace(TokenVal& t1, Ptree* p2, TokenVal& t3)
	: PtreeBlock(t1, p2, t3) {}
    PtreeBrace(TokenVal& t1, Ptree* p2, Ptree* p3)
	: PtreeBlock(t1, p2, p3) {}
    PtreeBrace(Ptree* p1, Ptree* p2, Ptree* p3)
	: PtreeBlock(p1, p2, p3) {}
    PtreeId WhatIs();
    void Print(int, int);
    int Write(ostream&, int);

    static Ptree* WalkStatements(Environment*, Ptree*);

protected:
    Ptree* WalkStatement(Environment*);
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeBracket : public PtreeBlock {
public:
    PtreeBracket(TokenVal& t1, TokenVal& t2, TokenVal& t3)
	: PtreeBlock(t1, t2, t3) {}
    PtreeBracket(Ptree* p1, Ptree* p2, Ptree* p3)
	: PtreeBlock(p1, p2, p3) {}
    PtreeId WhatIs();
};

class PtreeAngle : public PtreeBlock {
public:
    PtreeAngle(TokenVal& t1, TokenVal& t2, TokenVal& t3)
	: PtreeBlock(t1, t2, t3) {}
    PtreeAngle(Ptree* p1, Ptree* p2, Ptree* p3)
	: PtreeBlock(p1, p2, p3) {}
    PtreeId WhatIs();
};

class LeafPointer : public Leaf {
public:
    LeafPointer(TokenVal&);
    LeafPointer();
    PtreeId WhatIs();
};

class LeafReference : public Leaf {
public:
    LeafReference(TokenVal&);
    PtreeId WhatIs();
};

class PtreePtrToMember : public NonLeaf {
public:
    PtreePtrToMember(TokenVal&, TokenVal&, TokenVal&);
    PtreePtrToMember(Ptree*, Ptree*, Ptree*);
    PtreeId WhatIs();
};

class PtreeCvQualifier : public NonLeaf {
public:
    PtreeCvQualifier(TokenVal&);
    PtreeCvQualifier(TokenVal&, TokenVal&);
    PtreeId WhatIs();
};

class LeafName : public Leaf {
public:
    LeafName(TokenVal&);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeQname : public NonLeaf {
public:
    PtreeQname(TokenVal&, TokenVal&);
    PtreeQname(TokenVal&, TokenVal&, TokenVal&);
    PtreeQname(Ptree*);
    PtreeQname(Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeDestructor : public NonLeaf {
public:
    PtreeDestructor(TokenVal&, TokenVal&);
    BOOL IsA(PtreeId);	// IsA(LeafNameId) is also ture.
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeOperatorFunc : public NonLeaf {
public:
    PtreeOperatorFunc(TokenVal&, TokenVal&);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class LeafReserved : public Leaf {
public:
    LeafReserved(TokenVal&);
    PtreeId WhatIs();
};

class LeafReservedModifier : public LeafReserved {
public:
    LeafReservedModifier(TokenVal&);
    PtreeId WhatIs();
    BOOL IsA(PtreeId);
};

class LeafThis : public LeafReserved {
public:
    LeafThis(TokenVal&);
    PtreeId WhatIs();
    BOOL IsA(PtreeId);

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeTemplateName : public NonLeaf {
public:
    PtreeTemplateName(TokenVal&, TokenVal&, TokenVal&, TokenVal&);
    PtreeTemplateName(Ptree*, Ptree*, Ptree*, Ptree*);
    PtreeId WhatIs();
    Ptree* Name() { return this->Car(); }
};

class PtreeEnum : public NonLeaf {
public:
    PtreeEnum(TokenVal&, TokenVal&);
    PtreeEnum(TokenVal&, Ptree*);
    PtreeEnum(TokenVal&, TokenVal&, Ptree*);
    PtreeId WhatIs();
    Ptree* Name() { return this->Cadr(); }
};

class PtreeClass : public NonLeaf {
public:
    PtreeClass(TokenVal&, Ptree*);
    PtreeId WhatIs();
    Ptree* Name() { return this->Cadr(); }
};

class PtreeAccessCtrl : public NonLeaf {
public:
    PtreeAccessCtrl(TokenVal&, TokenVal&);
    PtreeId WhatIs();

protected:
    Ptree* WalkMemberDeclaration(Environment*, PtreeArray&, PtreeArray&);
    Ptree* NthMemberName(int&);
    BOOL IsMember(Ptree*, int&);
};

/*
  PtreeDeclaration is for type declarations and variable declaration.
  The null declaration ";" is also represented by this class.
  ";" is parsed into (nil ';').
*/
class PtreeDeclaration : public NonLeaf {
public:
    PtreeDeclaration(Ptree*, Ptree*, TokenVal&);
    PtreeDeclaration(Ptree*, Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    void WalkToplevel(Environment*);
    Ptree* WalkStatement(Environment*);
    Ptree* WalkMemberDeclaration(Environment*, PtreeArray&, PtreeArray&);
    Ptree* NthMemberName(int&);
    BOOL IsMember(Ptree*, int&);
};

class PtreeFunction : public NonLeaf {
public:
    PtreeFunction(Ptree*, Ptree*);
    PtreeFunction(Ptree*, Ptree*, Ptree*);
    PtreeFunction(Ptree*, Ptree*, Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    void WalkToplevel(Environment*);
    Ptree* WalkMemberDeclaration(Environment*, PtreeArray&, PtreeArray&);
    Ptree* NthMemberName(int&);
    BOOL IsMember(Ptree*, int&);
};

class PtreeTemplateDecl : public NonLeaf {
public:
    PtreeTemplateDecl(Ptree*, TokenVal&, Ptree*, TokenVal&, Ptree*);
    PtreeTemplateDecl(Ptree*, Ptree*, Ptree*, Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    void WalkToplevel(Environment*);
    Ptree* TranslateToplevel(Environment*);
    Ptree* TranslateTemplateFunction(Environment*, Ptree*);
    Ptree* TranslateTemplateClass(Environment*, Ptree*);
    Class* MakeMetaobject(Environment*, Ptree*, Ptree*);
};

class PtreeMetaclass : public NonLeaf {
public:
    PtreeMetaclass(TokenVal&, Ptree*, TokenVal&, TokenVal&, Ptree*,
		   TokenVal&);
    PtreeId WhatIs();
    static BOOL RegisterKeyword(char* keyword, char* class_name);
    static Ptree* TranslateClass(Environment*, Ptree*, Ptree*, Class*&);

protected:
    Ptree* WalkMemberDeclaration(Environment*, PtreeArray&, PtreeArray&);
    Ptree* WalkMemberDeclaration2(Environment*);
    Ptree* NthMemberName(int&);
    BOOL IsMember(Ptree*, int&);

private:
    static void RecordThisVariable(Environment*, Ptree*);
    static void RecordAllMemberTypes(Environment*, Class*);
    static void RecordMemberTypes(Environment*, Class*);

    void WalkToplevel(Environment*);
    static Class* MakeMetaobject(Environment*, Ptree*, Ptree*, Ptree*);
    static Class* MakeMetaobject2(Environment*, Ptree*, Ptree*, Ptree*, BOOL&);
    static Class* LookupDeclarations(Environment*, Ptree*, Ptree*, BOOL&);
    static Class* LookupKeyword(Environment*, Ptree*, Ptree*, Ptree*, BOOL&);
    static Class* LookupBaseMetaclass(Environment*, Ptree*, Ptree*, BOOL&);

    static HashTable* keywords;

friend class PtreeTemplateDecl;
};

class PtreeTypedef : public NonLeaf {
public:
    PtreeTypedef(TokenVal&, Ptree*, Ptree*, TokenVal&);
    PtreeId WhatIs();

protected:
    void WalkToplevel(Environment*);
    Ptree* WalkMemberDeclaration(Environment*, PtreeArray&, PtreeArray&);
    Ptree* NthMemberName(int&);
    BOOL IsMember(Ptree*, int&);
};

class PtreeLinkage : public NonLeaf {
public:
    PtreeLinkage(TokenVal& t1, TokenVal& t2, Ptree* p3);
    PtreeId WhatIs();

protected:
    void WalkToplevel(Environment*);
};

class PtreeUserKeyword : public NonLeaf {
public:
    PtreeUserKeyword(TokenVal& t1, Ptree* p2);
    PtreeUserKeyword(TokenVal& t1, Ptree* p2, TokenVal& t3);
    PtreeId WhatIs();

protected:
    Ptree* WalkMemberDeclaration(Environment*, PtreeArray&, PtreeArray&);
    Ptree* NthMemberName(int&);
    BOOL IsMember(Ptree*, int&);
};

class PtreeAssignExpr : public NonLeaf {
public:
    PtreeAssignExpr(Ptree*, Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeInfixExpr : public NonLeaf {
public:
    PtreeInfixExpr(TokenVal& t1, TokenVal& t2, TokenVal& t3);
    PtreeInfixExpr(Ptree*, Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeCondExpr : public NonLeaf {	/* <expr> ? <expr> : <expr> */
public:
    PtreeCondExpr(TokenVal& t1, TokenVal& t2, TokenVal& t3, TokenVal& t4,
		  TokenVal& t5);
    PtreeCondExpr(Ptree*, Ptree*, Ptree*, Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeCast : public NonLeaf {
public:
    PtreeCast(TokenVal& t1, TokenVal& t2, TokenVal& t3, TokenVal& t4);
    PtreeCast(Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeUnaryExpr : public NonLeaf {
public:
    PtreeUnaryExpr(TokenVal& t1, TokenVal& t2);
    PtreeUnaryExpr(Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreePostfixExpr : public NonLeaf {
public:
    PtreePostfixExpr(Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
    BOOL IsMemberCall(Environment*, Ptree*, TypeInfo&, Class*&);
    Ptree* WalkFuncall(Environment*, Ptree*, Ptree*, TypeInfo&);
};

class PtreeMemberAccessExpr : public NonLeaf {
public:
    PtreeMemberAccessExpr(Ptree*, TokenVal&, Ptree*);
    PtreeMemberAccessExpr(Ptree*, Ptree*, Ptree*);
    PtreeMemberAccessExpr(Ptree*, Ptree*, Ptree*, Ptree*);
    PtreeMemberAccessExpr(Ptree*, Ptree*, Ptree*, Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreePtrToMemExpr : public NonLeaf {
public:
    PtreePtrToMemExpr(Ptree*, TokenVal&, Ptree*);
    PtreePtrToMemExpr(Ptree*, Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeSizeofExpr : public NonLeaf {
public:
    PtreeSizeofExpr(Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeSizeofType : public NonLeaf {
public:
    PtreeSizeofType(Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class LeafConstValue : public Leaf {
public:
    LeafConstValue(TokenVal&);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeFstyleCast : public NonLeaf {
public:
    PtreeFstyleCast(Ptree*, TokenVal&, TokenVal&, TokenVal&);
    PtreeFstyleCast(Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeUserStatement : public NonLeaf {
public:
    PtreeUserStatement(TokenVal&, TokenVal&, TokenVal&);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeStaticUserStatement : public NonLeaf {
public:
    PtreeStaticUserStatement(TokenVal&, TokenVal&);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeExprStatement : public NonLeaf {
public:
    PtreeExprStatement(TokenVal&, TokenVal&);
    PtreeId WhatIs();

protected:
    Ptree* WalkStatement(Environment*);
};

class PtreeLabel : public NonLeaf {
public:
    PtreeLabel(Ptree*, Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    Ptree* WalkStatement(Environment*);
};

class PtreeCaseLabel : public NonLeaf {
public:
    PtreeCaseLabel(Ptree*, Ptree*, Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    Ptree* WalkStatement(Environment*);
};

class PtreeIfStatement : public NonLeaf {
public:
    PtreeIfStatement(TokenVal&);
    PtreeId WhatIs();
    void Print(int, int);
    int Write(ostream&, int);

protected:
    Ptree* WalkStatement(Environment*);
};

class PtreeSwitchStatement : public NonLeaf {
public:
    PtreeSwitchStatement(TokenVal&);
    PtreeId WhatIs();

protected:
    Ptree* WalkStatement(Environment*);
};

class PtreeWhileStatement : public NonLeaf {
public:
    PtreeWhileStatement(TokenVal&);
    PtreeId WhatIs();

protected:
    Ptree* WalkStatement(Environment*);
};

class PtreeDoStatement : public NonLeaf {
public:
    PtreeDoStatement(TokenVal&);
    PtreeId WhatIs();

protected:
    Ptree* WalkStatement(Environment*);
};

class PtreeForStatement : public NonLeaf {
public:
    PtreeForStatement(TokenVal&);
    PtreeId WhatIs();

protected:
    Ptree* WalkStatement(Environment*);
};

class PtreeBreakStatement : public NonLeaf {
public:
    PtreeBreakStatement(TokenVal&);
    PtreeId WhatIs();

protected:
    Ptree* WalkStatement(Environment*);
};

class PtreeContinueStatement : public NonLeaf {
public:
    PtreeContinueStatement(TokenVal&);
    PtreeId WhatIs();

protected:
    Ptree* WalkStatement(Environment*);
};

class PtreeReturnStatement : public NonLeaf {
public:
    PtreeReturnStatement(TokenVal&);
    PtreeId WhatIs();

protected:
    Ptree* WalkStatement(Environment*);
};

class PtreeGotoStatement : public NonLeaf {
public:
    PtreeGotoStatement(TokenVal&);
    PtreeId WhatIs();

protected:
    Ptree* WalkStatement(Environment*);
};

class PtreeNew : public NonLeaf {
public:
    PtreeNew(Ptree*, Ptree*, Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeDelete : public NonLeaf {
public:
    PtreeDelete(Ptree*, Ptree*, Ptree*, Ptree*);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

class PtreeActualArgs : public NonLeaf {
public:
    PtreeActualArgs(Ptree*);
    PtreeId WhatIs();

protected:
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

/*
  This internal node is used to mark a changed subtree.
  Token::MinimumReplace() always substitutes the subtree if it includes
  unchanged nodes.
*/
class PtreeChanged : public NonLeaf {
public:
    PtreeChanged(Ptree*);
    PtreeId WhatIs();

protected:
    void WalkToplevel(Environment*);
    Ptree* WalkStatement(Environment*);
    Ptree* WalkExpr(Environment*, TypeInfo&);
    void Type(Environment*, TypeInfo&);
};

//  PtreeNonExpr is used by Ptree::MarkStatement().

class PtreeNonExpr : public PtreeChanged {
public:
    PtreeNonExpr(Ptree*);
    PtreeId WhatIs();
    BOOL IsA(PtreeId);
};

#endif  /* _ptree_h */
