/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/

#include "ptree.h"

BOOL Ptree::Eq(Ptree* p, char c)
{
    if(p == nil)
	return FALSE;
    else
	return p->Eq(c);
}

BOOL Ptree::Eq(Ptree* p, char* str)
{
    if(p == nil)
	return FALSE;
    else
	return p->Eq(str);
}

BOOL Ptree::Eq(Ptree* p, Ptree* q)
{
    if(p == q)
	return TRUE;
    else if(p == nil || q == nil)
	return FALSE;
    else if(p->IsLeaf() && q->IsLeaf()){
	int plen = p->GetLength();
	int qlen = q->GetLength();
	if(plen == qlen){
	    char* pstr = p->GetPosition();
	    char* qstr = q->GetPosition();
	    while(--plen >= 0)
		if(pstr[plen] != qstr[plen])
		    return FALSE;

	    return TRUE;
	}
    }

    return FALSE;
}

Ptree* Ptree::Last(Ptree* p)	// return the last cons cell.
{
    Ptree* next;
    if(p == nil)
	return nil;

    while((next = p->Cdr()) != nil)
	p = next;

    return p;
}

Ptree* Ptree::First(Ptree* p)
{
    if(p != nil)
	return p->Car();
    else
	return p;
}

Ptree* Ptree::Rest(Ptree* p)
{
    if(p != nil)
	return p->Cdr();
    else
	return p;
}

Ptree* Ptree::Second(Ptree* p)
{
    if(p != nil){
	p = p->Cdr();
	if(p != nil)
	    return p->Car();
    }

    return p;
}

Ptree* Ptree::Third(Ptree* p)
{
    if(p != nil){
	p = p->Cdr();
	if(p != nil){
	    p = p->Cdr();
	    if(p != nil)
		return p->Car();
	}
    }

    return p;
}

/*
  Nth(lst, 0) is equivalent to First(lst).
*/
Ptree* Ptree::Nth(Ptree* p, int n)
{
    while(p != nil && n-- > 0)
	p = p->Cdr();

    if(p != nil)
	return p->Car();
    else
	return p;
}

/*
  Length() returns a negative number if p is not a list.
*/
int Ptree::Length(Ptree* p)
{
    int i = 0;

    if(p != nil && p->IsLeaf())
	return -2;	/* p is not a pair */

    while(p != nil){
	++i;
	if(p->IsLeaf())
	    return -1;	/* p is a pair, but not a list. */
	else
	    p = p->Cdr();
    }

    return i;
}

Ptree* Ptree::ListTail(Ptree* p, int k)
{
    while(p != nil && k-- > 0)
	p = p->Cdr();

    return p;
}

Ptree* Ptree::Cons(Ptree* p, Ptree* q)
{
    return new NonLeaf(p, q);
}

Ptree* Ptree::List(Ptree* p)
{
    return new NonLeaf(p, nil);
}

Ptree* Ptree::List()
{
    return nil;
}

Ptree* Ptree::List(Ptree* p, Ptree* q)
{
    return new NonLeaf(p, new NonLeaf(q, nil));
}

Ptree* Ptree::List(Ptree* p1, Ptree* p2, Ptree* p3)
{
    return new NonLeaf(p1, new NonLeaf(p2, new NonLeaf(p3, nil)));
}

Ptree* Ptree::List(Ptree* p1, Ptree* p2, Ptree* p3, Ptree* p4)
{
    return new NonLeaf(p1, List(p2, p3, p4));
}

Ptree* Ptree::List(Ptree* p1, Ptree* p2, Ptree* p3, Ptree* p4, Ptree* p5)
{
    return Nconc(List(p1, p2), List(p3, p4, p5));
}

Ptree* Ptree::List(Ptree* p1, Ptree* p2, Ptree* p3, Ptree* p4, Ptree* p5,
		   Ptree* p6)
{
    return Nconc(List(p1, p2, p3), List(p4, p5, p6));
}

Ptree* Ptree::List(Ptree* p1, Ptree* p2, Ptree* p3, Ptree* p4, Ptree* p5,
		   Ptree* p6, Ptree* p7)
{
    return Nconc(List(p1, p2, p3), List(p4, p5, p6, p7));
}

Ptree* Ptree::CopyList(Ptree* p)
{
    return Append(p, nil);
}

//   q may be a leaf
//
Ptree* Ptree::Append(Ptree* p, Ptree* q)
{
    Ptree *result, *tail;

    if(p == nil)
	if(q->IsLeaf())
	    return Cons(q, nil);
	else
	    return q;

    result = tail = Cons(p->Car(), nil);
    p = p->Cdr();
    while(p != nil){
	Ptree* cell = Cons(p->Car(), nil);
	tail->SetCdr(cell);
	tail = cell;
	p = p->Cdr();
    }

    if(q != nil && q->IsLeaf())
	tail->SetCdr(Cons(q, nil));
    else
	tail->SetCdr(q);

    return result;
}

Ptree* Ptree::Substitute(Ptree* list, int pos, Ptree* e, Ptree* orig)
{
    Ptree *result, *tail, *cell;

    if(list == nil || e == orig)
	return list;

    if(pos <= 1)
	return Cons(e, list->Cdr());

    result = tail = Cons(list->Car(), nil);
    list = list->Cdr();
    while(list != nil && --pos > 1){
	cell = Cons(list->Car(), nil);
	tail->SetCdr(cell);
	tail = cell;
	list = list->Cdr();
    }

    if(list != nil)
	tail->SetCdr(Cons(e, list->Cdr()));

    return result;
}

Ptree* Ptree::Substitute(Ptree* list, int pos1, Ptree* e1, Ptree* orig1,
			 int pos2, Ptree* e2, Ptree* orig2)
{
    Ptree *result, *tail, *cell;

    if(list == nil || (e1 == orig1 && e2 == orig2))
	return list;

    if(pos1 <= 1)
	return Cons(e1, Substitute(list->Cdr(), pos2 - pos1, e2, orig2));

    result = tail = Cons(list->Car(), nil);
    list = list->Cdr();
    int pos = pos1;
    while(list != nil && --pos > 1){
	cell = Cons(list->Car(), nil);
	tail->SetCdr(cell);
	tail = cell;
	list = list->Cdr();
    }

    if(list != nil)
	tail->SetCdr(Cons(e1, Substitute(list->Cdr(),
					 pos2 - pos1, e2, orig2)));

    return result;
}

Ptree* Ptree::Substitute(Ptree* list, int pos1, Ptree* e1, Ptree* orig1,
			 int pos2, Ptree* e2, Ptree* orig2,
			 int pos3, Ptree* e3, Ptree* orig3)
{
    Ptree *result, *tail, *cell;

    if(list == nil || (e1 == orig1 && e2 == orig2 && e3 == orig3))
	return list;

    if(pos1 <= 1)
	return Cons(e1, Substitute(list->Cdr(), pos2 - pos1, e2, orig2,
				   pos3 - pos1, e3, orig3));

    result = tail = Cons(list->Car(), nil);
    list = list->Cdr();
    int pos = pos1;
    while(list != nil && --pos > 1){
	cell = Cons(list->Car(), nil);
	tail->SetCdr(cell);
	tail = cell;
	list = list->Cdr();
    }

    if(list != nil)
	tail->SetCdr(Cons(e1, Substitute(list->Cdr(), pos2 - pos1, e2, orig2,
					 pos3 - pos1, e3, orig3)));

    return result;
}

Ptree* Ptree::Substitute(Ptree* list, int pos1, Ptree* e1, Ptree* orig1,
			 int pos2, Ptree* e2, Ptree* orig2,
			 int pos3, Ptree* e3, Ptree* orig3,
			 int pos4, Ptree* e4, Ptree* orig4)
{
    Ptree *result, *tail, *cell;

    if(list == nil
       || (e1 == orig1 && e2 == orig2 && e3 == orig3 && e4 == orig4))
	return list;

    if(pos1 <= 1)
	return Cons(e1, Substitute(list->Cdr(), pos2 - pos1, e2, orig2,
				   pos3 - pos1, e3, orig3,
				   pos4 - pos1, e4, orig4));

    result = tail = Cons(list->Car(), nil);
    list = list->Cdr();
    int pos = pos1;
    while(list != nil && --pos > 1){
	cell = Cons(list->Car(), nil);
	tail->SetCdr(cell);
	tail = cell;
	list = list->Cdr();
    }

    if(list != nil)
	tail->SetCdr(Cons(e1, Substitute(list->Cdr(), pos2 - pos1, e2, orig2,
					 pos3 - pos1, e3, orig3,
					 pos4 - pos1, e4, orig4)));

    return result;
}

/*
  ReplaceAll() substitutes SUBST for all occurences of ORIG in LIST.
  It recursively searches LIST for ORIG.
*/
Ptree* Ptree::ReplaceAll(Ptree* list, Ptree* orig, Ptree* subst)
{
    if(Eq(list, orig))
	return subst;
    else if(list == nil || list->IsLeaf())
	return list;
    else{
	PtreeArray newlist;
	BOOL changed = FALSE;
	Ptree* rest = list;
	while(rest != nil){
	    Ptree* p = rest->Car();
	    Ptree* q = ReplaceAll(p, orig, subst);
	    newlist.Append(q);
	    if(p != q)
		changed = TRUE;

	    rest = rest->Cdr();
	}

	if(changed)
	    return newlist.All();
	else
	    return list;
    }
}

//  Replace() makes a new list (.. new_element ..) from (.. orig_element ..).
//
Ptree* Ptree::Replace(Ptree* list, Ptree* orig_element, Ptree* new_element)
{
    Ptree *result, *tail, *cell;

    if(list == nil)
	return list;

    cell = list->Car();
    if(cell == orig_element)
	return Cons(new_element, list->Cdr());

    result = tail = Cons(cell, nil);
    list = list->Cdr();
    while(list != nil){
	cell = list->Car();
	if(cell == orig_element){
	    tail->SetCdr(Cons(new_element, list->Cdr()));
	    return result;
	}

	cell = Cons(cell, nil);
	tail->SetCdr(cell);
	tail = cell;
	list = list->Cdr();
    }

    return result;
}

Ptree* Ptree::ReplaceTail(Ptree* lst, Ptree* original, Ptree* subst)
{
    Ptree *result, *tail;

    if(lst == original)
	return subst;

    result = tail = Cons(lst->Car(), nil);
    lst = lst->Cdr();
    while(lst != nil){
	if(lst == original){
	    tail->SetCdr(subst);
	    break;
	}

	Ptree* cell = Cons(lst->Car(), nil);
	tail->SetCdr(cell);
	tail = cell;
	lst = lst->Cdr();
    }

    return result;
}

Ptree* Ptree::Map(Environment* env, Ptree* expr,
		  Ptree* (*fn)(Environment*, Ptree*))
{
    const N = 32;	// must be an even number
    Ptree* e2[N];
    Ptree *tail, *expr2;
    int tail_pos, i;

    tail = expr;
    tail_pos = 0;
    for(i = 0; expr != nil && i < N;){
	Ptree* e = expr->Car();
	e2[i] = (*fn)(env, e);
	expr = expr->Cdr();
	if(e != e2[i++]){
	    tail_pos = i;
	    tail = expr;
	}
    }

    if(expr == nil){
	i = tail_pos;
	expr2 = tail;
    }
    else{
	expr2 = Map(env, expr, fn);
	if(expr == expr2){
	    i = tail_pos;
	    expr2 = tail;
	}
    }

    for(--i; i >= 0; --i)
	expr2 = Cons(e2[i], expr2);

    return expr2;
}

Ptree* Ptree::Map2(Environment* env, Ptree* expr,
		   Ptree* (*fn)(Environment*, Ptree*))
{
    const N = 32;	// must be an even number
    Ptree* e2[N];
    Ptree *tail, *expr2;
    int tail_pos, i;

    tail = expr;
    tail_pos = 0;
    for(i = 0; expr != nil && i < N;){
	Ptree* e = expr->Car();
	e2[i] = (*fn)(env, e);
	expr = expr->Cdr();
	if(e != e2[i++]){
	    tail_pos = i;
	    tail = expr;
	}

	if(expr != nil){
	    e2[i++] = expr->Car();		// skip a separator
	    expr = expr->Cdr();
	}
    }

    if(expr == nil){
	i = tail_pos;
	expr2 = tail;
    }
    else{
	expr2 = Map2(env, expr, fn);
	if(expr == expr2){
	    i = tail_pos;
	    expr2 = tail;
	}
    }

    for(--i; i >= 0; --i)
	expr2 = Cons(e2[i], expr2);

    return expr2;
}

Ptree* Ptree::Snoc(Ptree* p, Ptree* q)
{
    return Nconc(p, Cons(q, nil));
}

/* Nconc is desctructive append */

Ptree* Ptree::Nconc(Ptree* p, Ptree* q)
{
    if(p == nil)
	return q;
    else{
	Last(p)->data.nonleaf.next = q;	// Last(p)->SetCdr(q);
	return p;
    }
}

Ptree* Ptree::Nconc(Ptree* p, Ptree* q, Ptree* r)
{
    Nconc(p, Nconc(q, r));
}

static struct { PtreeId id; char* name; } PtreeIdTable[] = {
    { NonLeafId, "NonLeafId" },
    { LeafId, "LeafId" },

    { LeafTypeNameId, "LeafTypeNameId" },
    { LeafClassNameId, "LeafClassNameId" },
    { PtreeQClassNameId, "PtreeQClassNameId" },

    { PtreeDeclarationId, "PtreeDeclarationId" },
    { PtreeFunctionId, "PtreeFunctionId" },
    { PtreeTemplateDeclId, "PtreeTemplateDeclId" },
    { PtreeMetaclassId, "PtreeMetaclassId" },
    { PtreeTypedefId, "PtreeTypedefId" },
    { PtreeLinkageId, "PtreeLinkageId" },

    { PtreeExprStatementId, "PtreeExprStatementId" },
    { PtreeLabelId, "PtreeLabelId" },
    { PtreeCaseLabelId, "PtreeCaseLabelId" },
    { PtreeIfStatementId, "PtreeIfStatementId" },
    { PtreeSwitchStatementId, "PtreeSwitchStatementId" },
    { PtreeWhileStatementId, "PtreeWhileStatementId" },

    { PtreeDoStatementId, "PtreeDoStatementId" },
    { PtreeForStatementId, "PtreeForStatementId" },
    { PtreeBreakStatementId, "PtreeBreakStatementId" },
    { PtreeContinueStatementId, "PtreeContinueStatementId" },
    { PtreeReturnStatementId, "PtreeReturnStatementId" },
    { PtreeGotoStatementId, "PtreeGotoStatementId" },

    { PtreeInfixExprId, "PtreeInfixExprId" },
    { PtreeCondExprId, "PtreeCondExprId" },
    { PtreeCastId, "PtreeCastId" },
    { PtreeUnaryExprId, "PtreeUnaryExprId" },

    { PtreePostfixExprId, "PtreePostfixExprId" },
    { PtreeMemberAccessExprId, "PtreeMemberAccessExprId" },
    { PtreeSizeofExprId, "PtreeSizeofExprId" },
    { PtreeAssignExprId, "PtreeAssignExprId" },
    { PtreeSizeofTypeId, "PtreeSizeofTypeId" },
    { PtreePtrToMemExprId, "PtreePtrToMemExprId" },
    { PtreeNewId, "PtreeNewId" },
    { PtreeDeleteId, "PtreeDeleteId" },
    { PtreeFstyleCastId, "PtreeFstyleCastId" },
    { PtreeUserStatementId, "PtreeUserStatementId" },
    { PtreeStaticUserStatementId, "PtreeStaticUserStatementId" },

    { PtreeEnumId, "PtreeEnumId" },
    { PtreeClassId, "PtreeClassId" },

    { PtreeParenId, "PtreeParenId" },
    { PtreeBracketId, "PtreeBracketId" },
    { PtreeBraceId, "PtreeBraceId" },
    { PtreeAngleId, "PtreeAngleId" },

    { LeafPointerId, "LeafPointerId" },
    { LeafReferenceId, "LeafReferenceId" },
    { LeafConstValueId, "LeafConstValueId" },
    { LeafNameId, "LeafNameId" },
    { LeafReservedId, "LeafReservedId" },
    { LeafThisId, "LeafThisId" },
    { LeafReservedModifierId, "LeafReservedModifierId" },
    { PtreePtrToMemberId, "PtreePtrToMemberId" },
    { PtreeCvQualifierId, "PtreeCvQualifierId" },
    { PtreeQnameId, "PtreeQnameId" },
    { PtreeDestructorId, "PtreeDestructorId" },
    { PtreeOperatorFuncId, "PtreeOperatorFuncId" },
    { PtreeTemplateNameId, "PtreeTemplateNameId" },
    { PtreeAccessCtrlId, "PtreeAccessCtrlId" },
    { PtreeUserKeywordId, "PtreeUserKeywordId" },
    { PtreeActualArgsId, "PtreeActualArgsId" },

    { PtreeChangedId, "PtreeChangedId" },
    { PtreeNonExprId, "PtreeNonExprId" },
};

char* Ptree::IdName(PtreeId id)
{
    int n = sizeof(PtreeIdTable) / sizeof(PtreeIdTable[0]);

    while(--n >= 0)
	if(PtreeIdTable[n].id == id)
	    return PtreeIdTable[n].name;

    return nil;
}
