/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/

#include "syntax.h"
#include "token.h"
#include "ptree.h"
#include "binding.h"
#include "typeinfo.h"
#include "function.h"

Ptree* Syntax::Constructor(Ptree* tspec, TokenVal& lp, Ptree* args,
			   TokenVal& rp, Ptree* colon,
			   Ptree* init, Ptree* body)
{
    Ptree* name;
    Ptree* p = tspec;
    if(tspec->Cdr() == nil){
	name = tspec;
	tspec = nil;
    }
    else{
	for(; p->Cddr() != nil; p = p->Cdr())
	    ;

	name = p->Cdr();
	p->SetCdr(nil);
    }

    PtreeIter next(tspec);
    while(p = next())
	if(p->WhatIs() != LeafReservedModifierId){
	    ErrorMessage("invalid constructor definition", tspec, nil);
	    break;
	}

    PtreeId whatis = name->Car()->WhatIs();
    if(whatis != LeafClassNameId && whatis != PtreeQClassNameId)
	ErrorMessage("invalid constructor name", name->Car(), nil);

    if(whatis == PtreeQClassNameId)
	name->SetCar(new PtreeQname(name->Car()));

    Ptree::Nconc(name, Ptree::List(new Leaf(lp), args, new Leaf(rp)));
    if(body == nil)
	return new PtreeFunction(tspec, name, colon);
    else
	return new PtreeFunction(tspec, name, Ptree::Cons(colon, init), body);
}

Ptree* Syntax::QualifiedConstructor(Ptree* head, TokenVal& lp, Ptree* args,
				    TokenVal& rp, Ptree* colon,
				    Ptree* init, Ptree* body)
{
    Ptree::Nconc(head->Cadr(),
		 Ptree::List(new Leaf(lp), args, new Leaf(rp)));
    if(body == nil)
	Ptree::Snoc(head, colon);
    else{
	Ptree::Snoc(head, Ptree::Cons(colon, init));
	Ptree::Snoc(head, body);
    }

    return head;
}

Ptree* Syntax::QualifiedConstructorHead(Ptree* spec, Ptree* name)
{
    Ptree* p;
    PtreeIter next(spec);

    while(p = next())
	if(p->WhatIs() != LeafReservedModifierId){
	    ErrorMessage("invalid constructor definition", spec, nil);
	    break;
	}

    Ptree* car = name->Car();
    Ptree* cdr = name->Cdr();
    ((NonLeaf*)name)->SetNil();
    delete name;
    return new PtreeFunction(spec, Ptree::List(new PtreeQname(car, cdr)));
}

void Syntax::Destructor(Ptree* p)
{
    if(!p->IsA(LeafClassNameId))
	ErrorMessage("invalid destructor name", p, nil);
}

Ptree* Syntax::OldFuncDeclaration(Ptree* decl, Ptree* body)
{
    BOOL is_func;

    DeclaredName(decl, is_func);
    if(is_func)
	return new PtreeFunction(NIL, decl, body);

    ErrorMessage("parse error");
    return new PtreeDeclaration(NIL, NIL, body);
}

Ptree* Syntax::FuncDeclaration(Ptree* typespec, Ptree* decl, Ptree* body)
{
    BOOL is_func;

    DeclaredName(decl, is_func);
    if(is_func)
	   return new PtreeFunction(typespec, decl, body);

    ErrorMessage("parse error");
    return new PtreeDeclaration(NIL, NIL, body);
}

Ptree* Syntax::Declaration(Ptree* typespec, Ptree* decl_list, Ptree* delimiter)
{
    BOOL is_func;

    if(Ptree::Length(decl_list) == 1){
	DeclaredName(Ptree::First(decl_list), is_func);
	if(is_func){
	    Ptree* declarator = Ptree::First(decl_list);
	    ((NonLeaf*)decl_list)->SetNil();
	    delete decl_list;
	    return new PtreeFunction(typespec, declarator, delimiter);
	}
    }

    return new PtreeDeclaration(typespec, decl_list, delimiter);
}

Ptree* Syntax::CastOperator(Ptree* p)
{
    if(Ptree::Length(p) != 1)
	return p;
    else{
	Ptree* q = p->Car();
	((NonLeaf*)p)->SetNil();
	delete p;
	return q;
    }
}

void Syntax::IsNotQualifiedClassName(Ptree* p)
{
    if(!p->IsA(LeafClassNameId))
	ErrorMessage("invalid class name", p, nil);
}

/*
   DeclaredName() looks up a name or a qualified name in the
   given declarator.  If the declarator is "**(*f)(int)", then
   the resulting value is "f".  If no name is found, the resulting
   value is nil.
*/
Ptree* Syntax::DeclaredName(Ptree* p)
{
    BOOL is_func;

    return DeclaredName(p, is_func);
}

Ptree* Syntax::DeclaredName(Ptree* p, BOOL& is_func)
{
    is_func = FALSE;
    while(p != nil){
	Ptree* q = p->Car();
	switch(q->WhatIs()){
	case PtreeParenId :
	    return DeclaredName(q->Cadr(), is_func);
	case LeafNameId :
	case LeafClassNameId :
	case PtreeDestructorId :
	case PtreeOperatorFuncId :
	case PtreeQnameId :
	    p = p->Cdr();
	    if(p != nil && p->Car()->Eq('(')){
	        p = p->Cadr();
		if(p == nil || p->WhatIs() != PtreeActualArgsId)
		    is_func = TRUE;
	    }

	    return q;
	default :
	    p = p->Cdr();
	    break;
	}
    }

    return nil;
}

Ptree* Syntax::TypedefName(Ptree* p)
{
    if(p != nil && (p->IsA(LeafTypeNameId) || p->IsA(LeafClassNameId)))
	return p;
    else{
	p = DeclaredName(p);
	if(p != nil && p->IsLeaf())
	    return p;
    }

    ErrorMessage("invalid typedef name", p, nil);
    return nil;
}

Ptree* Syntax::DeclaredTemplateName(Ptree* p)
{
    Ptree* decl;
    Ptree* name;

    if(p->IsA(PtreeDeclarationId) && Ptree::Match(p, "[[%?] ;]", &decl)){
	if(decl->IsA(PtreeClassId)
	   && Ptree::Match(decl, "[%* %? %_]", &name))
	    if(name != nil)
		return name;
    }
    else if(p->IsA(PtreeFunctionId))
	if(Ptree::Match(p, "[%* %* ;]"))
	    //
	    // prototype of a template function
	    // e.g. template <class T> T foo(T);
	    // recognize, but don't call Token::RecordTemplate().
	    //
	    return nil;
	else if(Ptree::Match(p, "[%* %? %* [{ %* }]]", &decl)){
	    //
	    // constructor with member initializers
	    // e.g. template <class T> X::X(T t) : Y(t) {}
	    //
	    name = DeclaredName(decl);
	    if(name != nil)
		return name;
	}
	else if(Ptree::Match(p, "[%* %? [{ %* }]]", &decl)){
	    name = DeclaredName(decl);
	    if(name != nil)
		return name;
	}

    ErrorMessage("invalid template declaration", nil, nil);
    return nil;
}

Ptree* Syntax::MakeAssignExpr(Ptree* p1, TokenVal& t2, Ptree* p3)
{
    PtreeId whatis = p1->WhatIs();
    if(whatis == PtreeMemberAccessExprId && Ptree::Length(p1) == 3)
	return Ptree::Nconc(p1, Ptree::List(new Leaf(t2), p3));
    else
	return new PtreeAssignExpr(p1, new Leaf(t2), p3);
}

Ptree* Syntax::MakeFuncallExpr(Ptree* p1, TokenVal& t2, Ptree* p3,
			       TokenVal& t4)
{
    Ptree* args = new PtreeParen(t2, p3, t4);
    PtreeId whatis = p1->WhatIs();
    if(whatis == PtreeMemberAccessExprId && Ptree::Length(p1) == 3)
	return Ptree::Snoc(p1, args);
    else
	return new PtreePostfixExpr(p1, args);
}

Ptree* Syntax::TranslateTempFuncall(TypeInfo& t, Environment* env,
				    Ptree* funcall)
{
    TemplateFunctionInfo* info = (TemplateFunctionInfo*)t.GetBinding();
    return info->GetTempFuncMetaobject()->TranslateFuncall(info->GetEnv(),
							   env, funcall);
}

TypeInfoId Syntax::IsMemberName(Ptree* qname, Environment* env, TypeInfo& t,
				Class*& c)
{
    static char this_name[] = "this";
    TypeInfoId whatis;
    Binding* b;
    Binding* b2;
    BOOL is_type;

    c = nil;
    if(qname->IsLeaf()){
	for(; env != nil; env = env->next)
	    if(env->LookupTop(qname->GetPosition(), qname->GetLength(), b))
		break;

	if(env == nil){
	    t.Undef();
	    return UndefType;
	}

	SetTypeInfo(b, env, is_type, t);
	whatis = b->WhatIs();
	if(whatis == DataMemberName || whatis == MemberFunctionName){
	    b2 = env->MostEffectiveClass();
	    c = b2->GetMetaobject();
	}

	return whatis;
    }
    else if(qname->WhatIs() == PtreeQnameId){
	b = BindQualifiedName(env, qname, &b2);
	if(b != nil){
	    SetTypeInfo(b, env, is_type, t);
	    if(b2 != nil)
		c = b2->GetMetaobject();

	    return b->WhatIs();
	}
    }

    // if PtreeDestructor or PtreeOperatorFunc
    t.Undef();
    return UndefType;
}

// TYPE specifies the class that has the member.

TypeInfoId Syntax::LookupMemberType(Ptree* qname, TypeInfo& type,
				    TypeInfo& member_type)
{
    Binding* b;
    BOOL is_type;

    if(!qname->IsLeaf()){
	if(qname->WhatIs() == PtreeQnameId)
	    qname = Ptree::First(Ptree::Last(qname));
	else{
	    member_type.Undef();
	    return UndefType;
	}
    }

    Environment* env = type.ClassEnvironment();
    if(env != nil && ((b = env->LookupAll(qname)) != nil)){
	SetTypeInfo(b, env, is_type, member_type);
	return b->WhatIs();
    }
    else{
	member_type.Undef();
	return UndefType;
    }
}

BOOL Syntax::LookupType(Ptree* qname, Environment* env,
			BOOL& is_type_name, TypeInfo& t)
{
    Binding* b;
    PtreeId whatis;
    Ptree* name;

    whatis = qname->WhatIs();
    if(whatis == PtreeTemplateNameId || qname->IsLeaf()){
	if(whatis == PtreeTemplateNameId)
	    qname = qname->Car();

	if((b = env->LookupAll(qname)) != nil){
	    SetTypeInfo(b, env, is_type_name, t);
	    return TRUE;
	}
	else{
	    t.Undef();
	    return FALSE;
	}
    }
    else if(whatis == PtreeQnameId || whatis == PtreeQClassNameId){
	b = BindQualifiedName(env, qname);
	if(b != nil){
	    SetTypeInfo(b, env, is_type_name, t);
	    return TRUE;
	}
	else{
	    t.Undef();
	    return FALSE;
	}
    }
    else{	// this function is called only by Environment::Lookup()
	MopErrorMessage("Environment::Lookup()", "the type of NAME is wrong");
	t.Undef();
	return FALSE;
    }
}

void Syntax::SetTypeInfo(Binding* b, Environment* env,
			 BOOL& is_type_name, TypeInfo& t)
{
    TypeInfoId whatis;
    switch(whatis = b->WhatIs()){
    case VariableName :
    case DataMemberName :
    case ImmutableVarName :
    case FunctionName :
    case MemberFunctionName :
	AnalyzeVariableType((NameInfo*)b, env, t);
	is_type_name = FALSE;
	break;
    case TypedefType :
	AnalyzeVariableType((NameInfo*)b, env, t);
	is_type_name = TRUE;
	break;
    case ThisVariable :
	t.Set(whatis, ((ThisVariableInfo*)b)->ClassInfo(), env);
	is_type_name = FALSE;
	break;
    case ClassType :
    case EnumType :
    case TemplateType :
	t.Set(whatis, b, env);
	is_type_name = TRUE;
	break;
    case TemplateFunctionType :
	t.Set(whatis, b, env);
	is_type_name = FALSE;
	break;
    case TypeVariable :
	t.Set(((TypeVarInfo*)b)->GetTypeInfo());
	is_type_name = TRUE;
	break;
    default :
	MopErrorMessage("Syntax::SetTypeInfo()", "bomb!");
    }
}

void Syntax::SetPseudoClassTypeInfo(Environment* env, TypeInfo& t)
{
    t.Set(ClassType, nil, env);
}

void Syntax::AnalyzeConstType(Ptree*, Environment* env, TypeInfo& t)
{
    t.Set(BuiltInType, nil, env);
}

// for cast, sizeof, and new expressions.
//
void Syntax::AnalyzeCastType(Ptree* cast_type,
			     Environment* env, TypeInfo& t)
{
    AnalyzeVariableType(new ImmutableVarNameInfo(cast_type->Car(),
						 cast_type->Cdr()),
			cast_type->Cdr(), env, t);
}

void Syntax::AnalyzeFstyleCast(Ptree* cast_type,
			       Environment* env, TypeInfo& t)
{
    AnalyzeVariableType(new ImmutableVarNameInfo(cast_type, nil),
			NIL, env, t);
}

void Syntax::AnalyzeVariableType(NameInfo* name, Environment* env,
				 TypeInfo& t)
{
    AnalyzeVariableType(name, name->declarator, env, t);
}

void Syntax::AnalyzeVariableType(NameInfo* name, Ptree* declarator,
				 Environment* env, TypeInfo& t)
{
    PtreeId whatis;
    Ptree* full_type_name = nil;

    TypeInfoId derived = UndefType;
    while(declarator != nil){
	Ptree* head = declarator->Car();
	switch(whatis = head->WhatIs()){
	case PtreeParenId :
	    AnalyzeVariableType(name, ((PtreeBlock*)head)->Body(), env, t);
	    return;
	case LeafNameId :
	case LeafClassNameId :
	case PtreeDestructorId :
	case PtreeOperatorFuncId :
	case PtreeQnameId :
	    declarator = declarator->Cdr();
	    if(declarator != nil){
		head = declarator->Car();
		if(head->Eq('(')){
		    head = declarator->Cadr();
		    if(head == nil || head->WhatIs() != PtreeActualArgsId){
			t.Set(FunctionType, name, env);
			return;
		    }
		}
		else if(head->Eq('[')){
		    t.Set(ArrayType, name, env);
		    return;
		}
		else if(!head->Eq('=') && !head->Eq(':')) {
		    ErrorMessage("invalid type declarator",
				 name->declarator, nil);
		    t.Undef();
		    return;
		}
	    }

	    if(derived == UndefType)
		AnalyzeTypespecifier(name, env, t);
	    else
		t.Set(derived, name, env);

	    return;
	case LeafPointerId :
	    derived = PointerType;
	    goto do_repeat;
	case LeafReferenceId :
	    derived = ReferenceType;
	    goto do_repeat;
	case PtreePtrToMemberId :
	    derived = PointerToMemberType;
	    full_type_name = head;
	    goto do_repeat;
	default :
	    if(head->Eq('(')){
		head = declarator->Cadr();
		if(head == nil || head->WhatIs() != PtreeActualArgsId){
		    t.Set(FunctionType, name, env);
		    return;
		}
		else{	// initializer  e.g. Point x(1, 3);
		    declarator = nil;
		    break;
		}
	    }
	    else if(head->Eq('[')){
		t.Set(ArrayType, name, env);
		return;
	    }
	    else if(head->Eq('=') || head->Eq(':')){
		declarator = nil;
		break;
	    }
	do_repeat :
	    declarator = declarator->Cdr();
	    break;
	}
    }

    // only arg.declarator, cast.declarator, etc may reach here.
    if(derived == UndefType)
	AnalyzeTypespecifier(name, env, t);
    else{
	t.Set(derived, name, env);
	if(derived == PointerToMemberType)
	    t.SetFullTypeName(full_type_name);
    }
}

void Syntax::AnalyzeTypespecifier(Binding* name, Environment* env,
				  TypeInfo& t)
{
    PtreeId whatis;
    Ptree* p;

    for(p = name->typespec; p != nil; p = p->Cdr()){
	whatis = p->Car()->WhatIs();
	if(whatis != LeafReservedModifierId && whatis != PtreeUserKeywordId)
	    break;
    }

    if(p == nil)
	t.Undef();
    else
	switch(whatis){
	case LeafTypeNameId :
	case LeafClassNameId :
	    ResolveTypedefName(p->Car(), env, t);
	    break;
	case PtreeQClassNameId :
	    AnalyzeQClass(p->Car(), env, t);
	    break;
	case LeafReservedId :
	    t.Set(BuiltInType, name, env);
	    t.SetFullTypeName(p->Car());
	    break;
	case PtreeClassId :
	    // We cannot deal with no-name classes.
	    ResolveTypedefName(((PtreeClass*)(p->Car()))->Name(), env, t);
	    break;
	case PtreeEnumId :
	    ResolveTypedefName(((PtreeEnum*)(p->Car()))->Name(), env, t);
	    break;
	case PtreeTemplateNameId :
	    ResolveTypedefName(p->Car()->Car(), env, t);
	    break;
	default :
	    MopErrorMessage("Syntax::AnalyzeTypespecifier()", "bomb!");
	    break;
	}
}

void Syntax::ResolveTypedefName(Ptree* name, Environment* env, TypeInfo& t)
{
    Binding* b;

    if(name == nil){
	t.Undef();
	return;
    }

    if((b = env->LookupAll(name)) != nil){
	TypeInfoId whatis = b->WhatIs();
	if(whatis == TypedefType){
	    AnalyzeVariableType((NameInfo*)b, ((NameInfo*)b)->declarator,
				env, t);
	    return;
	}
	else if(whatis == ClassType || whatis == EnumType
		|| whatis == TemplateType){
	    t.Set(whatis, b, env);
	    t.SetFullTypeName(name);
	    return;
	}
	else if(whatis == TypeVariable){
	    t.Set(((TypeVarInfo*)b)->GetTypeInfo());
	    t.SetFullTypeName(name);
	    return;
	}
    }

    ErrorMessage("cannot resolve the type name", name, nil);
    t.Undef();
}

void Syntax::AnalyzeQClass(Ptree* qclass, Environment* env, TypeInfo& t)
{
    BOOL is_type;

    Binding* b = BindQualifiedName(env, qclass);
    if(b == nil)
	SetPseudoClassTypeInfo(env, t);
    else
	SetTypeInfo(b, env, is_type, t);

    t.SetFullTypeName(qclass);
}

static TypeInfoId whatIsInDereferencedType;
static Ptree* fullnameOfDereferencedType;

/*
  t becomes UndefType if type cannot be dereferenced any more.
  type and t may be the same object.
*/
void Syntax::DereferencedType(TypeInfo& type, TypeInfo& t)
{
    switch(type.WhatIs()){
    case PointerType :
    case ReferenceType :
    case PointerToMemberType :
    case ArrayType :
    case FunctionType :
      {
	int deref_level = type.DerefLevel() + 1;
	Binding* b = type.GetBinding();
	int depth = DereferencedType2(((NameInfo*)b)->declarator,
				       deref_level);
	if(depth == 0)
	    AnalyzeTypespecifier(b, type.Env(), t);
	else if(depth > 0)
	    t.Undef();
	else{
	    t.Set(whatIsInDereferencedType, b, type.Env(), deref_level);
	    if(whatIsInDereferencedType == PointerToMemberType)
		t.SetFullTypeName(fullnameOfDereferencedType);
	}
      }
	break;
    case TemplateFunctionType :
	t.Undef();
	break;
    default :
	t.Undef();
	break;
    }
}

int Syntax::DereferencedType2(Ptree* declarator, int depth)
{
    PtreeId whatis;

    if(declarator == nil)
	return depth;

    Ptree* head = declarator->Car();
    switch(whatis = head->WhatIs()){
    case LeafPointerId :
	depth = DereferencedType2(declarator->Cdr(), depth);
	if(depth-- == 0)
	    whatIsInDereferencedType = PointerType;

	return depth;
    case LeafReferenceId :
	depth = DereferencedType2(declarator->Cdr(), depth);
	if(depth-- == 0)
	    whatIsInDereferencedType = ReferenceType;

	return depth;
    case PtreePtrToMemberId :
	depth = DereferencedType2(declarator->Cdr(), depth);
	if(depth-- == 0){
	    whatIsInDereferencedType = PointerToMemberType;
	    fullnameOfDereferencedType = head;
	}

	return depth;
    case PtreeParenId :
	depth = DereferencedType2(((PtreeBlock*)head)->Body(), depth);
	/* not break here */
    case LeafNameId :
    case LeafClassNameId :
    case PtreeDestructorId :
    case PtreeOperatorFuncId :
    case PtreeQnameId :
	declarator = declarator->Cdr();
	/* not break here */
    case LeafId :
	while(declarator != nil){
	    head = declarator->Car();
	    if(head->Eq('(')){
		head = declarator->Cadr();
		if(head == nil || head->WhatIs() != PtreeActualArgsId){
		    if(depth-- == 0)
			whatIsInDereferencedType = FunctionType;
		}
		break;
	    }
	    else if(head->Eq('[')){
		if(depth-- == 0)
		    whatIsInDereferencedType = ArrayType;

		declarator = declarator->Cddr()->Cdr();
	    }
	    /* if the rest of the expr is an initializer, a bit field,
	       or member initializers */
//	    else if(head->Eq('=') || head->Eq(':'))
//		return depth;	/* this if is redundant. */
	    else
		break;
	}

	return depth;
    default :
	return DereferencedType2(declarator->Cdr(), depth);
    }
}

/*
  This returns nil if the qname is the symbol name or the qualified
  class is not found.
*/
ClassTypeInfo* Syntax::LookupClassOfQname(Ptree* qname, Environment* env)
{
    Binding* b;

    if(qname->WhatIs() == PtreeQnameId){
	BindQualifiedName(env, qname, &b);
	if(b != nil && b->WhatIs() == ClassType)
	    return (ClassTypeInfo*)b;
    }

    return nil;		// error
}

/*
  AnalyzeNthArgType() returns FALSE if the type is not a function type
  or the NTH argument doesn't exist.  To analyze the first argument,
  give this function 1 as NTH.  If the argument is ellipses, the type
  is UndefType.
*/
BOOL Syntax::AnalyzeNthArgType(TypeInfo& type, int nth, Environment* env,
			       TypeInfo& argtype)
{
    if(nth < 1)
	nth = 1;

    Ptree* args = GetArgList(type);
    if(args != nil){
	args = Ptree::Nth(args->Cadr(), nth * 2 - 2);
	if(args != nil)
	    if(args->IsLeaf()){		// ellipses ...
		argtype.Undef();
		return TRUE;
	    }
	    else{
		AnalyzeVariableType(new ImmutableVarNameInfo(args->Car(),
							     args->Cadr()),
				    args->Cadr(), env, argtype);
		return TRUE;
	    }
    }

    argtype.Undef();
    return FALSE;
}

static Ptree* foundArgList;

Ptree* Syntax::GetArgList(TypeInfo& type)
{
    if(type.WhatIs() != FunctionType)
	return nil;	// error
    else{
	Binding* b = type.GetBinding();
	foundArgList = nil;
	GetArgList2(((NameInfo*)b)->declarator, type.DerefLevel());
	return foundArgList;
    }
}

int Syntax::GetArgList2(Ptree* declarator, int depth)
{
    PtreeId whatis;

    if(declarator == nil)
	return depth;

    Ptree* head = declarator->Car();
    switch(whatis = head->WhatIs()){
    case LeafPointerId :
    case LeafReferenceId :
    case PtreePtrToMemberId :
	depth = GetArgList2(declarator->Cdr(), depth) - 1;
	return depth;
    case PtreeParenId :
	depth = GetArgList2(((PtreeBlock*)head)->Body(), depth);
	/* not break here */
    case LeafNameId :
    case LeafClassNameId :
    case PtreeDestructorId :
    case PtreeOperatorFuncId :
    case PtreeQnameId :
	declarator = declarator->Cdr();
	/* not break here */
    case LeafId :
	while(declarator != nil){
	    head = declarator->Car();
	    if(head->Eq('(')){
		head = declarator->Cadr();
		if(head == nil || head->WhatIs() != PtreeActualArgsId){
		    if(depth-- == 0)
			foundArgList = declarator;	// found!
		}
		break;
	    }
	    else if(head->Eq('[')){
		--depth;
		declarator = declarator->Cddr()->Cdr();
	    }
	    /* if the rest of the expr is an initializer, a bit field,
	       or member initializers */
//	    else if(head->Eq('=') || head->Eq(':'))
//		return depth;	/* this if is redundant. */
	    else
		break;
	}

	return depth;
    default :
	return GetArgList2(declarator->Cdr(), depth);
    }
}

/*
  BindQualifiedName() returns the Binding of the rightmost name.
  It sets classinfo to the most local class included in the given name.
  For example, if the name is Rect::Point::x, classinfo is set to Point.
*/
Binding* Syntax::BindQualifiedName(Environment* env, Ptree* qname,
				   Binding** classinfo)
{
    Binding* b;
    PtreeId whatis = qname->WhatIs();
    Ptree* last;

    if(classinfo != nil)
	*classinfo = nil;

    if(whatis != PtreeQnameId && whatis != PtreeQClassNameId)
	return nil;
    else{
	last = Ptree::Last(qname)->Car();
	PtreeIter next(qname);
	Ptree* p = next();
	if(p->Eq("::")){
	    p = next();		// skip ::
	    env = env->GetBottom();
	}

	while(p != nil){
	    char* name;
	    int len;
	    if(p->WhatIs() == PtreeTemplateNameId){
		name = p->Car()->GetPosition();
		len = p->Car()->GetLength();
	    }
	    else if(p->IsLeaf()){
		name = p->GetPosition();
		len = p->GetLength();
	    }
	    else
		return nil;

	    if(env != nil && env->LookupAll(name, len, b))
		switch(b->WhatIs()){
		case ClassType :
		case TemplateType :
		    env = b->GetEnv();
		    break;
		case TypeVariable :
		    b = ((TypeVarInfo*)b)->GetTypeInfo().GetBinding();
		    env = b->GetEnv();
		    break;
		default :
		    break;
		}
	    else
		return nil;

	    next();		// skip ::
	    p = next();
	    if(p == last && classinfo != nil)
		*classinfo = b;
	}

	return b;
    }
}
