/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/

#include <stdlib.h>
#include <string.h>
#include "buffer.h"
#include "token.h"
#include "hash.h"
#include "ptree.h"
#include "binding.h"
#include "syntax.h"

#ifdef DEBUG

#include <stdio.h>
#ifdef __GNUG__
#define token(x)	(int)#x
#else
#define token(x)	(int)"x"
#endif

main()
{
    Token::SetSource(new ProgramFromStdin);
    for(;;){
	int t = yylex();
	if(t < 128){
	    if(t == 0)
		break;

	    printf("%c (%x): ", t, t);
	}
	else
	    printf("%-10.10s (%x): ", (char*)t, t);

	printf("\"%s\"\n",
	   Token::GetIdentifier(Token::Tokenp(), Token::TokenLen()));
    }
}

#else

#define token(x)	x

#endif /* DEBUG */

#define YYSTYPE		TokenVal

#include "y.tab.h"

extern TokenVal yylval;

#define is_blank(c)	((c) == ' ' || (c) == '\t' || (c) == '\f' \
			 || (c) == '\r')
#define is_letter(c)	('A' <= (c) && (c) <= 'Z' || 'a' <= (c) && (c) <= 'z'\
			 || (c) == '_' || (c) == '$')
#define is_digit(c)	('0' <= (c) && (c) <= '9')
#define is_xletter(c)	((c) == 'X' || (c) == 'x')
#define is_eletter(c)	((c) == 'E' || (c) == 'e')
#define is_hexdigit(c)	(is_digit(c) || 'A' <= (c) && (c) <= 'F'\
			 || 'a' <= (c) && (c) <= 'f')
#define is_int_suffix(c)	((c) == 'U' || (c) == 'u' || (c) == 'L'\
				 || (c) == 'l')
#define is_float_suffix(c)	((c) == 'F' || (c) == 'f' || (c) == 'L'\
				 || (c) == 'l')

#define UNDEF		' '

Program* Token::file = nil;
int Token::last_token = '\n';
int Token::next_token = UNDEF;
int Token::marked_position = 0;
int Token::marked_line = 0;
uint Token::tokenp = 0;
int Token::token_len = 0;
Binding* Token::last_identifier_type = nil;
int Token::line_number = 0;
char* Token::file_name = nil;
Environment* Token::type_env = nil;
HashTable* Token::user_keywords = nil;
BOOL Token::ignore_line_directive = FALSE;

int yylex()
{
    Ptree* p;
    BOOL is_constructor;

    int t = Token::GetToken();
    yylval.leaf.position = Token::TokenPosition();
    yylval.leaf.length = Token::TokenLen();
    yylval.leaf.token = t;

    /*
       If the token is ClassName, we look ahead next tokens.
       Qualified class names and constructor names make a single token.
       Note: MarkPosition() takes care of this fact.
    */
    if(t == ClassName || t == TemplateName){
	yylval.ptree = Token::ReadClassName(is_constructor);
	if(is_constructor)
	    return QConstructorName;
	else if(yylval.ptree->WhatIs() == PtreeQnameId)
	    return QName;
	else
	    return ClassName;
    }
#ifdef __GNUG__
    else if(t == ATTRIBUTE){	// if the token is __attribute__
	Token::SkipAttributeToken();
	return yylex();
    }
#endif
    else
	return t;
}

/*
   SetSource() should not be called more than once.
*/
void Token::SetSource(Program* prog, char* name)
{
    file = prog;
    file_name = DuplicateString(name);
    line_number = 1;
    prog->Rewind();
    last_token = '\n';
    next_token = UNDEF;
    type_env = new Environment;		// the global name space
    if(user_keywords == nil)
	user_keywords = new HashTable;

#ifdef __GNUG__
    // We below typedef some predefined types in GNU g++.

    static char* builtin_types[] = { "__wchar_t", "wchar_t", "bool" };
    for(int i = 0;
	i < sizeof(builtin_types) / sizeof(builtin_types[0]);
	++i)
    {
	RecordType(builtin_types[i], strlen(builtin_types[i]),
		   new ClassTypeInfo(nil, nil));
    }
#endif
}

Environment* Token::GetEnvironment()
{
    return type_env;
}

char* Token::TokenPosition()
{
    return (char*)file->Read(Tokenp());
}

char* Token::File()
{
    if(file_name != nil)
	return file_name;
    else
	return "unknown";
}

char Token::Ref(uint i)
{
    return file->Ref(i);
}

BOOL Token::RecordKeyword(char* str, int token)
{
    int index;

    if(user_keywords == nil)
	user_keywords = new HashTable;

    if(user_keywords->AddEntry(str, (HashValue)token, &index) >= 0)
	return TRUE;
    else
	return BOOL(user_keywords->Peek(index) == (HashValue)token);
}

//   SkipAttributeToken() skips __attribute__(...).

void Token::SkipAttributeToken()
{
    char c;

    do{
	c = file->Get();
    }while(c != '(');

    int i = 1;
    do{
	c = file->Get();
	if(c == '(')
	    ++i;
	else if(c == ')')
	    --i;
    } while(i > 0);
}

/*
   --- Scope control ---

   To distinguish TypeName and ClassName from Identifier, the following
   functions have to be appropriately called.
*/

void Token::EnterNewScope()
{
    type_env = new Environment(type_env);
}

/*
  ExitScope() exits the current scope and deletes the current environment.
  If you want to keep that environment, call ExitClassBody().
*/
void Token::ExitScope()
{
    Environment* e = type_env;
    type_env = type_env->UnlinkTop();
    delete e;
}

/*
  ExitClasBody() exits the current scope and returns the current environment
  if exists.  Deallocating the returned environment is responsibility of
  the caller.  Note that the returned environment may have a parent
  environment.  If you simply deallocate the returned environment, the parent
  one is deallocated together.
*/
Environment* Token::ExitClassBody()
{
    Environment* e = type_env;
    if(e->IsEmpty()){
	type_env = e->UnlinkTop();
	delete e;
	return nil;
    }
    else{
	type_env = e->Next();
	return e;
    }
}

/*
   EnterFunctionScope() returns the previous environment.
*/
Environment* Token::EnterFunctionScope(Ptree* declarator)
{
    Ptree* qname = Syntax::DeclaredName(declarator);
    if(qname != nil && !qname->IsA(PtreeQnameId))
	return type_env;
    else
	return EnterNestedClassScope(qname);
}

/*
   EnterNestedClassScope() returns the previous environment.
*/
Environment* Token::EnterNestedClassScope(Ptree* qname)
{
    Binding* type;

    int len = Ptree::Length(qname);
    Environment* env = type_env;
    Syntax::BindQualifiedName(env, qname, &type);
    if(type == nil)
	return type_env;
    else{
	Environment* previous_env = type_env;
	type_env = type->GetEnv();
	return previous_env;
    }
}

void Token::ExitFunctionScope(Environment* previous_env)
{
    ExitNestedClassScope(previous_env);
}

void Token::ExitNestedClassScope(Environment* previous_env)
{
    type_env = previous_env;
}


/* --- Record class names, enum names, typedef names,
   and template names to the hash table. --- */

void Token::RecordClass(Ptree* name, Ptree* definition, Environment* local_env)
{
    Binding* type;

    if(!name->IsLeaf())
	MopErrorMessage("RecordClass()", "bomb!");

    if(type_env->LookupTop(name->GetPosition(), name->GetLength(), type))
	if(type->WhatIs() != ClassType)
	    ErrorMessage("invalid class name", nil, nil);
	else{
	    if(local_env != nil)
		local_env->RecordClassInfo(type);

	    ((ClassTypeInfo*)type)->UpdateInfo(definition, local_env);
	}
    else{
	type = new ClassTypeInfo(definition, local_env);
	if(local_env != nil)
	    local_env->RecordClassInfo(type);

	RecordType(name->GetPosition(), name->GetLength(), type);
    }
}

void Token::RecordEnum(TokenVal& t, Ptree* definition)
{
    RecordType(t.leaf.position, t.leaf.length, new EnumTypeInfo(definition));
}

void Token::RecordTypedef(Ptree* definition, Ptree* declarator)
{
    Ptree* name = Syntax::TypedefName(declarator);
    if(name == nil)
	return;
    else
	RecordType(name->GetPosition(), name->GetLength(),
		   new TypedefTypeInfo(definition->Cadr(), declarator));
}

void Token::RecordTemplate(Ptree* name, Ptree* definition, Environment* env)
{
    Binding *b, *type;

    if(!name->IsLeaf())
	return;		// template operator functions etc are ignored.

    if(Ptree::Nth(definition, 4)->IsA(PtreeFunctionId))
	/* if this is a constructor of a template class, it
	   should not be recorded because it overrides the template
	   class in the record.  Recall that the two have the same name. */
	type = new TemplateFunctionInfo(definition, env);
    else{
	/* a class template makes two environments.  One is made when the
	   class declration is parsed, while the other is made when the
	   outer template declaration is parsed.  The parent of the former
	   environment is the latter.  So we record the former one. */

	b = env->LookupAll(name);
	if(b != nil)
	    env = b->GetEnv();	// get the env for the class declration.

	type = new TemplateInfo(definition, env);
	env->RecordClassInfo(type);
    }

    RecordType(name->GetPosition(), name->GetLength(), type);
}

void Token::RecordType(char* position, int len, Binding* type)
{
    if(type_env->AddEntry(position, len, type) < 0)
	delete type;
}

/* --- Read qualified class names --- */

Ptree* Token::ReadClassName(BOOL& is_constructor)
{
    is_constructor = FALSE;
    Ptree* p = ReadQualifiedClass(nil, nil);
    int n = Ptree::Length(p);
    if(n == 1){
	Ptree* leaf = p->Car();
	((NonLeaf*)p)->SetNil();
	delete p;
	return leaf;	// a LeafClassName object (single class name)
    }
    else if(n >= 3 && Ptree::Eq(Ptree::Nth(p, n - 3), Ptree::Nth(p, n - 1))){
	is_constructor = TRUE;
	return p;	// a PtreeQClassName object (constructor)
    }
    else
	return p;	// a PtreeQClassName object (qualified class)
			// or a PtreeQname object (qualified name)
}

/*
  ReadQualifiedClass() returns:
    - a list (<class1> :: <class2> ... <class-n> :: )
      if the next sequence of tokens is:
	  <class1> :: <class2> ... <class-n> :: <name>
      The next token is <name>.

    - a list ( <class1> :: )
      if the next sequence is:
	  <class1> :: <name>

    - a list ( <class> :: <class> )
      if the next sequence is:
	  <class> :: <class>

    - a list ( <class1> :: ... :: <type name> )
      if the next sequence is:
          <class1> :: ... :: <type name>.
*/
Ptree* Token::ReadQualifiedClass(Ptree* class_list, Environment* env)
{
    Binding* type;
    uint class_name;
    int  class_name_len;
    int  kind;

    if(env == nil){
	env = last_identifier_type->GetEnv();
	if(env == nil)
	    env = type_env;
    }

    class_name = Tokenp();
    class_name_len = TokenLen();
    Ptree* p = new LeafClassName((char*)file->Read(class_name),
				 class_name_len);
    p = CheckTemplateName(p);
    if(class_list == nil)
	class_list = new PtreeQClassName(p);
    else
	class_list = Ptree::Snoc(class_list, p);

    type = NextIsClassName(env, class_name, class_name_len, kind);

    if(kind == 0)
	return class_list;

    GetToken();		// read a Scope
    class_list = Ptree::Snoc(class_list, new Leaf(TokenPosition(),
						  TokenLen()));
    if(kind == 1){	// qualified class name
	GetToken();
	return ReadQualifiedClass(class_list,
				  (type != nil ? type->GetEnv() : env));
    }
    else if(kind == 2){	// constructor name
	GetToken();
	return Ptree::Snoc(class_list,
			   new LeafClassName(TokenPosition(),
					     TokenLen()));
    }
    else if(kind == 3)	// qualified name
	return new PtreeQname(class_list);
    else if(kind == 4){	// qualified type
	GetToken();	// read a TypeName
	return Ptree::Snoc(class_list,
			   new LeafTypeName(TokenPosition(),
					    TokenLen()));
    }
    else{		// if(kind == 5)
	GetToken();
	return ReadQualifiedClass(class_list, env);
    }
}

/*
  NextIsClassName() changes KIND to:
    1 if the next tokens are ":: <class>",
    2 if                     ":: <constructor name>",
    3 if                     ":: <identifier>",
    4 if                     ":: <type name>",
    5 if		     ":: <class?> <identifier?>" or ":: <class?> ::",
    0 otherwise.
*/
Binding* Token::NextIsClassName(Environment* env, uint class_name,
				int class_name_len, int& kind)
{
    char c;
    Binding* type = nil;
    kind = 0;
    uint pos = file->GetNextPos();

    do{
	c = file->Get();
    } while(is_blank(c) || c == '\n');

    if(c == ':' && file->Get() == ':'){
	kind = 3;		// qualified name?
	c = file->Get();
	while(is_blank(c) || c == '\n')
	    c = file->Get();

	uint top = file->GetCurPos();
	if(is_letter(c)){
	    do{
		c = file->Get();
	    }while(is_letter(c) || is_digit(c));

	    int len = file->GetCurPos() - top;
	    if(env != nil && env->LookupAll(file->Read(top), len, type)){
		int t = type->GetToken();
		if(t == ClassName || t == TemplateName)
		    kind = 1;	// qualified class
		else if(t == TypeName)
		    kind = 4;	// qualified type name
	    }
	    else if(len == class_name_len
		    && strncmp((char*)file->Read(top),
			       (char*)file->Read(class_name), len) == 0){
		type = nil;
		kind = 2;	// constructor name
	    }
	    else if(ProbablyTypeName((char*)file->Read(top), len, c)){
		type = nil;
		kind = 5;	// qualified type name?
	    }
	}
	else if(c == '*')	// <qualified_class>::*
	    kind = 0;
    }

    file->Rewind(pos);
    return type;
}

BOOL Token::ProbablyTypeName(char* name, int len, char c)
{
    static char rkeyword[] = "operator";

    if(len == sizeof(rkeyword) - 1 && strncmp(name, rkeyword, len) == 0)
	return FALSE;

    while(is_blank(c) || c == '\n')
	c = file->Get();

    if(is_letter(c))
	return TRUE;
    else if(c == ':' && file->Get() == ':')
	return TRUE;
    else
	return FALSE;
}

Ptree* Token::CheckTemplateName(Ptree* name)
{
    char c;
    int n;
    char *lpos, *rpos;
    uint pos = file->GetNextPos();

    do{
	c = file->Get();
    } while(is_blank(c) || c == '\n');

    if(c == '<'){
	lpos = (char*)file->Read(file->GetCurPos());
	n = 1;
	do{
	    c = file->Get();
	    switch(c){
	    case '<' :
		++n;
		break;
	    case '>' :
		--n;
		break;
	    case ';' :
	    case '\0' :
		goto not_template;
	    default :
		break;
	    }
	}while(n > 0);

	rpos = (char*)file->Read(file->GetCurPos());
	name = new PtreeTemplateName(name, new Leaf(lpos, 1),
				     new Leaf(lpos + 1, rpos - lpos - 1),
				     new Leaf(rpos, 1));
	return name;
    }

not_template :
    file->Rewind(pos);
    return name;
}


/*
  Lexical Analyzer
*/

int Token::GetToken()
{
    int t;

    if(next_token != UNDEF){
	t = next_token;
	next_token = UNDEF;
	return t;
    }

    for(;;){
	t = ReadLine();
	last_token = t;
	if(t == '\n')
	    ++line_number;
	else
	    break;
    }

    return t;
}

void Token::MarkPosition(int lookahead, BOOL next_token)
{
    Ptree* head;

    if(next_token)	// the look-ahead buffer is empty.
	marked_position = file->GetNextPos();
    else{
	switch(lookahead){
	case QConstructorName :
	case QName :
	case ClassName :
	    head = yylval.ptree;
	    while(!head->IsLeaf())
		head = head->Car();

	    marked_position = head->GetPosition() - (char*)file->Read(0);
	    break;
	default :
	    marked_position = Tokenp();
	    break;
	}
    }

    marked_line = line_number;
}

void Token::BackTrack(int token)
{
    last_token = '\n';
    next_token = token;
    line_number = marked_line;
    file->Rewind(marked_position);
}

int Token::ReadLine()
{
    char c;
    uint top;

    do{
	c = file->Get();
    } while(is_blank(c));

    tokenp = top = file->GetCurPos();
    if(c == '\0'){
	file->Unget();
	return '\0';
    }
    else if(c == '\n')
	return '\n';
    else if(c == '#' && last_token == '\n'){
	if(ReadLineDirective())
	    return '\n';
	else{
	    file->Rewind(top + 1);
	    token_len = 1;
	    return SingleCharOp(c);
	}
    }
    else if(c == '\'' || c == '"'){
	if(c == '\''){
	    if(ReadCharConst(top))
		return token(CharConst);
	}
	else{
	    if(ReadStrConst(top))
		return token(StringL);
	}

	file->Rewind(top + 1);
	token_len = 1;
	return SingleCharOp(c);
    }
    else if(is_digit(c))
	return ReadNumber(c, top);
    else if(c == '.'){
	c = file->Get();
	if(is_digit(c))
	    return ReadFloat(top);
	else{
	    file->Unget();
	    return ReadSeparator('.', top);
	}
    }
    else if(is_letter(c))
	return ReadIdentifier(top);
    else
	return ReadSeparator(c, top);
}

BOOL Token::ReadCharConst(uint top)
{
    char c;

    for(;;){
	c = file->Get();
	if(c == '\\'){
	    c = file->Get();
	    if(c == '\0')
		return FALSE;
	}
	else if(c == '\''){
	    token_len = file->GetCurPos() - top + 1;
	    return TRUE;
	}
	else if(c == '\n' || c == '\0')
	    return FALSE;
    }
}

/*
  If text is a sequence of string constants like:
	"string1" "string2"
  then the string constants are delt with as a single constant.
*/
BOOL Token::ReadStrConst(uint top)
{
    char c;

    for(;;){
	c = file->Get();
	if(c == '\\'){
	    c = file->Get();
	    if(c == '\0')
		return FALSE;
	}
	else if(c == '"'){
	    uint pos = file->GetCurPos() + 1;
	    int nline = 0;
	    do{
		c = file->Get();
		if(c == '\n')
		    ++nline;
	    } while(is_blank(c) || c == '\n');

	    if(c == '"')
		line_number += nline;
	    else{
		token_len = pos - top;
		file->Rewind(pos);
		return TRUE;
	    }
	}
	else if(c == '\n' || c == '\0')
	    return FALSE;
    }
}

int Token::ReadNumber(char c, uint top)
{
    char c2 = file->Get();

    if(c == '0' && is_xletter(c2)){
	do{
	    c = file->Get();
	} while(is_hexdigit(c));
	while(is_int_suffix(c))
	    c = file->Get();

	file->Unget();
	token_len = file->GetCurPos() - top + 1;
	return token(Constant);
    }

    while(is_digit(c2))
	c2 = file->Get();

    if(is_int_suffix(c2))
	do{
	    c2 = file->Get();
	}while(is_int_suffix(c2));
    else if(c2 == '.')
	return ReadFloat(top);
    else if(is_eletter(c2)){
	file->Unget();
	return ReadFloat(top);
    }

    file->Unget();
    token_len = file->GetCurPos() - top + 1;
    return token(Constant);
}

int Token::ReadFloat(uint top)
{
    char c;

    do{
	c = file->Get();
    }while(is_digit(c));
    if(is_float_suffix(c))
	do{
	    c = file->Get();
	}while(is_float_suffix(c));
    else if(is_eletter(c)){
	uint p = file->GetCurPos();
	c = file->Get();
	if(c == '+' || c == '-'){
	     c = file->Get();
	     if(!is_digit(c)){
		file->Rewind(p);
		token_len = p - top;
		return token(Constant);
	    }
	}
	else if(!is_digit(c)){
	    file->Rewind(p);
	    token_len = p - top;
	    return token(Constant);
	}

	do{
	    c = file->Get();
	}while(is_digit(c));

	while(is_float_suffix(c))
	    c = file->Get();
    }

    file->Unget();
    token_len = file->GetCurPos() - top + 1;
    return token(Constant);
}

BOOL Token::ReadLineDirective()
{
    static char keyword[] = "pragma";
    char c;

    do{
	c = file->Get();
    }while(is_blank(c));

    if(c == 'p'){	/* #pragma */
	for(int i = 1; i < sizeof(keyword) - 1; ++i){
	    c = file->Get();
	    if(c != keyword[i])
		return FALSE;
	}

	c = file->Get();
	if(!is_blank(c) && c != '\n' && c != '\0')
	    return FALSE;

	do{
	    c = file->Get();
	}while(c != '\n' && c != '\0');
    }
    else if(is_digit(c)){	/* # <line> <file> */
	int num = c - '0';
	for(;;){
	    c = file->Get();
	    if(is_digit(c))
		num = num * 10 + c - '0';
	    else
		break;
	}

	if(!ignore_line_directive)
	    line_number = num - 1; /* line_number'll be incremented soon */

	if(is_blank(c)){
	    do{
		c = file->Get();
	    }while(is_blank(c));
	    uint fname = file->GetCurPos();
	    while(!is_blank(c) && c != '\n' && c != '\0')
		c = file->Get();

	    int len = file->GetCurPos() - fname;
	    if(len > 0 && !ignore_line_directive){
		delete file_name;
		file_name = GetIdentifier(fname, len);
	    }

	    while(c != '\n' && c != '\0')
		c = file->Get();
	}
	else if(c != '\n' && c != '\0')
	    return FALSE;
    }
    else
	return FALSE;

    return TRUE;
}

char* Token::GetIdentifier(uint position, int n)
{
    char* buf = new char[n + 1];
    bcopy((char*)file->Read(position), buf, n);
    buf[n] = '\0';
    return buf;
}

int Token::ReadIdentifier(uint top)
{
    char c;

    do{
	c = file->Get();
    }while(is_letter(c) || is_digit(c));

    uint len = file->GetCurPos() - top;
    token_len = len;
    file->Unget();

    return Screening((char*)file->Read(top), len);
}

/*
  This table is a list of reserved key words.
  Note: alphabetical order!
*/
static struct rw_table {
    char*	name;
    int		value;
} table[] = {
#ifdef __GNUG__
    "__attribute__", token(ATTRIBUTE),
    "__const",	token(CONST),
    "__signed",	token(SIGNED),
#endif
    "auto",	token(AUTO),
    "break",	token(BREAK),
    "case",	token(CASE),
    "char",	token(CHAR),
    "class",	token(CLASS),
    "const",	token(CONST),
    "continue",	token(CONTINUE),
    "default",	token(DEFAULT),
    "delete",	token(DELETE),
    "do",	token(DO),
    "double",	token(DOUBLE),
    "else",	token(ELSE),
    "enum",	token(ENUM),
    "extern",	token(EXTERN),
    "float",	token(FLOAT),
    "for",	token(FOR),
    "friend",	token(FRIEND),
    "goto",	token(GOTO),
    "if",	token(IF),
    "inline",	token(INLINE),
    "int",	token(INT),
    "long",	token(LONG),
    "metaclass",token(METACLASS),	// OpenC++
    "mutable",  token(MUTABLE),
    "new",	token(NEW),
    "operator",	token(OPERATOR),
    "private",	token(PRIVATE),
    "protected",token(PROTECTED),
    "public",	token(PUBLIC),
    "register", token(REGISTER),
    "return",	token(RETURN),
    "short",	token(SHORT),
    "signed",	token(SIGNED),
    "sizeof",	token(SIZEOF),
    "static",	token(STATIC),
    "struct",	token(STRUCT),
    "switch",	token(SWITCH),
    "template", token(TEMPLATE),
    "this",	token(THIS),
#ifdef __GNUG__
    // Under g++ 2.7 on Solaris, timezone is first defined as a type name,
    // then it is redefined as an extern variable (sys/time.h).  Because
    // the parser cannot process this redefinition, we makes timezone be
    // a reserved word.

    "timezone", token(Identifier),
#endif
    "typedef",	token(TYPEDEF),
    "union",	token(UNION),
    "unsigned", token(UNSIGNED),
    "virtual",	token(VIRTUAL),
    "void",	token(VOID),
    "volatile",	token(VOLATILE),
    "while",	token(WHILE),
};

int Token::Screening(char *identifier, int len)
{
    struct rw_table	*low, *high, *mid;
    int			c, token;

    low = table;
    high = &table[sizeof(table) / sizeof(table[0]) - 1];
    while(low <= high){
	mid = low + (high - low) / 2;
	if((c = strncmp(mid->name, identifier, len)) == 0)
	    if(mid->name[len] == '\0')
		return(mid->value);
	    else
		high = mid - 1;
	else if(c < 0)
	    low = mid + 1;
	else
	    high = mid - 1;
    }

    if(user_keywords->Lookup(identifier, len, (HashValue*)&token))
	return token;

    if(type_env->LookupAll(identifier, len, last_identifier_type))
	return last_identifier_type->GetToken();
    else
	return token(Identifier);
}

int Token::ReadSeparator(char c, uint top)
{
    char c1 = file->Get();

    token_len = 2;
    if(c1 == '='){
	switch(c){
	case '*' :
	case '/' :
	case '%' :
	case '+' :
	case '-' :
	case '&' :
	case '^' :
	case '|' :
	    return token(AssignOp);
	case '=' :
	case '!' :
	    return token(EqualOp);
	case '<' :
	case '>' :
	    return token(RelOp);
	default :
	    file->Unget();
	    token_len = 1;
	    return SingleCharOp(c);
	}
    }
    else if(c == c1){
	switch(c){
	case '<' :
	case '>' :
	    if(file->Get() != '='){
		file->Unget();
		return token(ShiftOp);
	    }
	    else{
		token_len = 3;
		return token(AssignOp);
	    }
	case '|' :
	    return token(LogOrOp);
	case '&' :
	    return token(LogAndOp);
	case '+' :
	case '-' :
	    return token(IncOp);
	case ':' :
	    return token(Scope);
	case '.' :
	    if(file->Get() == '.'){
		token_len = 3;
		return token(Ellipsis);
	    }
	    else
		file->Unget();
	default :
	    file->Unget();
	    token_len = 1;
	    return SingleCharOp(c);
	}
    }
    else if(c == '.' && c1 == '*')
	return token(PmOp);
    else if(c == '-' && c1 == '>')
	if(file->Get() == '*'){
	    token_len = 3;
	    return token(PmOp);
	}
	else{
	    file->Unget();
	    return token(ArrowOp);
	}
    else{
	file->Unget();
	token_len = 1;
	return SingleCharOp(c);
    }

    cerr << "*** An invaid character has been found! ("
	 << (int)c << ',' << (int)c1 << ")\n";
    return token(BadToken);
}

int Token::SingleCharOp(unsigned char c)
{
			/* !"#$%&'()*+,-./0123456789:;<=>? */
    static char valid[] = "x   xx xxxxxxxx          xxxxxx";

    if('!' <= c && c <= '?' && valid[c - '!'] == 'x')
	return c;
    else if(c == '[' || c == ']')
	return c;
    else if('{' <= c && c <= '~')
	return c;
    else
	return token(BadToken);
}

BOOL Token::Reify(Leaf* t, unsigned int& value)
{
    char* p = t->Position();
    int len = t->Length();
    value = 0;
    if(len > 2 && *p == '0' && is_xletter(p[1])){
	for(int i = 2; i < len; ++i){
	    char c = p[i];
	    if(is_digit(c))
		value = value * 0x10 + (c - '0');
	    else if('A' <= c && c <= 'F')
		value = value * 0x10 + (c - 'A' + 10);
	    else if('a' <= c && c <= 'f')
		value = value * 0x10 + (c - 'a' + 10);
	    else if(is_int_suffix(c))
		break;
	    else
		return FALSE;
	}

	return TRUE;
    }
    else if(len > 0 && is_digit(*p)){
	for(int i = 0; i < len; ++i){
	    char c = p[i];
	    if(is_digit(c))
		value = value * 10 + c - '0';
	    else if(is_int_suffix(c))
		break;
	    else
		return FALSE;
	}

	return TRUE;
    }
    else
	return FALSE;
}

// Reify() doesn't interpret an escape character.

BOOL Token::Reify(Leaf* t, char*& str)
{
    char* p = t->Position();
    int length = t->Length();
    if(*p != '"')
	return FALSE;
    else{
#ifdef DONT_GC
	str = new char[length];
#else
	str = (char*)GC_MALLOC_ATOMIC(length);
#endif
	char* sp = str;
	for(int i = 1; i < length; ++i)
	    if(p[i] != '"')
		*sp++ = p[i];
	    else
		while(++i < length && p[i] != '"')
		    ;

	*sp = '\0';
	return TRUE;
    }
}

/*

line directive:
^"#"{blank}*{digit}+({blank}+.*)?\n

pragma directive:
^"#"{blank}*"pragma".*\n

Constant	{digit}+{int_suffix}*
		"0"{xletter}{hexdigit}+{int_suffix}*
		{digit}*\.{digit}+{float_suffix}*
		{digit}+\.{float_suffix}*
		{digit}*\.{digit}+"e"("+"|"-")*{digit}+{float_suffix}*
		{digit}+\."e"("+"|"-")*{digit}+{float_suffix}*
		{digit}+"e"("+"|"-")*{digit}+{float_suffix}*

CharConst	\'([^'\n]|\\[^\n])\'

StringL		\"([^"\n]|\\["\n])*\"

Identifier
TypeName	{letter}+({letter}|{digit})*

ClassName	{typename}("::"{typename})*

AssignOp	*= /= %= += -= &= ^= <<= >>=

EqualOp		== !=

RelOp		<= >=

ShiftOp		<< >>

LogOrOp		||

LogAndOp	&&

IncOp		++ --

Scope		::

Ellipsis	...

PmOp		.* ->*

ArrowOp		->

others		!%^&*()-+={}|~[];:<>?,./

BadToken	others

*/
