/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/

#ifndef _token_h
#define _token_h

#include <iostream.h>
#include "types.h"

class Program;
class Ptree;
class Leaf;
class Environment;
class Binding;
class HashTable;

int yylex();

/* these two are defined in parser.y */
void ShowProgram(BOOL);
int NumOfErrors();
int yyparse();

/* semantic value used by the parser */

union TokenVal {
    /* for terminal symbols */
    struct {
	char* position;
	int length;
	int token;
    } leaf;
    /* for non-terminal symbols */
    Ptree* ptree;
    /* for class.body */
    struct {
	Ptree* ptree;
	Environment* env;
    } class_body;
    struct {
	Ptree* ptree;
	Environment* env;
    } func_decl;
};

/*
  class Token

  Before using this class, SetSource() must be called only once.

  This token analizer is a context-depend one.  The user must call
  EnterNewScope(), ExitScope(), RecordClass(), etc to give semantic
  information to the token analizer.

  TokenPosition() returns the location of the current token.
  The length of the token is given from TokenLen().  Note that a token
  is not a null-terminated string.

  GetIdentifier() is a convenient function to convert a token
  into a null-terminated string.  The user must be responsible to delete
  the resulting string.
*/
class Token {
public:
    static void SetSource(Program*, char* = nil);
    static Environment* GetEnvironment();
    static BOOL RecordKeyword(char*, int);

    static void EnterNewScope();
    static void ExitScope();
    static Environment* ExitClassBody();
    static Environment* EnterFunctionScope(Ptree*);
    static Environment* EnterNestedClassScope(Ptree*);
    static void ExitFunctionScope(Environment*);
    static void ExitNestedClassScope(Environment*);

    static void RecordClass(Ptree*, Ptree*, Environment*);
    static void RecordEnum(TokenVal&, Ptree*);
    static void RecordTypedef(Ptree*, Ptree*);
    static void RecordTemplate(Ptree*, Ptree*, Environment*);

    static int GetToken();
    static void MarkPosition(int, BOOL);
    static void BackTrack(int token);
    static void PushToken(int t) { next_token = t; }

    static char* TokenPosition();
    static uint Tokenp() { return tokenp; }
    static int TokenLen() { return token_len; }
    static char* GetIdentifier(uint, int);
    static int Line() { return line_number; }
    static char* File();
    static char Ref(uint i);

    static BOOL Reify(Leaf*, unsigned int&);
    static BOOL Reify(Leaf* t, char*&);

private:
    static void SkipAttributeToken();
    static void RecordType(char*, int, Binding*);
    static Ptree* ReadClassName(BOOL&);
    static Ptree* ReadQualifiedClass(Ptree*, Environment*);
    static Binding* NextIsClassName(Environment*, uint, int, int&);
    static BOOL ProbablyTypeName(char*, int, char);
    static Ptree* CheckTemplateName(Ptree*);
    static int ReadLine();
    static BOOL ReadCharConst(uint top);
    static BOOL ReadStrConst(uint top);
    static int ReadNumber(char c, uint top);
    static int ReadFloat(uint top);
    static BOOL ReadLineDirective();
    static int ReadIdentifier(uint top);
    static int Screening(char *identifier, int len);
    static int ReadSeparator(char c, uint top);
    static int SingleCharOp(unsigned char c);

// in output.cc
public:
    static void WriteText(ostream&, char*);
    static void ReplaceText(Ptree* oldtext, Ptree* newtext);
    static void InsertText(Ptree* oldtext, Ptree* before_text,
			   Ptree* after_text);
    static void MinimumReplace(Ptree* oldtext, Ptree* newtext);

private:
    static char* PtreeLeftMost(Ptree*);
    static char* PtreeRightMost(Ptree*);
    static BOOL MinimumReplace2(Ptree*, Ptree*);


public:
    static BOOL ignore_line_directive;	// ignore # <line> <file>

private:
    static Program* file;
    static int last_token;
    static int next_token;
    static int marked_position;
    static int marked_line;

    static uint tokenp;
    static int token_len;

    static Binding* last_identifier_type;

    static int line_number;
    static char* file_name;

    static Environment* type_env;
    static HashTable* user_keywords;

    friend int yylex();
};

#endif /* _token_h */
