/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/

#include "typeinfo.h"
#include "syntax.h"
#include "binding.h"
#include "ptree.h"

Ptree* TypeInfo::star = new Leaf("*", 1);
Ptree* TypeInfo::ampersand = new Leaf("&", 1);
Ptree* TypeInfo::brackets = new Leaf("[]", 2);
Ptree* TypeInfo::comma = new Leaf(",", 1);
Ptree* TypeInfo::dots = new Leaf("...", 3);
Ptree* TypeInfo::left_paren = new Leaf("(", 1);
Ptree* TypeInfo::right_paren = new Leaf(")", 1);

TypeInfo::TypeInfo(TypeInfoId k, Binding* b, Environment* e)
{
    deref_level = 0;
    ref_level = 0;
    kind = (unsigned short)k;
    env = e;
    full_typename = nil;
#ifdef DONT_GC
    if(b != nil && b->WhatIs() == ImmutableVarName)
	binding = new ImmutableVarNameInfo((ImmutableVarNameInfo*)b);
    else
	binding = b;
#else
    binding = b;
#endif
}

TypeInfo::~TypeInfo()
{
#ifdef DONT_GC
    if(binding != nil && binding->WhatIs() == ImmutableVarName)
	delete binding;
#endif
}

void TypeInfo::Set(TypeInfoId k, Binding* b, Environment* e, uint level)
{
    deref_level = level;
    ref_level = 0;
    kind = (unsigned short)k;
    env = e;
    full_typename = nil;
#ifdef DONT_GC
    if(b != nil && b->WhatIs() == ImmutableVarName)
	binding = new ImmutableVarNameInfo((ImmutableVarNameInfo*)b);
    else
	binding = b;
#else
    binding = b;
#endif
}

void TypeInfo::Set(TypeInfo& t)
{
    deref_level = t.deref_level;
    ref_level = t.ref_level;
    kind = t.kind;
    original_kind = t.original_kind;
    env = t.env;
    full_typename = t.full_typename;
#ifdef DONT_GC
    if(t.binding != nil && t.binding->WhatIs() == ImmutableVarName)
	binding = new ImmutableVarNameInfo((ImmutableVarNameInfo*)t.binding);
    else
	binding = t.binding;
#else
    binding = t.binding;
#endif
}

void TypeInfo::Reference()
{
    if(ref_level++ == 0)
	original_kind = kind;

    kind = PointerType;
}

void TypeInfo::Reference(TypeInfo& t)
{
    t.deref_level = deref_level;
    t.ref_level = ref_level + 1;
    if(t.ref_level == 0)
	t.original_kind = kind;
    else
	t.original_kind = original_kind;

    t.kind = PointerType;
    t.env = env;
    t.full_typename = full_typename;

#ifdef DONT_GC
    if(binding != nil && binding->WhatIs() == ImmutableVarName)
	t.binding = new ImmutableVarNameInfo((ImmutableVarNameInfo*)binding);
    else
	t.binding = binding;
#else
    t.binding = binding;
#endif
}

void TypeInfo::Dereference()
{
    if(ref_level > 0){
	if(--ref_level == 0)
	    kind = original_kind;
    }
    else
	Syntax::DereferencedType(*this, *this);
}

void TypeInfo::Dereference(TypeInfo& t)
{
    if(ref_level > 0){
	t.deref_level = deref_level;
	t.ref_level = ref_level - 1;
	if(t.ref_level == 0)
	    t.kind = original_kind;
	else{
	    t.kind = PointerType;
	    t.original_kind = original_kind;
	}

	t.env = env;
	t.full_typename = full_typename;
#ifdef DONT_GC
	if(binding != nil && binding->WhatIs() == ImmutableVarName)
	    t.binding = new ImmutableVarNameInfo((ImmutableVarNameInfo*)
						 binding);
	else
	    t.binding = binding;
#else
	t.binding = binding;
#endif
    }
    else
	Syntax::DereferencedType(*this, t);
}

void TypeInfo::DereferenceIfRefType()
{
    while(WhatIs() == ReferenceType)
	Dereference();
}

/*
  NthArgument() returns FALSE if the type is not a function type or
  the N-th argument doesn't exist.  If the argument is ellipses, the type
  is UndefType.  To analyze the first argument, give this function 0 as N.  
*/
BOOL TypeInfo::NthArgument(int n, TypeInfo& t)
{
    return Syntax::AnalyzeNthArgType(*this, n + 1, env, t);
}

Class* TypeInfo::ClassMetaobject()
{
    switch(WhatIs()){
    case ClassType :
    case TemplateType :
	if(GetBinding() == nil)
	    return nil;
	else
	    return GetBinding()->GetMetaobject();
    default :
	return nil;
    }
}

Environment* TypeInfo::ClassEnvironment()
{
    switch(WhatIs()){
    case ClassType :
    case TemplateType :
	if(GetBinding() == nil)
	    return nil;
	else
	    return GetBinding()->GetEnv();
    default :
	return nil;
    }
}

/*
  IsBuiltInType() returns a bitmap if the type is a built-in type.
  The bitmap indicates details of the type.  Otherwise it returns
  FALSE.
*/
uint TypeInfo::IsBuiltInType()
{
    uint bitmap = 0;

    if(WhatIs() != BuiltInType || binding == nil)
	return bitmap;

    Ptree* typespec = binding->typespec;
    while(typespec != nil){
	Ptree* p = Ptree::First(typespec);
	typespec = Ptree::Rest(typespec);
	if(p->Eq("char"))
	    bitmap |= CharType;
	else if(p->Eq("int"))
	    bitmap |= IntType;
	else if(p->Eq("short"))
	    bitmap |= ShortType;
#ifdef __GNUG__
	else if(p->Eq("__short"))
	    bitmap |= ShortType;
#endif
	else if(p->Eq("long"))
	    bitmap |= LongType;
	else if(p->Eq("signed"))
	    bitmap |= SignedType;
	else if(p->Eq("unsigned"))
	    bitmap |= UnsignedType;
	else if(p->Eq("float"))
	    bitmap |= FloatType;
	else if(p->Eq("double"))
	    bitmap |= DoubleType;
	else if(p->Eq("void"))
	    bitmap |= VoidType;
    }

    return bitmap;
}

// FullTypeName() returns the full names of qualified types and
// template names.

Ptree* TypeInfo::FullTypeName()
{
    switch(WhatIs()){
    case BuiltInType :
    case ClassType :
    case EnumType :
    case TemplateType :
    case PointerToMemberType :
	return full_typename;
    default :
	return nil;
    }
}

Ptree* TypeInfo::MakePtree(Ptree* variable_name)
{
    TypeInfo t;
    Ptree* op;

    switch(WhatIs()){
    case BuiltInType :
    case ClassType :
    case EnumType :
    case TemplateType :
    case PointerToMemberType :
	return Ptree::List(FullTypeName(), variable_name);
    case PointerType :
	op = star;
	goto jump_label;
    case ReferenceType :
	op = ampersand;
jump_label :
	Dereference(t);
	if(t.WhatIs() == FunctionType)
	    variable_name = Ptree::List(left_paren, op, variable_name,
					right_paren);
	else
	    variable_name = Ptree::List(op, variable_name);

	return t.MakePtree(variable_name);
    case ArrayType :
	Dereference(t);
	return t.MakePtree(Ptree::List(variable_name, brackets));
    case FunctionType :
	return MakeFunctionTypePtree(variable_name);
    case TemplateFunctionType :
	MopErrorMessage("TypeInfo::MakePtree()",
			"sorry, cannot handle a template function");
	return nil;
    case UndefType :
	MopErrorMessage("TypeInfo::MakePtree()",
			"sorry, the type is not specified (UndefType).");
	return nil;
    default :
	MopErrorMessage("TypeInfo::MakePtree()", "bomb!");
        return nil;
    }
}

Ptree* TypeInfo::MakeFunctionTypePtree(Ptree* name)
{
    TypeInfo argtype;
    Ptree* p = nil;
    int i = 0;
    while(NthArgument(i++, argtype)){
	if(p != nil)
	    p = Ptree::Snoc(p, comma);

	if(argtype.WhatIs() == UndefType){
	    p = Ptree::Snoc(p, dots);
	    break;
	}
	else
	    p = Ptree::Snoc(p, argtype.MakePtree(nil));
    }

    name = Ptree::List(name, left_paren, p, right_paren);
    Dereference(argtype);
    if(argtype.WhatIs() == UndefType)
	return name;
    else
	return argtype.MakePtree(name);
}
