/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/

#include <string.h>
#include "ptree.h"
#include "token.h"
#include "binding.h"
#include "typeinfo.h"
#include "syntax.h"
#include "hash.h"
#include "mop.h"
#include "metaclass.h"
#include "template.h"
#include "function.h"

Ptree* Ptree::program_tree = nil;
Ptree* Ptree::last_appended_subtree = nil;

Ptree* Ptree::before_toplevel = nil;
Ptree* Ptree::after_toplevel = nil;

HashTable* PtreeMetaclass::keywords = nil;

void Ptree::AppendToplevel(Ptree* toplevel)
{
#ifdef DONT_GC
    Ptree* top = List(toplevel);
    if(last_appended_subtree == nil)
	program_tree = last_appended_subtree = top;
    else{
	last_appended_subtree->SetCdr(top);
	last_appended_subtree = top;
    }
#endif

    if(NumOfErrors() == 0)
	toplevel->WalkToplevel(Token::GetEnvironment());
}

Environment* Ptree::MakeEnvironment(Environment* env)
{
    return new Environment(env);
}

void Ptree::DeleteTopEnvironment(Environment* env)
{
    env->UnlinkTop();
    delete env;
}

void Ptree::ClearBeforeAfterToplevel()
{
    before_toplevel = after_toplevel = nil;
}

void Ptree::InsertBeforeAfterToplevel(Ptree* toplevel)
{
    Token::InsertText(toplevel, before_toplevel, after_toplevel);
}

void Ptree::RecordBeforeToplevel(Ptree* p)
{
    if(p != nil)
	before_toplevel = Snoc(before_toplevel, p);
}

void Ptree::RecordAfterToplevel(Ptree* p)
{
    if(p != nil)
	after_toplevel = Snoc(after_toplevel, p);
}

void Ptree::WalkToplevel(Environment* env)
{
    this->Display();
    MopErrorMessage("Ptree::WalkToplevel()", "bomb!");
}

void UserLeaf::WalkToplevel(Environment*) {}

void PtreeChanged::WalkToplevel(Environment* env)
{
    Ptree* child = Car();
    if(child != nil)
	child->WalkToplevel(env);
}

void PtreeDeclaration::WalkToplevel(Environment* env)
{
    ClearBeforeAfterToplevel();
    Ptree* this2 = WalkStatement(env);
    Token::MinimumReplace(this, this2);
    InsertBeforeAfterToplevel(this);
}

void PtreeFunction::WalkToplevel(Environment* env)
{
    PtreeArray	before, after;

    ClearBeforeAfterToplevel();

    Token::MinimumReplace(this, WalkMemberDeclaration(env, before, after));

    Token::InsertText(this, before.All(), after.All());
    InsertBeforeAfterToplevel(this);
}

void PtreeTemplateDecl::WalkToplevel(Environment* env)
{
    ClearBeforeAfterToplevel();
    Ptree* this2 = TranslateToplevel(env);
    if(this != this2)
	Token::ReplaceText(this, this2);

    InsertBeforeAfterToplevel(this);
}

Ptree* PtreeTemplateDecl::TranslateToplevel(Environment* env)
{
    Ptree* decl = Nth(this, 4);
    if(decl == nil)
	return this;
    else if(decl->IsA(PtreeFunctionId))
	return TranslateTemplateFunction(env, decl);
    else
	return TranslateTemplateClass(env, decl);
}

Ptree* PtreeTemplateDecl::TranslateTemplateFunction(Environment* env,
						    Ptree* decl)
{
    Ptree* name = Syntax::DeclaredName(Second(decl));
    if(!name->IsLeaf())
	return this;	// operator function?	we cannot handle it.

    TemplateFunctionInfo* info = (TemplateFunctionInfo*)env->LookupAll(name);
    if(info == nil)
	return this;	// the control reaches here if this is a function
			// prototype.  See Syntax::DeclaredTemplateName().

    TemplateFunction* metaobject = info->GetTempFuncMetaobject();
    if(metaobject == nil)
	return this;
    else
	return metaobject->TranslateSelf(info->GetEnv());
}

Ptree* PtreeTemplateDecl::TranslateTemplateClass(Environment* env,
						 Ptree* decl)
{
    Ptree *name, *body, *classdef, *userkey, *classdef2;

    classdef = First(decl);
    if(classdef == nil)
	return this;

    userkey = First(classdef);
    if(userkey != nil && userkey->WhatIs() == PtreeUserKeywordId)
	classdef = Second(classdef);
    else{
	classdef = userkey;
	userkey = nil;
    }

    if(!Match(classdef, "[class %? %* %?]", &name, &body))
	return this;

    if(name == nil || body == nil)
	return this;

    TemplateInfo* info = (TemplateInfo*)env->LookupAll(name);
    if(info == nil || info->WhatIs() != TemplateType
       || info->GetMetaobject() != nil)
	return this;

    Class* metaobject = MakeMetaobject(env, userkey, name);
    if(metaobject == nil)
	return this;

    info->SetMetaobject(metaobject);
    env = info->GetEnv();
    PtreeMetaclass::RecordAllMemberTypes(env, metaobject);
    PtreeMetaclass::RecordThisVariable(env, name);

    classdef2 = metaobject->TranslateSelf(env);
    if(classdef != classdef2){
	Ptree *keyword, *lbracket, *args, *rbracket, *semicolon;
	if(!Ptree::Match(this, "[%? %? %? %? [%* %?]]",
			&keyword, &lbracket, &args, &rbracket, &semicolon)){
	    this->Display();
	    MopErrorMessage("PtreeTemplateDecl::TranslateTemplateClass()",
			    "bomb!");
	}

	return new PtreeTemplateDecl(keyword, lbracket, args, rbracket,
				     Ptree::List(Ptree::List(classdef2),
						 semicolon));
    }

    return this;
}

Class* PtreeTemplateDecl::MakeMetaobject(Environment* env,
					 Ptree* userkey, Ptree* name)
{
    BOOL error;
    Class* metaobject = PtreeMetaclass::MakeMetaobject2(env, userkey, this,
							name, error);
    if(error)
	return metaobject;
    else if(metaobject != nil)
	if(metaobject->AcceptTemplate())
	    return metaobject;
	else if(strcmp(metaobject->MetaclassName(), "Class") != 0){
	    ErrorMessage("the metaclass cannot handle the template", name);
	    return nil;
	}

    return new TemplateClass(this, nil);	// default metaobject
}

void PtreeMetaclass::WalkToplevel(Environment* env)
{
    ClearBeforeAfterToplevel();
    Token::ReplaceText(this, WalkMemberDeclaration2(env));
    InsertBeforeAfterToplevel(this);
}

Ptree* PtreeMetaclass::TranslateClass(Environment* env, Ptree* userkey,
				      Ptree* def, Class*& metaobject)
{
    Ptree* name;
    Ptree* body;
    Ptree* decl;
    ClassTypeInfo* c;

    metaobject = nil;
    if(!Match(def, "[%* %? %* %?]", &name, &body))
	return def;

    if(name == nil || body == nil)
	return def;

    c = (ClassTypeInfo*)env->LookupAll(name);
    if(c == nil || c->WhatIs() != ClassType || c->GetMetaobject() != nil)
	return def;

    if(name->Eq("Class") || name->Eq("Metaclass")
       || name->Eq("TemplateClass")){
	// For bootstrapping

	metaobject = new Metaclass(def, nil);
	c->SetMetaobject(metaobject);
	env = c->GetEnv();
	RecordAllMemberTypes(env, metaobject);
	RecordThisVariable(env, name);
	return def;	// don't call TranslateSelf()
    }
    else{
	metaobject = MakeMetaobject(env, userkey, def, name);
	if(metaobject == nil)
	    return def;
	else{
	    c->SetMetaobject(metaobject);
	    env = c->GetEnv();
	    RecordAllMemberTypes(env, metaobject);
	    RecordThisVariable(env, name);
	    return metaobject->TranslateSelf(env);
	}
    }
}

Class* PtreeMetaclass::MakeMetaobject(Environment* env, Ptree* userkey,
				      Ptree* def, Ptree* name)
{
    BOOL error;
    Class* metaobject = MakeMetaobject2(env, userkey, def, name, error);
    if(metaobject != nil || error)
	return metaobject;

    return new Class(def, nil);		// default metaobject
}

Class* PtreeMetaclass::MakeMetaobject2(Environment* env, Ptree* userkey,
				       Ptree* def, Ptree* name, BOOL& error)
{
    Class* metaobject;

    metaobject = LookupDeclarations(env, def, name, error);
    if(metaobject != nil && userkey != nil)
	ErrorMessage("The metaclass declaration conflicts with a keyword.",
		     name);

    if(metaobject != nil || error)
	return metaobject;

    metaobject = LookupKeyword(env, userkey, def, name, error);
    if(metaobject != nil || error)
	return metaobject;

    metaobject = LookupBaseMetaclass(env, def, name, error);
    if(metaobject != nil || error)
	return metaobject;

    error = FALSE;
    return nil;
}

Class* PtreeMetaclass::LookupDeclarations(Environment* env, Ptree* def,
					  Ptree* name, BOOL& error)
{
    Ptree* decl;
    Class* metaobject;

    decl = env->LookupMetaclassDeclaration(name);
    if(decl == nil)
	metaobject = nil;
    else{
	metaobject = opcxx_ListOfMetaclass::New(Nth(decl, 3), def,
						Second(Nth(decl, 4)));
	if(metaobject == nil){
	    ErrorMessage("the metaclass is not loaded:", Nth(decl, 3));
	    error = TRUE;
	    return nil;
	}
    }

    error = FALSE;
    return metaobject;
}

Class* PtreeMetaclass::LookupKeyword(Environment* env, Ptree* userkey,
				     Ptree* def, Ptree* name, BOOL& error)
{
    Ptree *key, *margs;
    Class* metaobject;
    char* mclass_name;

    if(userkey == nil){
	error = FALSE;
	return nil;
    }

    key = First(userkey);
    margs = Second(Second(userkey));
    if(key != nil && keywords != nil
       && keywords->Lookup(((Leaf*)key)->Position(), ((Leaf*)key)->Length(),
			   (HashValue*)&mclass_name)){
	metaobject = opcxx_ListOfMetaclass::New(mclass_name, def, margs);
	if(metaobject == nil)
	    ErrorMessage("the metaclass associated with the keyword"
			 "is not loaded:", userkey);
	else{
	    error = FALSE;
	    return metaobject;
	}
    }
    else
	ErrorMessage("invalid keyword:", userkey);

    error = TRUE;
    return nil;
}

Class* PtreeMetaclass::LookupBaseMetaclass(Environment* env, Ptree* def,
					   Ptree* name, BOOL& error)
{
    Ptree* p;
    BOOL is_type;
    TypeInfo t;
    Class* metaobject = nil;
    Ptree* supers;

    if(def->WhatIs() != PtreeTemplateDeclId)
	supers = Third(def);
    else{
	if(!Match(def, "[%* < %* > [[[%* %* %? %_]] %_]]", &supers))
	    MopErrorMessage("PtreeMetaclass::LookupBaseMetaclass()", "bomb!");
    }

    PtreeIter next(supers);

    next();	// skip :
    while(p = next()){
	if(env->Lookup(First(Last(p)), is_type, t)){
	    Class* m = t.ClassMetaobject();
	    if(metaobject == nil)
		metaobject = m;
	    else if(m == nil
		    || strcmp(metaobject->MetaclassName(),
			      m->MetaclassName()) != 0){
		ErrorMessage("inherited metaclasses conflict.", name);
		error = TRUE;
		return nil;
	    }
	}

	next();	// skip ,
    }

    error = FALSE;
    if(metaobject == nil)
	return nil;
    else
	return opcxx_ListOfMetaclass::New(metaobject->MetaclassName(),
					  def, nil);
}

void PtreeMetaclass::RecordThisVariable(Environment* env, Ptree* class_name)
{
    static char this_name[] = "this";

    Binding* b = new VariableNameInfo(List(class_name),
				      List(new LeafPointer()));
    if(env->AddEntry(this_name, sizeof(this_name) - 1, b) < 0)
	delete b;
}

void PtreeMetaclass::RecordAllMemberTypes(Environment* env, Class* subclass)
{
    Class* base;
    Binding* b;

    for(int i = 0; (base = subclass->NthBaseClass(env, i)) != nil; ++i)
	RecordAllMemberTypes(env, base);

    RecordMemberTypes(env, subclass);
}

void PtreeMetaclass::RecordMemberTypes(Environment* env, Class* c)
{
    Ptree* mem;
    BOOL is_function;
    Binding* b;
    PtreeIter next(c->Members());

    while(mem = next())
	switch(mem->WhatIs()){
	case PtreeDeclarationId :
	case PtreeFunctionId :
	    {
	      Ptree* type = First(mem);
	      PtreeIter next_decl(Rest(mem));
	      Ptree* decl = next_decl();
	      if(!decl->Eq(';')){
		  do{
		      Ptree* qname = Syntax::DeclaredName(decl, is_function);
		      if(qname != nil && qname->WhatIs() != LeafClassNameId){
			  if(is_function){
			      b = new MemberFunctionNameInfo(type, decl);
			      if(env->AddEntry(qname, b) < 0)
				  delete b;

			      break;
			  }
			  else{
			      b = new DataMemberNameInfo(type, decl);
			      if(env->AddEntry(qname, b) < 0)
				  delete b;
			  }
		      }
		      else
			  break;	// error?

		      next_decl();	// skip ,
		  } while(decl = next_decl());
	      }
	    }
	    break;
	case PtreeAccessCtrlId :
	case PtreeUserKeywordId :
	default :
	    break;
	}
}

void PtreeTypedef::WalkToplevel(Environment* env)
{
    ClearBeforeAfterToplevel();

    Ptree* type = Second(this);
    Ptree* type2 = WalkTypeSpecifier(env, type);
    if(type != type2)
	Token::ReplaceText(this, Replace(this, type, type2));

    InsertBeforeAfterToplevel(this);
}

void PtreeLinkage::WalkToplevel(Environment* env)
{
    ClearBeforeAfterToplevel();
    // nothing to do??
    InsertBeforeAfterToplevel(this);
}

Ptree* Ptree::WalkClassBody(Environment* env, Ptree* body,
			    PtreeArray& before, PtreeArray& after)
{
    Ptree *tail;
    int tail_pos, i;
    PtreeArray mem2(32);

    for(tail = body; tail != nil; tail = tail->Cdr())
	mem2.Append(tail->Car()->WalkMemberDeclaration(env, before, after));

    tail = body;
    tail_pos = 0;
    for(i = 0; body != nil; ++i){
	Ptree* mem = body->Car();
	body = body->Cdr();
	if(mem != mem2[i]){
	    mem2[i] = new PtreeChanged(mem2[i]);
	    tail_pos = i + 1;
	    tail = body;
	}
    }

    for(--tail_pos; tail_pos >= 0; --tail_pos)
	tail = Cons(mem2[tail_pos], tail);

    return tail;
}

Ptree* Ptree::WalkMemberDeclaration(Environment* env,
				    PtreeArray& before, PtreeArray& after)
{
    this->Display();
    MopErrorMessage("Ptree::WalkMemberDeclaration()", "bomb!");
}

/*
  WalkMemberDeclaration() trys to record declared names as non-member
  variables or functions.  But this attempt fails because those names
  must be already recorded by RecordMemberTypes().
*/
Ptree* PtreeDeclaration::WalkMemberDeclaration(Environment* env,
				PtreeArray& before, PtreeArray& after)
{
    int i;
    Ptree* this2;

    Ptree* type = this->Car();
    Ptree* type2 = WalkTypeSpecifier(env, type);

    Ptree* p = this->Cadr();
    if(p->Eq(';'))
	if(type == type2)
	    return this;
	else
	    return List(type2, p);
    else{
	Ptree* rest = this->Cdr();
	Ptree* rest2 = WalkDeclarators(env, type, rest,
				       before, after);
	if(type == type2 && rest == rest2)
	    return this;
	else
	    return Cons(type2, rest2);
    }
}

Ptree* PtreeFunction::WalkMemberDeclaration(Environment* env,
				    PtreeArray& before, PtreeArray& after)
{
    Ptree *type, *type2, *name, *declarator, *declarator2;
    Ptree *args, *args2, *init, *init2, *sts, *sts2, *body, *body2, *rest;
    Ptree *this2;
    BOOL	is_func;
    Binding*	c;
    FunctionEnvironment *newenv;
    int		i;

    if(!Match(this, "[%? %? %? %r]", &type, &declarator, &body, &rest)){
	this->Display();
	MopErrorMessage("Ptree::WalkFunctionImple()", "bomb!");
    }

    type2 = WalkTypeSpecifier(env, type);

    name = Syntax::DeclaredName(declarator, is_func);
    if(name->WhatIs() == PtreeQnameId)
	c = Syntax::LookupClassOfQname(name, env);
    else
	c = env->MostEffectiveClass();

    if(c == nil)
	newenv = new FunctionEnvironment(env);
    else
	newenv = new FunctionEnvironment(c->GetEnv());	// member function

    RecordDeclarator(env, type, declarator);
    declarator2 = WalkDeclarator(env, declarator, before, after);

    if(body->IsLeaf()){
	body2 = body;			// ';'
	init = init2 = nil;
    }
    else{
	if(Eq(First(body), ':')){
	    init = body;
	    init2 = WalkMemberInitializers(newenv, init);
	    body = First(rest);
	}
	else
	    init = init2 = nil;

	sts = Second(body);
	if(c == nil || c->GetMetaobject() == nil)
	    sts2 = PtreeBrace::WalkStatements(newenv, sts);
	else{
	    BOOL inlined;
	    if(name->IsLeaf())
		inlined = TRUE;
	    else{
		name = Ptree::First(Ptree::Last(name));
		inlined = Ptree::Eq(Ptree::First(type), "inline");
	    }

	    sts2 = c->GetMetaobject()
		    ->TranslateMemberFunctionBody(newenv, name, inlined, sts);
	    if(sts2 != nil && sts2->IsLeaf())
		sts2 = List(sts2);
	}

	Ptree* ins_decls = newenv->InsertedDeclarations();
	if(sts == sts2 && ins_decls == nil)
	    body2 = body;
	else if(ins_decls == nil)
	    body2 = new PtreeBrace(First(body), sts2, Third(body));
	else
	    body2 = new PtreeBrace(First(body), Cons(ins_decls, sts2),
				   Third(body));
    }

    if(type == type2 && declarator == declarator2 && init == init2
       && body == body2)
	this2 = this;
    else if(init2 == nil)
	this2 = List(type2, declarator2, body2);
    else
	this2 = List(type2, declarator2, init2, body2);

    DeleteTopEnvironment(newenv);
    return this2;
}

Ptree* PtreeUserKeyword::WalkMemberDeclaration(Environment* env,
				    PtreeArray& before, PtreeArray& after)
{
    return this;
}

Ptree* PtreeAccessCtrl::WalkMemberDeclaration(Environment* env,
				    PtreeArray& before, PtreeArray& after)
{
    return this;
}

Ptree* PtreeTypedef::WalkMemberDeclaration(Environment* env,
				PtreeArray& before, PtreeArray& after)
{
    Ptree* type = Second(this);
    Ptree* type2 = WalkTypeSpecifier(env, type);
    if(type != type2)
	return Replace(this, type, type2);
    else
	return this;
}

Ptree* PtreeMetaclass::WalkMemberDeclaration(Environment* env,
				    PtreeArray& before, PtreeArray& after)
{
    return WalkMemberDeclaration2(env);
}

Ptree* PtreeMetaclass::WalkMemberDeclaration2(Environment* env)
{
    Binding* info;
    Ptree* class_name = Second(this);
    if(!class_name->IsLeaf()){
	ErrorMessage("invalid class name", class_name);
	return nil;
    }

    info = env->LookupAll(class_name);
    if(info != nil){
	TypeInfoId id = info->WhatIs();
	if((id == ClassType || id == TemplateType)
	   && info->GetMetaobject() != nil){
	    ErrorMessage("the metaclass must appear before the class"
			 "is declared",
			 class_name);
	    return nil;
	}
    }

    if(env->LookupMetaclassDeclaration(class_name) == nil)
	env->AddMetaclassDeclaration(this);
    else
	ErrorMessage("the metaclass was declared more than twice",
		     class_name);

    return nil;
}

Ptree* Ptree::FindNthMemberName(Ptree* body, int nth)
{
    PtreeIter next(body);
    Ptree* p;
    while(p = next()){
	Ptree* name = p->NthMemberName(nth);
	if(name != nil)
	    return name;
	else if(nth <= 0)
	    return nil;		// error?
    }

    return nil;
}

Ptree* Ptree::NthMemberName(int& nth)
{
    this->Display();
    MopErrorMessage("Ptree::NthMemberName()", "bomb!");
    return nil;
}

Ptree* PtreeDeclaration::NthMemberName(int& nth)
{
    BOOL is_func;
    Ptree* p = this->Cadr();
    if(p->Eq(';'))
	return nil;
    else{
	for(PtreeIter i = this->Cdr(); !i.Empty(); i++){
	    if(--nth == 0)
		return Syntax::DeclaredName(*i, is_func);
	    i++;	// skip ,
	}
    }

    return nil;
}

Ptree* PtreeFunction::NthMemberName(int& nth)
{
    BOOL is_func;

    if(--nth == 0)
	return Syntax::DeclaredName(Second(this), is_func);
    else
	return nil;
}

Ptree* PtreeUserKeyword::NthMemberName(int& nth)
{
    return nil;
}

Ptree* PtreeAccessCtrl::NthMemberName(int& nth)
{
    return nil;
}

Ptree* PtreeTypedef::NthMemberName(int& nth)
{
    return nil;
}

Ptree* PtreeMetaclass::NthMemberName(int& nth)
{
    return nil;
}

int Ptree::FindMemberName(Ptree* body, Ptree* name)
{
    PtreeIter next(body);
    Ptree* p;
    int n = 0;
    while(p = next()){
	if(p->IsMember(name, n))
	    return n;
    }

    return 0;	// not found
}

BOOL Ptree::IsMember(Ptree*, int&)
{
    this->Display();
    MopErrorMessage("Ptree::IsMember()", "bomb!");
    return FALSE;
}

BOOL PtreeDeclaration::IsMember(Ptree* name, int& nth)
{
    BOOL is_func;
    Ptree* p = this->Cadr();
    if(p->Eq(';'))
	return FALSE;
    else{
	for(PtreeIter i = this->Cdr(); !i.Empty(); i++){
	    nth++;
	    if(Eq(name, Syntax::DeclaredName(*i, is_func)))
		return TRUE;

	    i++;	// skip ,
	}
    }

    return FALSE;
}

BOOL PtreeFunction::IsMember(Ptree* name, int& nth)
{
    BOOL is_func;

    nth++;
    return Eq(name, Syntax::DeclaredName(Second(this), is_func));
}

BOOL PtreeUserKeyword::IsMember(Ptree*, int&)
{
    return FALSE;
}

BOOL PtreeAccessCtrl::IsMember(Ptree*, int&)
{
    return FALSE;
}

BOOL PtreeTypedef::IsMember(Ptree*, int&)
{
    return FALSE;
}

BOOL PtreeMetaclass::IsMember(Ptree*, int&)
{
    return FALSE;
}

Ptree* Ptree::WalkStatement(Environment* env)
{
    this->Display();
    // because this is invoked by Class::TranslateStatement()
    MopErrorMessage("Class::TranslateStatement()", "not a statement");
    return this;
}

Ptree* UserLeaf::WalkStatement(Environment* env)
{
    return this;
}

Ptree* PtreeChanged::WalkStatement(Environment* env)
{
    Ptree* child = Car();
    if(child == nil)
	return child;
    else
	return child->WalkStatement(env);
}

Ptree* PtreeDeclaration::WalkStatement(Environment* env)
{
    int i;
    PtreeArray before, after;

    Ptree* this2 = WalkMemberDeclaration(env, before, after);

    this2 = Nconc(before.All(), this2);

    if(after.Number() > 0)
	this2 = Append(this2, after.All());

    return this2;	// don't return a PtreeChanged object because
			// this2 may be a class declaration.
}

Ptree* PtreeExprStatement::WalkStatement(Environment* env)
{
    Ptree* expr = Car();
    Ptree* expr2 = WalkExpr(env, expr);
    if(expr == expr2)
	return this;
    else if(expr2 == nil || expr2->WhatIs() == PtreeNonExprId)
	return expr2;
    else
	return new PtreeChanged(Substitute(this, 1, expr2, expr));
}

/* compound statement */
Ptree* PtreeBrace::WalkStatement(Environment* env)
{
    Environment* new_env = MakeEnvironment(env);
    Ptree* body = this->Cadr();
    Ptree* body2 = WalkStatements(new_env, body);
    DeleteTopEnvironment(new_env);
    return Substitute(this, 2, body2, body);
}

Ptree* PtreeBrace::WalkStatements(Environment* env, Ptree* body)
{
    const N = 16;
    Ptree* st2[N];
    Ptree *tail, *body2;
    int tail_pos, i;

    tail = body;
    tail_pos = 0;
    for(i = 0; body != nil && i < N; ++i){
	Ptree* st = body->Car();
	st2[i] = st->WalkStatement(env);
	body = body->Cdr();
	if(st != st2[i]){
	    tail_pos = i + 1;
	    tail = body;
	}
    }

    if(body == nil){
	i = tail_pos;
	body2 = tail;
    }
    else{
	body2 = WalkStatements(env, body);
	if(body == body2){
	    i = tail_pos;
	    body2 = tail;
	}
    }

    for(--i; i >= 0; --i)
	body2 = Cons(st2[i], body2);

    return body2;
}

Ptree* PtreeLabel::WalkStatement(Environment* env)
{
    Ptree* st = Third(this);
    if(st == nil)
	return this;
    else{
	Ptree* st2 = st->WalkStatement(env);
	return Substitute(this, 3, st2, st);
    }
}

Ptree* PtreeCaseLabel::WalkStatement(Environment* env)
{
    Ptree *expr, *expr2, *st, *st2;
    expr = Second(this);
    expr2 = WalkExpr(env, expr);
    st = Nth(this, 3);
    st2 = st->WalkStatement(env);
    return Substitute(this, 2, expr2, expr, 4, st2, st);
}

Ptree* PtreeIfStatement::WalkStatement(Environment* env)
{
    Ptree *cond, *cond2, *then_part, *then_part2;
    Ptree *rest;

    if(!Ptree::Match(this, "[%* %* %? %* %? %r]",
		     &cond, &then_part, &rest))
	MopErrorMessage("PtreeIfStatement::WalkStatement()", "bomb!");

    cond2 = WalkExpr(env, cond);
    then_part2 = then_part->WalkStatement(env);
    if(rest == nil)
	return Substitute(this, 3, cond2, cond, 5, then_part2, then_part);
    else{
	Ptree* else_part = Ptree::Second(rest);
	Ptree* else_part2 = else_part->WalkStatement(env);
	return Substitute(this, 3, cond2, cond, 5, then_part2, then_part,
			  7, else_part2, else_part);
    }
}

Ptree* PtreeSwitchStatement::WalkStatement(Environment* env)
{
    Ptree *cond, *cond2, *body, *body2;

    if(!Ptree::Match(this, "[%* %* %? %* %?]", &cond, &body))
	MopErrorMessage("PtreeSwitchStatement::WalkStatement()", "bomb!");

    cond2 = WalkExpr(env, cond);
    body2 = body->WalkStatement(env);
    return Substitute(this, 3, cond2, cond, 5, body2, body);
}

Ptree* PtreeWhileStatement::WalkStatement(Environment* env)
{
    Ptree *cond, *cond2, *body, *body2;

    if(!Ptree::Match(this, "[%* %* %? %* %?]", &cond, &body))
	MopErrorMessage("PtreeWhileStatement::WalkStatement()", "bomb!");

    cond2 = WalkExpr(env, cond);
    body2 = body->WalkStatement(env);
    return Substitute(this, 3, cond2, cond, 5, body2, body);
}

Ptree* PtreeDoStatement::WalkStatement(Environment* env)
{
    Ptree *body, *body2, *cond, *cond2;

    if(!Ptree::Match(this, "[%* %? %* %* %? %_]", &body, &cond))
	MopErrorMessage("PtreeDoStatement::WalkStatement()", "bomb!");

    body2 = body->WalkStatement(env);
    cond2 = WalkExpr(env, cond);
    return Substitute(this, 2, body2, body, 5, cond2, cond);
}

Ptree* PtreeForStatement::WalkStatement(Environment* env)
{
    Ptree *init, *init2, *cond, *cond2, *iter, *iter2, *body, *body2;

    if(!Match(this, "[%* %* %? %? %* %? %* %?]",
		     &init, &cond, &iter, &body))
	MopErrorMessage("PtreeForStatement::WalkStatement()", "bomb!");

    Environment* new_env = MakeEnvironment(env);

    init2 = init->WalkStatement(new_env);
    cond2 = WalkExpr(new_env, cond);
    iter2 = WalkExpr(new_env, iter);
    body2 = body->WalkStatement(env);

    DeleteTopEnvironment(new_env);

    return Substitute(this, 3, init2, init, 4, cond2, cond, 6, iter2, iter,
		      8, body2, body);
}

Ptree* PtreeReturnStatement::WalkStatement(Environment* env)
{
    Ptree* expr = Ptree::Second(this);
    Ptree* expr2 = WalkExpr(env, expr);
    return Substitute(this, 2, expr2, expr);
}

Ptree* PtreeBreakStatement::WalkStatement(Environment*)
{
    return this;
}

Ptree* PtreeContinueStatement::WalkStatement(Environment*)
{
    return this;
}

Ptree* PtreeGotoStatement::WalkStatement(Environment*)
{
    return this;
}

/*
  TYPESPEC may be nil.
  Curently, we ignore types locally defined in a function scope.
*/
Ptree* Ptree::WalkTypeSpecifier(Environment* env, Ptree* typespec)
{
    const N = 16;
    Ptree* type[N];
    Ptree *car, *car2, *rest, *rest2, *userkey;
    int i;
    Class* metaobject = nil;

    userkey = nil;
    rest = typespec;
    for(i = 0; rest != nil && i < N; ++i, rest = rest->Cdr()){
	type[i] = car = rest->Car();
	switch(car->WhatIs()){
	case PtreeEnumId :
	    return typespec;
	case PtreeClassId :	// class NAME { ... }
	    car2 = PtreeMetaclass::TranslateClass(env, userkey, car,
						  metaobject);
	    if(car == car2 && userkey == nil)
		return typespec;
	    else{
		rest2 = Cons(car2, rest->Cdr());
		for(--i; i >= 0; --i)
		    rest2 = Cons(type[i], rest2);

		return rest2;
	    }
	case PtreeTemplateNameId :
	case LeafTypeNameId :
	case PtreeQClassNameId :
	case LeafClassNameId :
	    metaobject = env->LookupClassMetaobject(car);
	    if(metaobject != nil){
		car2 = metaobject->TranslateClassName(env, userkey, car);
		if(car == car2 && userkey == nil)
		    return typespec;
		else{
		    rest2 = Cons(car2, rest->Cdr());
		    for(--i; i >= 0; --i)
			rest2 = Cons(type[i], rest2);

		    return rest2;
		}
	    }

	    return typespec;
	case PtreeUserKeywordId :
	    if(userkey == nil)
		userkey = car;
	    else
		ErrorMessage("too many user keywords", typespec, nil);

	    --i;
	    break;
	case LeafReservedId :
	case LeafReservedModifierId :
	default :
	    break;
	}
    }

    if(rest == nil)
	return typespec;
    else{
	Ptree* rest2 = WalkTypeSpecifier(env, rest);
	if(rest == rest2)
	    return typespec;
	else{
	    for(--i; i >= 0; --i)
		rest2 = Cons(type[i], rest2);

	    return rest2;
	}
    }
}

Ptree* Ptree::WalkMemberInitializers(Environment* env, Ptree* init)
{
    Ptree* mems = init->Cdr();
    Ptree* mems2 = Map2(env, mems, &Ptree::WalkMemberInitializersCore);
    if(mems == mems2)
	return init;
    else
	return Cons(init->Car(), mems2);
}

Ptree* Ptree::WalkMemberInitializersCore(Environment* env, Ptree* mem)
{
    Ptree *name, *name2, *args, *args2;
    Class* metaobject;

    name = name2 = First(mem);
    switch(name->WhatIs()){
    case LeafClassNameId :
    case PtreeTemplateNameId :
	metaobject = env->LookupClassMetaobject(name);
	if(metaobject != nil)
	    name2 = metaobject->TranslateClassName(env, nil, name);
	break;
    default :
	break;
    }

    args = Third(mem);
    args2 = WalkExprList(env, args);

    if(name == name2 && args == args2)
	return mem;
    else
	return Substitute(mem, 1, name2, name, 3, args2, args);
}

Ptree* Ptree::WalkDeclarators(Environment* env, Ptree* type, Ptree* decl,
			      PtreeArray& before, PtreeArray& after)
{
    const N = 32;	// must be an even number
    Ptree* d2[N];
    Ptree *tail, *decl2;
    int tail_pos, i;

    tail = decl;
    tail_pos = 0;
    for(i = 0; decl != nil && i < N;){
	Ptree* d = decl->Car();
	RecordDeclarator(env, type, d);
	d2[i] = WalkDeclarator(env, d, before, after);
	decl = decl->Cdr();
	if(d != d2[i++]){
	    tail_pos = i;
	    tail = decl;
	}

	if(decl != nil){
	    d2[i++] = decl->Car();		// skip ',' and ';'
	    decl = decl->Cdr();
	}
    }

    if(decl == nil){
	i = tail_pos;
	decl2 = tail;
    }
    else{
	decl2 = WalkDeclarators(env, type, decl, before, after);
	if(decl == decl2){
	    i = tail_pos;
	    decl2 = tail;
	}
    }

    for(--i; i >= 0; --i)
	decl2 = Cons(d2[i], decl2);

    return decl2;
}

Ptree* Ptree::WalkArguments(Environment* env, Ptree* args,
			    PtreeArray& before, PtreeArray& after)
{
    const N = 32;	// must be an even number
    Ptree* a2[N];
    Ptree *tail, *args2, *a, *type, *type2, *decl, *decl2;
    int tail_pos, i;

    tail = args;
    tail_pos = 0;
    for(i = 0; args != nil && i < N;){
	a = args->Car();
	args = args->Cdr();
	if(a->Eq("...")){
	    a2[i++] = a;
	    break;
	}

	type = a->Car();
	type2 = WalkTypeSpecifier(env, type);
	decl = a->Cadr();
	RecordDeclarator(env, type, decl);
	decl2 = WalkDeclarator(env, decl, before, after);
	if(type == type2 && decl == decl2)
	    a2[i++] = a;
	else{
	    a2[i++] = List(type2, decl2);
	    tail_pos = i;
	    tail = args;
	}

	if(args != nil){
	    a2[i++] = args->Car();		// skip ','
	    args = args->Cdr();
	}
    }

    if(args == nil){
	i = tail_pos;
	args2 = tail;
    }
    else{
	args2 = WalkArguments(env, args, before, after);
	if(args == args2){
	    i = tail_pos;
	    args2 = tail;
	}
    }

    for(--i; i >= 0; --i)
	args2 = Cons(a2[i], args2);

    return args2;
}

void Ptree::RecordDeclarator(Environment* env, Ptree* type, Ptree* decl)
{
    BOOL is_function;
    Binding* b;

    Ptree* qname = Syntax::DeclaredName(decl, is_function);
    if(qname != nil){
	if(is_function)
	    b = new FunctionNameInfo(type, decl);
	else
	    b = new VariableNameInfo(type, decl);

	if(env->AddEntry(qname, b) < 0)
	    delete b;
    }
}

Ptree* Ptree::WalkDeclarator(Environment* env, Ptree* decl,
			     PtreeArray& before, PtreeArray& after)
{
    Ptree *p, *p2, *expr, *expr2;

    if(decl == nil)
	return decl;
    else{
	p = p2 = decl->Car();
	if(p == nil)
	    return decl;	// the rest is cv.qualifiers

	switch(p->WhatIs()){
	case LeafClassNameId :
	case PtreeQnameId :
	    p2 = WalkQualifiedName(env, p);
	    // don't break here.
	case LeafNameId :
	case PtreeDestructorId :
	case PtreeOperatorFuncId :
	    expr = decl->Cdr();
	    expr2 = WalkInitializerEtc(env, p, expr, before, after);
	    if(p == p2 && expr == expr2)
		return decl;
	    else if(expr2 != nil && expr2->IsLeaf())
		return List(p2, expr2);
	    else
		return Cons(p2, expr2);
	case PtreeParenId :
	    expr = p->Cadr();
	    expr2 = WalkDeclarator(env, expr, before, after);
	    if(expr != expr2)
		p2 = new PtreeParen(p->Car(), expr2, p->Cddr()->Car());

	    // don't break here.
	default :
	    expr = decl->Cdr();
	    expr2 = WalkDeclarator(env, expr, before, after);
	    if(p == p2 && expr == expr2)
		return decl;
	    else if(expr2 != nil && expr2->IsLeaf())
		return List(p2, expr2);
	    else
		return Cons(p2, expr2);
	}
    }
}

Ptree* Ptree::WalkInitializerEtc(Environment* env,
				 Ptree* qname, Ptree* expr,
				 PtreeArray& before, PtreeArray& after)
{
    Ptree *p, *q, *q2;
    TypeInfo tinfo;
    BOOL is_type;
    Class* metaobject;

    if(expr == nil)
	return expr;

    p = expr->Car();
    q = expr->Cadr();
    if(p->Eq(':')){			// bitfield
	q2 = WalkExpr(env, q);
	if(q == q2)
	    return expr;
	else
	    return Cons(p, q2);
    }
    else if(p->Eq('(')			// formal arguments
	    && (q == nil || q->WhatIs() != PtreeActualArgsId)){
	q2 = WalkArguments(env, q, before, after);
	if(q == q2)
	    return expr;
	else
	    return Cons(p, Cons(q2, expr->Cddr()));
    }
    else{				// initializer
	env->Lookup(qname, is_type, tinfo);
	metaobject = tinfo.ClassMetaobject();
	if(metaobject != nil)
	    return metaobject->TranslateInitializer(env, expr, before, after);
	else{
	    q2 = WalkExpr(env, q);
	    if(q == q2)
		return expr;
	    else
		return Cons(p, Cons(q2, expr->Cddr()));
	}
    }
}

Ptree* Ptree::WalkQualifiedName(Environment* env, Ptree* qname)
{
    switch(qname->WhatIs()){
    case LeafClassNameId :
    case PtreeTemplateNameId :
	return WalkClassName(env, qname);
    case PtreeQnameId :
	if(qname->Car()->Eq("::"))
	    env = env->GetBottom();

	return WalkQualifiedName2(env, qname);
    default :
	return qname;
    }
}

Ptree* Ptree::WalkClassName(Environment*& env, Ptree* name)
{
    Binding* b;
    Class *metaobject;

    if(name->WhatIs() == PtreeTemplateNameId)
	name = name->Car();

    b = env->LookupAll(name);
    if(b != nil)
	switch(b->WhatIs()){
	case ClassType :
	case TemplateType :
	    metaobject = b->GetMetaobject();
	    env = b->GetEnv();
	    if(metaobject == nil)
		return name;
	    else
		return metaobject->TranslateClassName(env, nil, name);
	default :
	    break;
	}

    return name;	// error
}

Ptree* Ptree::WalkQualifiedName2(Environment* env, Ptree* qname)
{
    Ptree* h = qname->Car();
    Ptree* t = qname->Cdr();
    if(t == nil)
	return qname;
    else{
	if(h->Eq("::")){
	    Ptree* t2 = WalkQualifiedName2(env, t);
	    if(t == t2)
		return qname;
	    else
		return Ptree::Cons(h, t2);
	}
	else{
	    Ptree* h2 = WalkClassName(env, h);
	    Ptree* t2 = WalkQualifiedName2(env, t);
	    if(h == h2 && t == t2)
		return qname;
	    else
		return Ptree::Cons(h2, t2);
	}
    }
}
