/*
  Copyright (c) 1995, 1996 Xerox Corporation.
  All Rights Reserved.

  Use and copying of this software and preparation of derivative works
  based upon this software are permitted. Any copy of this software or
  of any derivative work must include the above copyright notice of
  Xerox Corporation, this paragraph and the one after it.  Any
  distribution of this software or derivative works must comply with all
  applicable United States export control laws.

  This software is made available AS IS, and XEROX CORPORATION DISCLAIMS
  ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
  LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
  EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
  NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGES.

  Author: Shigeru Chiba		chiba@is.s.u-tokyo.ac.jp
*/

#include "ptree.h"
#include "token.h"
#include "typeinfo.h"
#include "syntax.h"
#include "mop.h"

Ptree* Ptree::WalkExpr(Environment* env, Ptree* expr)
{
    TypeInfo t;

    if(expr == nil)
	return expr;

    t.Undef();
    Ptree* new_expr = expr->WalkExpr(env, t);
    if(new_expr == expr)
	return expr;
    else if(new_expr == nil || new_expr->WhatIs() == PtreeNonExprId)
	return new_expr;
    else
	return new PtreeChanged(new_expr);
}

Ptree* Ptree::WalkExpr(Environment* env, TypeInfo& t)
{
    this->Display();
    // because this is invoked by Class::TranslateExpression()
    MopErrorMessage("Class::TranslateExpression()", "not an expression");
    return this;
}

void Ptree::Type(Environment* env, TypeInfo& t)
{
    this->Display();
    MopErrorMessage("Ptree::Type()", "bomb!");
}

Ptree* UserLeaf::WalkExpr(Environment*, TypeInfo& t)
{
    t.Undef();
    return this;
}

void UserLeaf::Type(Environment*, TypeInfo& t)
{
    t.Undef();
}

Ptree* PtreeChanged::WalkExpr(Environment* env, TypeInfo& t)
{
    Ptree* child = Car();
    if(child == nil){
	t.Undef();
	return child;
    }
    else
	return child->WalkExpr(env, t);
}

void PtreeChanged::Type(Environment* env, TypeInfo& t)
{
    Ptree* child = Car();
    if(child == nil)
	t.Undef();
    else
	child->Type(env, t);
}

Ptree* PtreeAssignExpr::WalkExpr(Environment* env, TypeInfo& t)
{
    uint whatis;
    Class* metaobject;

    Ptree* lexpr = First(this);
    switch(lexpr->WhatIs()){
    case LeafNameId :
    case PtreeQnameId :
	whatis = Syntax::IsMemberName(lexpr, env, t, metaobject);
	t.DereferenceIfRefType();
	if(metaobject != nil && whatis == DataMemberName){
	    Ptree* expr = metaobject->TranslateMemberWrite(env, lexpr,
							   Second(this),
							   Third(this));
	    return CheckEquiv(this, expr);
	}
	break;
    default :
	lexpr->Type(env, t);
	break;
    }

    metaobject = t.ClassMetaobject();
    if(metaobject == nil){
	TypeInfo t1;
	Ptree* lexpr2 = lexpr->WalkExpr(env, t);
	Ptree* rexpr = Third(this);
	Ptree* rexpr2 = rexpr->WalkExpr(env, t1);
	if(lexpr != lexpr2 || rexpr != rexpr2)
	    return new PtreeAssignExpr(lexpr2, Second(this), rexpr2);
	else
	    return this;
    }
    else{
	Ptree* expr = metaobject->TranslateAssign(env, lexpr,
						  Second(this), Third(this));
	return CheckEquiv(this, expr);
    }
}

void PtreeAssignExpr::Type(Environment* env, TypeInfo& t)
{
    Class* metaobject;

    Ptree* lexpr = First(this);
    switch(lexpr->WhatIs()){
    case LeafNameId :
    case PtreeQnameId :
	Syntax::IsMemberName(lexpr, env, t, metaobject);
	break;
    default :
	lexpr->Type(env, t);
	break;
    }

    t.DereferenceIfRefType();
}

// the resulting value of t is not collect if an operator is overloaded.
//
Ptree* PtreeInfixExpr::WalkExpr(Environment* env, TypeInfo& t)
{
    Ptree* lexpr = First(this);
    Ptree* op = Second(this);
    Ptree* rexpr = Third(this);
    rexpr->Type(env, t);
    Class* metaobject = t.ClassMetaobject();
    if(metaobject != nil){
	Ptree* this2 =  metaobject->TranslateBinary(env, lexpr, op, rexpr);
	return CheckEquiv(this, this2);
    }
    else{
	TypeInfo t1;
	Ptree* lexpr2 = lexpr->WalkExpr(env, t1);
	Ptree* rexpr2 = rexpr->WalkExpr(env, t1);
	if(lexpr != lexpr2 || rexpr != rexpr2)
	    return new PtreeInfixExpr(lexpr2, op, rexpr2);
	else
	    return this;
    }
}

void PtreeInfixExpr::Type(Environment* env, TypeInfo& t)
{
    Ptree* lexpr = First(this);
    lexpr->Type(env, t);
}

Ptree* PtreeCondExpr::WalkExpr(Environment* env, TypeInfo& t)
{
    TypeInfo t1;

    Ptree* cexpr = First(this);
    Ptree* cexpr2 = cexpr->WalkExpr(env, t1);
    Ptree* thenexpr = Third(this);
    Ptree* thenexpr2 = thenexpr->WalkExpr(env, t);
    Ptree* elseexpr = Nth(this, 4);
    Ptree* elseexpr2 = elseexpr->WalkExpr(env, t);

    if(cexpr != cexpr2 || thenexpr != thenexpr2 || elseexpr != elseexpr2)
	return new PtreeCondExpr(cexpr2, Second(this), thenexpr2,
				 Nth(this, 3), elseexpr2);
    else
	return this;
}

void PtreeCondExpr::Type(Environment* env, TypeInfo& t)
{
    Third(this)->Type(env, t);
}

Ptree* PtreeCast::WalkExpr(Environment* env, TypeInfo& t)
{
    Ptree* type;
    Ptree* expr;
    TypeInfo t1;

    if(!Match(this, "[[%* %? %*] %?]", &type, &expr)){
	this->Display();
	MopErrorMessage("PtreeCast::WalkExpr()", "bomb!");
    }

    Ptree* expr2 = expr->WalkExpr(env, t1);
    Syntax::AnalyzeCastType(type, env, t);
    t.DereferenceIfRefType();
    Ptree* type2 = WalkCastType(env, type);
    Ptree* castexpr = First(this);
    if(type != type2)
	return new PtreeCast(new PtreeParen(First(castexpr), type2,
					    Third(castexpr)),
			     expr2);
    else if(expr != expr2)
	return new PtreeCast(castexpr, expr2);
    else
	return this;
}

void PtreeCast::Type(Environment* env, TypeInfo& t)
{
    Syntax::AnalyzeCastType(Second(First(this)), env, t);
    t.DereferenceIfRefType();
}

Ptree* Ptree::WalkCastType(Environment* env, Ptree* type)
{
    Ptree* typespec = First(type);
    Ptree* typespec2 = WalkTypeSpecifier(env, typespec);
    if(typespec == typespec2)
	return type;
    else
	return Cons(typespec2, type->Cdr());
}

Ptree* PtreeUnaryExpr::WalkExpr(Environment* env, TypeInfo& t)
{
    Ptree* op = this->Car();
    Ptree* expr = this->Cadr();
    expr->Type(env, t);
    Class* metaobject = t.ClassMetaobject();
    if(op->Eq('*'))
	t.Dereference();
    else if(op->Eq('&'))
	t.Reference();
    else if(op->Eq('!'))
	t.BeBoolean();

    t.DereferenceIfRefType();

    if(metaobject != nil){
	Ptree* this2 =  metaobject->TranslateUnary(env, op, expr);
	return CheckEquiv(this, this2);
    }
    else{
	TypeInfo t1;
	Ptree* expr2 = expr->WalkExpr(env, t1);
	if(expr == expr2)
	    return this;
	else
	    return new PtreeUnaryExpr(op, expr2);
    }
}

void PtreeUnaryExpr::Type(Environment* env, TypeInfo& t)
{
    this->Cadr()->Type(env, t);
    Ptree* op = this->Car();
    if(op->Eq('*'))
	t.Dereference();
    else if(op->Eq('&'))
	t.Reference();
    else if(op->Eq('!'))
	t.BeBoolean();

    t.DereferenceIfRefType();
}

/*
  new expression ::=

  [keyword | "::"] "new" ["(" placement ")"] new-type ["(" initializer ")"]
*/
Ptree* PtreeNew::WalkExpr(Environment* env, TypeInfo& t)
{
    TypeInfo t1;
    Ptree* header;
    Ptree* op;
    Ptree* placement;
    Ptree* new_type;
    Ptree* rest;

    if(!Match(this, "[%? %? %? %? %r]",
	      &header, &op, &placement, &new_type, &rest)){
	this->Display();
	MopErrorMessage("PtreeNew::WalkExpr()", "bomb!");
    }

    if(new_type->WhatIs() == PtreeParenId)
	Syntax::AnalyzeCastType(Second(new_type), env, t);
    else
	Syntax::AnalyzeCastType(new_type, env, t);

    Class* metaobject = t.ClassMetaobject();
    if(metaobject != nil){
	Ptree* expr = metaobject->TranslateNew(env, header, op, placement,
					       new_type, First(rest));
	return CheckEquiv(this, expr);
    }
    else{
	Ptree* placement2;
	Ptree* arglist;
	Ptree* arglist2;

	if(placement != nil)
	    placement2 = placement->WalkExpr(env, t1);
	else
	    placement2 = nil;

	if(rest == nil)
	    arglist = arglist2 = nil;
	else{
	    arglist = First(rest);
	    arglist2 = Class::TranslateArgumentList(env, arglist);
	}

	if(placement != placement2 || arglist != arglist2)
	    if(arglist2 == nil)
		return List(header, op, placement2, new_type);
	    else
		return List(header, op, placement2, new_type, arglist2);
	else
	    return this;
    }
}

void PtreeNew::Type(Environment* env, TypeInfo& t)
{
    Ptree* new_type = Nth(this, 3);
    if(new_type->WhatIs() == PtreeParenId)
	Syntax::AnalyzeCastType(Second(new_type), env, t);
    else
	Syntax::AnalyzeCastType(new_type, env, t);
}

// delete expression ::= ["::"] "delete" ["[" "]"] cast-expr
//
Ptree* PtreeDelete::WalkExpr(Environment* env, TypeInfo& t)
{
    Ptree* expr = Nth(this, 3);
    Ptree* expr2 = expr->WalkExpr(env, t);
    t.BeVoid();
    if(expr == expr2)
	return this;
    else
	return new PtreeDelete(First(this), Second(this), Third(this),
			       expr2);
}

void PtreeDelete::Type(Environment* env, TypeInfo& t)
{
    t.BeVoid();
}

Ptree* LeafConstValue::WalkExpr(Environment* env, TypeInfo& t)
{
    Syntax::AnalyzeConstType(this, env, t);
    return this;
}

void LeafConstValue::Type(Environment* env, TypeInfo& t)
{
    Syntax::AnalyzeConstType(this, env, t);
}

Ptree* LeafThis::WalkExpr(Environment* env, TypeInfo& t)
{
    BOOL is_type;
    env->Lookup(this, is_type, t);
    return this;
}

void LeafThis::Type(Environment* env, TypeInfo& t)
{
    BOOL is_type;
    env->Lookup(this, is_type, t);
}

Ptree* PtreeQname::WalkExpr(Environment* env, TypeInfo& t)
{
    Class* metaobject;

    uint whatis = Syntax::IsMemberName(this, env, t, metaobject);
    t.DereferenceIfRefType();
    if(metaobject != nil && whatis == DataMemberName){
	Ptree* expr = metaobject->TranslateMemberRead(env, this);
	if(this != expr)
	    return expr;
    }

    return this;
}

void PtreeQname::Type(Environment* env, TypeInfo& t)
{
    Class* metaobject;

    Syntax::IsMemberName(this, env, t, metaobject);
    t.DereferenceIfRefType();
}

Ptree* LeafName::WalkExpr(Environment* env, TypeInfo& t)
{
    Class* metaobject;

    uint whatis = Syntax::IsMemberName(this, env, t, metaobject);
    t.DereferenceIfRefType();
    if(metaobject != nil && whatis == DataMemberName){
	Ptree* expr = metaobject->TranslateMemberRead(env, this);
	if(this != expr)
	    return expr;
    }

    return this;
}

void LeafName::Type(Environment* env, TypeInfo& t)
{
    Class* metaobject;

    Syntax::IsMemberName(this, env, t, metaobject);
    t.DereferenceIfRefType();
}

Ptree* PtreeDestructor::WalkExpr(Environment* env, TypeInfo& t)
{
    t.BeVoid();
    return this;
}

void PtreeDestructor::Type(Environment* env, TypeInfo& t)
{
    t.BeVoid();
}

Ptree* PtreeOperatorFunc::WalkExpr(Environment* env, TypeInfo& t)
{
    t.Undef();	// not implemented yet.
    return this;
}

void PtreeOperatorFunc::Type(Environment* env, TypeInfo& t)
{
    t.Undef();	// not implemented yet.
}

Ptree* PtreeParen::WalkExpr(Environment* env, TypeInfo& t)
{
    Ptree* expr = Second(this);
    Ptree* expr2 = expr->WalkExpr(env, t);
    if(expr != expr2)
	return new PtreeParen(First(this), expr2, Third(this));
    else
	return this;
}

void PtreeParen::Type(Environment* env, TypeInfo& t)
{
    Second(this)->Type(env, t);
}

Ptree* PtreePostfixExpr::WalkExpr(Environment* env, TypeInfo& t)
{
    Ptree *expr, *expr2, *this2;
    Ptree *op, *op2;
    Class* metaobject;
    uint whatis;
    TypeInfo t1;

    expr = First(this);
    op = Second(this);

    switch(op->WhatIs()){
    case PtreeParenId :		// function
	return WalkFuncall(env, expr, op, t);
    case PtreeBracketId :	// array
	expr->Type(env, t);
	metaobject = t.ClassMetaobject();
	t.Dereference();
	t.DereferenceIfRefType();
	if(metaobject != nil){
	    this2 = metaobject->TranslateSubscript(env, expr, op);
	    return CheckEquiv(this, this2);
	}
	else{
	    expr2 = expr->WalkExpr(env, t1);
	    op2 = Class::TranslateIndex(env, op);
	}
	break;
    default :	// IncOp
	expr->Type(env, t);
	metaobject = t.ClassMetaobject();
	if(metaobject != nil){
	    this2 = metaobject->TranslatePostfix(env, expr, op);
	    return CheckEquiv(this, this2);
	}
	else{
	    expr2 = expr->WalkExpr(env, t);
	    op2 = op;
	}
	break;
    }

    if(expr != expr2 || op != op2)
	return new PtreePostfixExpr(expr2, op2);
    else
	return this;
}

void PtreePostfixExpr::Type(Environment* env, TypeInfo& t)
{
    Ptree *expr, *op;
    Class* metaobject;
    uint whatis;

    expr = First(this);
    op = Second(this);

    switch(op->WhatIs()){
    case PtreeParenId :		// function
	if(!IsMemberCall(env, expr, t, metaobject))
	    expr->Type(env, t);

	if(t.WhatIs() == PointerType)
	    t.Dereference();	// fptr() is an alias of (*fptr)().

	t.Dereference();
	t.DereferenceIfRefType();
	break;
    case PtreeBracketId :	// array
	expr->Type(env, t);
	t.Dereference();
	t.DereferenceIfRefType();
	break;
    default :	// IncOp
	expr->Type(env, t);
	break;
    }
}

BOOL PtreePostfixExpr::IsMemberCall(Environment* env, Ptree* expr,
				      TypeInfo& t, Class*& metaobject)
{
    TypeInfoId whatis;

    switch(expr->WhatIs()){
    case LeafNameId :
    case PtreeQnameId :
	whatis = Syntax::IsMemberName(expr, env, t, metaobject);
	t.DereferenceIfRefType();
	return BOOL(whatis == MemberFunctionName);
    default :
	return FALSE;
    }
}

Ptree* PtreePostfixExpr::WalkFuncall(Environment* env, Ptree* expr, Ptree* op,
				     TypeInfo& t)
{
    Class* metaobject;
    Ptree *expr2, *op2, *this2;
    TypeInfo t1;

    if(IsMemberCall(env, expr, t, metaobject))
	if(metaobject == nil)
	    expr2 = expr;
	else{
	    t.Dereference();
	    t.DereferenceIfRefType();
	    this2 = metaobject->TranslateMemberCall(env, expr, op);
	    return CheckEquiv(this, this2);
	}
    else{
	expr->Type(env, t1);
	if(t1.WhatIs() == TemplateFunctionType)
	    return Syntax::TranslateTempFuncall(t1, env, this);
	else{
	    metaobject = t1.ClassMetaobject();
	    if(metaobject == nil)
		expr2 = expr->WalkExpr(env, t);
	    else{
		t.Dereference();
		t.DereferenceIfRefType();
		this2 = metaobject->TranslateFunctionCall(env, expr, op);
		return CheckEquiv(this, this2);
	    }
	}
    }

    if(t.WhatIs() == PointerType)
	t.Dereference();	// fptr() is an alias of (*fptr)().

    t.Dereference();
    t.DereferenceIfRefType();
    op2 = Class::TranslateArgumentList(env, op);

    if(expr != expr2 || op != op2)
	return new PtreePostfixExpr(expr2, op2);
    else
	return this;
}

Ptree* PtreeMemberAccessExpr::WalkExpr(Environment* env, TypeInfo& t)
{
    TypeInfo t1;
    Ptree* object;
    Ptree* op;
    Ptree* member;
    Ptree* rest;
    Ptree* expr;

    if(!Match(this, "[%? %? %? %r]", &object, &op, &member, &rest))
	MopErrorMessage("PtreeMemberAccessExpr::WalkExpr()", "bomb!");

    object->Type(env, t1);
    if(!op->Eq('.'))
	t1.Dereference();	// the operator is ->

    uint whatis = Syntax::LookupMemberType(member, t1, t);
    if(whatis == DataMemberName || whatis == MemberFunctionName)
	t.DereferenceIfRefType();
    else
	t.Undef();

    Class* metaobject = t1.ClassMetaobject();
    if(metaobject != nil){
	if(rest == nil)
	    expr = metaobject->TranslateMemberRead(env, object, op,
						   member);
	else if(Length(rest) == 2)
	    expr = metaobject->TranslateMemberWrite(env, object, op,
						    member, First(rest),
						    Second(rest));
	else{
	    t.Dereference();
	    t.DereferenceIfRefType();
	    expr = metaobject->TranslateMemberCall(env, object, op,
						   member, First(rest));
	}

	return CheckEquiv(this, expr);
    }

    Ptree* object2 = object->WalkExpr(env, t1);
    if(rest == nil)
	if(object != object2)
	    return new PtreeMemberAccessExpr(object2, op, member);
	else
	    return this;
    else if(Length(rest) == 2){
	TypeInfo t2;
	Ptree* rexpr = Second(rest);
	Ptree* rexpr2 = rexpr->WalkExpr(env, t2);
	if(object != object2 || rexpr != rexpr2)
	    return new PtreeMemberAccessExpr(object2, op, member,
					     First(rest), rexpr2);
	else
	    return this;
    }
    else{
	t.Dereference();
	t.DereferenceIfRefType();
	Ptree* arglist = First(rest);
	Ptree* arglist2 = Class::TranslateArgumentList(env, arglist);
	if(object != object2 || arglist != arglist2)
	    return new PtreeMemberAccessExpr(object2, op, member, arglist2);
	else
	    return this;
    }
}

void PtreeMemberAccessExpr::Type(Environment* env, TypeInfo& t)
{
    TypeInfo t1;
    Ptree* object;
    Ptree* op;
    Ptree* member;
    Ptree* rest;

    if(!Match(this, "[%? %? %? %r]", &object, &op, &member, &rest))
	MopErrorMessage("PtreeMemberAccessExpr::Type()", "bomb!");

    object->Type(env, t1);
    if(!op->Eq('.'))
	t1.Dereference();	// the operator is ->

    uint whatis = Syntax::LookupMemberType(member, t1, t);
    if(whatis == DataMemberName || whatis == MemberFunctionName){
	t.DereferenceIfRefType();
	if(rest != nil && Length(rest) != 2){
	    t.Dereference();
	    t.DereferenceIfRefType();
	}
    }
    else
	t.Undef();
}

Ptree* PtreePtrToMemExpr::WalkExpr(Environment* env, TypeInfo& t)
{
    TypeInfo t1;
    Ptree* lexpr = First(this);
    Ptree* lexpr2 = lexpr->WalkExpr(env, t1);
    Ptree* rexpr = Third(this);
    Ptree* rexpr2 = rexpr->WalkExpr(env, t);
    t.Dereference();

    if(lexpr != lexpr2 || rexpr != rexpr2)
	return new PtreePtrToMemExpr(lexpr2, Second(this), rexpr2);
    else
	return this;
}

void PtreePtrToMemExpr::Type(Environment* env, TypeInfo& t)
{
    Third(this)->Type(env, t);
    t.Dereference();
}

Ptree* PtreeSizeofExpr::WalkExpr(Environment* env, TypeInfo& t)
{
    TypeInfo t1;
    Ptree* expr = Second(this);
    Ptree* expr2 = expr->WalkExpr(env, t1);
    t.BeInt();
    if(expr != expr2)
	return new PtreeSizeofExpr(this->Car(), expr2);
    else
	return this;
}

void PtreeSizeofExpr::Type(Environment* env, TypeInfo& t)
{
    t.BeInt();
}

Ptree* PtreeSizeofType::WalkExpr(Environment* env, TypeInfo& t)
{
    Ptree* paren = Second(this);
    Ptree* type = Second(paren);
    Ptree* type2 = WalkCastType(env, type);
    t.BeInt();
    if(type != type2)
	return new PtreeSizeofType(this->Car(),
				   new PtreeParen(First(paren), type2,
						  Third(paren)));
    else
	return this;
}

void PtreeSizeofType::Type(Environment* env, TypeInfo& t)
{
    t.BeInt();
}

Ptree* PtreeFstyleCast::WalkExpr(Environment* env, TypeInfo& t)
{
    Ptree* type = First(this);
    Syntax::AnalyzeFstyleCast(type, env, t);
    Ptree* type2 = WalkTypeSpecifier(env, type);

    Ptree* args = Second(this);
    Ptree* expr = Second(args);
    Ptree* expr2 = WalkExprList(env, expr);

    if(type == type2 && expr == expr2)
	return this;
    else
	return new PtreeFstyleCast(type2,
				   new PtreeParen(First(args), expr2,
						  Third(args)));
}

void PtreeFstyleCast::Type(Environment* env, TypeInfo& t)
{
    Syntax::AnalyzeFstyleCast(First(this), env, t);
}

Ptree* PtreeUserStatement::WalkExpr(Environment* env, TypeInfo& t)
{
    TypeInfo t1;
    Ptree* object;
    Ptree* op;
    Ptree* keyword;
    Ptree* rest;

    t.BeVoid();

    if(!Match(this, "[%? %? %? %r]", &object, &op, &keyword, &rest))
	MopErrorMessage("PtreeUserStatement::WalkExpr()", "bomb!");

    object->Type(env, t1);

    if(!op->Eq('.'))
	t1.Dereference();	// the operator is ->

    Class* metaobject = t1.ClassMetaobject();
    if(metaobject == nil){
	ErrorMessage("the metaobject cannot handle this statement", this);
	return this;
    }
    else
	return metaobject->TranslateUserStatement(env, object, op,
						  keyword, rest);
}

void PtreeUserStatement::Type(Environment* env, TypeInfo& t)
{
    t.BeVoid();
}

Ptree* PtreeStaticUserStatement::WalkExpr(Environment* env, TypeInfo& t)
{
    TypeInfo t1;
    Ptree* qclass;
    Ptree* keyword;
    Ptree* rest;

    t.BeVoid();

    if(!Match(this, "[%? [%? %r]]", &qclass, &keyword, &rest))
	MopErrorMessage("PtreeStaticUserStatement::WalkExpr()", "bomb!");

    Syntax::AnalyzeQClass(qclass, env, t1);
    Class* metaobject = t1.ClassMetaobject();
    if(metaobject == nil){
	ErrorMessage("the metaobject cannot handle this statement", this);
	return this;
    }
    else
	return metaobject->TranslateStaticUserStatement(env, keyword, rest);
}

void PtreeStaticUserStatement::Type(Environment* env, TypeInfo& t)
{
    t.BeVoid();
}

// initialize expression
//
Ptree* PtreeBrace::WalkExpr(Environment* env, TypeInfo& t)
{
    Ptree* expr = Second(this);
    Ptree* expr2 =  WalkExprList(env, expr);
    t.Undef();		// not implemented yet
    if(expr == expr2)
	return this;
    else
	return new PtreeBrace(First(this), expr2, Third(this));
}

void PtreeBrace::Type(Environment* env, TypeInfo& t)
{
    t.Undef();		// not implemented yet
}

Ptree* PtreeActualArgs::WalkExpr(Environment* env, TypeInfo& t)
{
    t.Undef();
    return WalkExprList(env, this);
}

void PtreeActualArgs::Type(Environment* env, TypeInfo& t)
{
    t.Undef();
}

Ptree* Ptree::WalkExprList(Environment* env, Ptree* expr)
{
    return Map2(env, expr, &Ptree::WalkExprListCore);
}

Ptree* Ptree::WalkExprListCore(Environment* env, Ptree* subexpr)
{
    TypeInfo t;

    if(subexpr == nil)
	return subexpr;
    else
	return subexpr->WalkExpr(env, t);
}

/*
  CheckEquiv() compares all the elements of the two lists,
  and returns origlst if they are the same, otherwise
  returns newlst.
*/
Ptree* Ptree::CheckEquiv(Ptree* origlst, Ptree* newlst)
{
    Ptree* lst1 = origlst;
    Ptree* lst2 = newlst;
    while(lst1 != nil && lst2 != nil)
	if(lst1->Car() != lst2->Car())
	    return newlst;
	else{
	    lst1 = lst1->Cdr();
	    lst2 = lst2->Cdr();
	}

    if(lst1 == nil && lst2 == nil)
	return origlst;
    else
	return newlst;
}
